package nextapp.echooff;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.mozilla.javascript.ErrorReporter;
import org.mozilla.javascript.EvaluatorException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import com.yahoo.platform.yui.compressor.JavaScriptCompressor;

public class EchoOff {
    
    private static final String DEFAULT_VIEWPORT_TAG 
            = "<meta name = \"viewport\" content = \"width=device-width, user-scalable=no\"/>";
    
    private static final Map EXTENSION_TO_CONTENT_TYPE;
    static {
        Map m = new HashMap();
        m.put("png", "image/png");
        m.put("jpg", "image/jpeg");
        m.put("jpeg", "image/jpeg");
        m.put("gif", "image/gif");
        m.put("html", "text/html");
        m.put("htm", "text/html");
        EXTENSION_TO_CONTENT_TYPE = Collections.unmodifiableMap(m);
    }
    
    private static Element buildElement;
    private static File workingDirectory;
    
    private static String compress(String js) 
    throws IOException {

        JavaScriptCompressor compressor = new JavaScriptCompressor(new StringReader(js), new ErrorReporter() {

            public void warning(String message, String sourceName,
                    int line, String lineSource, int lineOffset) {
                if (line < 0) {
                    System.err.println("\n[WARNING] " + message);
                } else {
                    System.err.println("\n[WARNING] " + line + ':' + lineOffset + ':' + message);
                }
            }

            public void error(String message, String sourceName,
                    int line, String lineSource, int lineOffset) {
                if (line < 0) {
                    System.err.println("\n[ERROR] " + message);
                } else {
                    System.err.println("\n[ERROR] " + line + ':' + lineOffset + ':' + message);
                }
            }

            public EvaluatorException runtimeError(String message, String sourceName,
                    int line, String lineSource, int lineOffset) {
                error(message, sourceName, line, lineSource, lineOffset);
                return new EvaluatorException(message);
            }
        });
        
        StringWriter w = new StringWriter();
        compressor.compress(w, -1, true, false, true, false);
        return w.toString();
    }
    
    private static String createApplicationHtml()
    throws IOException, SAXException {
        StringBuffer html = new StringBuffer();

        html.append("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\""
                + "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n");
        html.append("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n");
        html.append("<head>\n");
        
        Element titleElement = DomUtil.getChildElementByTagName(buildElement, "title");
        if (titleElement != null) {
            html.append(DEFAULT_VIEWPORT_TAG);
            html.append("<title>");
            html.append(DomUtil.getElementText(titleElement));
            html.append("</title>\n");
        }
        
        Element iconElement = DomUtil.getChildElementByTagName(buildElement, "icon");
        if (iconElement != null) {
            html.append("<link rel=\"apple-touch-icon\" href=\"");
            html.append(toData(DomUtil.getElementText(iconElement)));
            html.append("\"/>\n");
        }

        html.append("<script type=\"text/javascript\">\n");
        
        Element[] libElements = DomUtil.getChildElementsByTagName(buildElement, "lib");
        StringBuffer js = new StringBuffer();
        for (int i = 0; i < libElements.length; ++i) {
            js.append(getLibrary(DomUtil.getElementText(libElements[i])));
        }
        html.append(compress(js.toString()));
        
        Element[] imageNamespaceElements = DomUtil.getChildElementsByTagName(buildElement, "image-namespace");
        for (int i = 0; i < imageNamespaceElements.length; ++i) {
            html.append(imageNamespaceElements[i].getAttribute("name"));
            html.append(" = {\n");
            Element[] imageElements = DomUtil.getChildElementsByTagName(imageNamespaceElements[i], "image");
            for (int j = 0; j < imageElements.length; ++j) {
                String name = imageElements[j].getAttribute("name");
                html.append(name);
                String imageUrl = DomUtil.getElementText(imageElements[j]);
                html.append(": \"");
                html.append(toData(imageUrl));
                html.append("\"");
                if (j < imageElements.length - 1) {
                    html.append(",\n");
                } else {
                    html.append("\n");
                }
            }
            html.append("};\n");
        }
        html.append("</script>\n");
        
        html.append("</head>\n");
        html.append("<body onload=\"init();\">\n");
        html.append("<div id=\"rootArea\"");
        html.append(">\n");
        html.append("</body>\n");
        html.append("</html>\n");
        
        return html.toString();
    }
    
    private static String createApplicationUrl(String appHtml) 
    throws IOException, SAXException {
        String base64 = Base64.encodeBytes(appHtml.getBytes(), Base64.DONT_BREAK_LINES);
        
        StringBuffer out = new StringBuffer();
        out.append("data:text/html;base64,");
        out.append(base64);
        
        return out.toString();
    }
    
    private static String createRedirectHtml(String appUrl)
    throws IOException, SAXException {
        StringBuffer html = new StringBuffer();

        html.append("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\""
                + "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n");
        html.append("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n");
        html.append("<head>\n");
        
        Element titleElement = DomUtil.getChildElementByTagName(buildElement, "title");
        if (titleElement != null) {
            html.append("<title>");
            html.append(DomUtil.getElementText(titleElement));
            html.append("</title>\n");
        }
        html.append("<script type=\"text/javascript\">\n");
        
        html.append("function init() {\n");
        html.append("window.location.href = \"" + appUrl + "\";");
        html.append("}\n");
        
        html.append("</script>\n");
        html.append("</head>\n");
        html.append("<body onload=\"init();\">\n");
        html.append("</body>\n");
        html.append("</html>\n");
        
        return html.toString();
    }
    
    private static String createStartPage(File redirectFile) {
        StringBuffer html = new StringBuffer();

        html.append("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\""
                + "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n");
        html.append("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n");
        html.append("<head>\n");
        
        Element titleElement = DomUtil.getChildElementByTagName(buildElement, "title");
        String title = null;
        if (titleElement != null) {
            title = DomUtil.getElementText(titleElement);
            html.append(DEFAULT_VIEWPORT_TAG);
            html.append("<title>");
            html.append(title);
            html.append("</title>\n");
        }
        html.append("</head>\n");
        html.append("<body>\n");
        html.append("<a href=\"");
        html.append(redirectFile == null ? "#" : redirectFile.getName());
        html.append("\">Start ");
        html.append(title == null ? "Application" : title);
        html.append("</a>\n");
        html.append("</body>\n");
        html.append("</html>\n");
        
        return html.toString();
    }

    private static String getLibrary(String url) 
    throws IOException {
        File libraryFile = new File(workingDirectory, url);
        BufferedReader reader = new BufferedReader(new FileReader(libraryFile));
        String line;
        StringBuffer out = new StringBuffer();
        while ((line = reader.readLine()) != null) {
            out.append(line);
            out.append("\n");
        }
        return out.toString();
    }
    
    private static String getContentType(String imageUrl) {
        String extension = imageUrl.substring(imageUrl.lastIndexOf(".") + 1).toLowerCase();
        return (String) EXTENSION_TO_CONTENT_TYPE.get(extension);
    }
    
    private static byte[] getFileData(String url) 
    throws IOException {
        File libraryFile = new File(workingDirectory, url);
        byte[] data = new byte[(int) libraryFile.length()];
        
        int offset = 0;
        int readBytes = 0;
        InputStream in = new FileInputStream(libraryFile);
        while (offset < data.length && (readBytes = in.read(data, offset, data.length - offset)) >= 0) {
            offset += readBytes;
        }
        
        return data;
    }
    
    private static String toData(String url) 
    throws IOException {
        
        String base64 = Base64.encodeBytes(getFileData(url), Base64.DONT_BREAK_LINES);
        StringBuffer out = new StringBuffer();
        out.append("data:");
        out.append(getContentType(url));
        out.append(";base64,");
        out.append(base64);
        return out.toString();
    }
    
    public static void saveText(String text, File file) 
    throws IOException {
        File parent = file.getParentFile();
        if (parent != null && !parent.exists()) {
            parent.mkdirs();
        }
        FileWriter fw = new FileWriter(file);
        fw.write(text);
        fw.flush();
        fw.close();
    }
    
    public static void main(String[] args) 
    throws Exception {
        if (args.length != 1) {
            System.err.println("Invalid arguments: specify location of offline-build.xml file.");
            System.exit(1);
        }
        
        File file = new File(args[0]);
        if (!file.exists()) {
            System.err.println("offline-build.xml file \"" + args[0] + "\" cannot be found.");
            System.exit(1);
        }
        workingDirectory = file.getParentFile();
        
        InputStream in = new FileInputStream(file);
        Document dom = DomUtil.load(in);
        
        buildElement = dom.getDocumentElement();
        
        String appHtml = createApplicationHtml();
        
        Element saveHtmlElements[] = DomUtil.getChildElementsByTagName(buildElement, "save-html");
        for (int i = 0; i < saveHtmlElements.length; ++i) {
            File outputFile = new File(workingDirectory, DomUtil.getElementText(saveHtmlElements[i]));
            saveText(appHtml, outputFile);
            System.out.println("Stored application HTML in file: " + outputFile.getAbsolutePath());
        }

        String url = createApplicationUrl(appHtml);
        Element saveUrlElements[] = DomUtil.getChildElementsByTagName(buildElement, "save-url");
        for (int i = 0; i < saveUrlElements.length; ++i) {
            File outputFile = new File(workingDirectory, DomUtil.getElementText(saveUrlElements[i]));
            saveText(url, outputFile);
            System.out.println("Stored URL in file: " + outputFile.getAbsolutePath());
        }
        
        File redirectFile = null;
        String redirectHtml = createRedirectHtml(url);
        Element saveRedirectElements[] = DomUtil.getChildElementsByTagName(buildElement, "save-redirect");
        for (int i = 0; i < saveRedirectElements.length; ++i) {
            File outputFile = new File(workingDirectory, DomUtil.getElementText(saveRedirectElements[i]));
            saveText(redirectHtml, outputFile);
            if (i == 0) {
                redirectFile = outputFile;
            }
            System.out.println("Stored redirect HTML in file: " + outputFile.getAbsolutePath());
        }

        String startPageHtml = createStartPage(redirectFile);
        Element saveStartPageElements[] = DomUtil.getChildElementsByTagName(buildElement, "save-start-page");
        for (int i = 0; i < saveStartPageElements.length; ++i) {
            File outputFile = new File(workingDirectory, DomUtil.getElementText(saveStartPageElements[i]));
            saveText(startPageHtml, outputFile);
            System.out.println("Stored start page HTML in file: " + outputFile.getAbsolutePath());
        }
    }
}
