/**
 * @fileoverview
 * Provides low-level core functionality.  Non-instantiable object.  Requires nothing.
 * <p>
 * Provides core APIs for creating object-oriented and event-driven JavaScript code.  Features include:
 * <ul>
 *  <li>Provides API for declaring JavaScript classes which includes support for
 *   specifying abstract and virtual properties and validating subtypes to such
 *   specification.</li>
 *  <li>Provides a "Method Wrapper" function (Core.method()) to create a function which will invoke 
 *    a member function of a specific object instance (enabling invocation with the "this pointer" set
 *    appropriately).</li>
 *  <li>Provides event/listener management framework.  Event-listeners which invoke
 *    methods of an object instance may be created using the Core.method() function.</li>
 *  <li>Provides a "Large Map" useful for managing an associative array that is frequently modified
 *    and will exist for a long period of time.  This object is unfortunately necessary due to
 *    issues present in certain clients (Internet Explorer 6 memory leak / performance degredation).</li>
 *  <li>Provides array manipulation utilities.<li>
 *  <li>Provides some simple debugging utilities, e.g., a pseudo-console output.</li>
 *  <li>Does not provide any web-specific functionality.</li>
 * </ul>
 */

/**
 * Namespace for core functionality.
 */
Core = {

    /**
     * Creates a duplicate copy of a function.
     * Per the ECMA-262 v3 specification, Function.toString() is required to return an (implementation specific)
     * string representation of the function.
     * Creating a copy of a constructor is more efficient than invoking Function.apply() in certain browsers
     * (a significant performance improvement was observed in Internet Explorer 6).
     *
     * @param f the function
     * @return an identical copy
     * @private
     */
    _copyFunction: function(f) {
        var fCopy;
        eval("fCopy = " + f.toString() + ";");
        return fCopy;
    },
    
    /**
     * Creates an empty function.
     *
     * @private
     */
    _createFunction: function() {
        return function() { };
    },
    
    /**
     * Creates a new class, optionally extending an existing class.
     * This method may be called with one or two parameters as follows:
     * <p>
     * Core.extend(definition)
     * Core.extend(baseClass, definition)
     * <p>
     * Each property of the definition object will be added to the prototype of the returned defined class.
     * Properties that begin with a dollar-sign ($) will be processed specially:
     * <p>
     * <ul>
     * <li>The $construct property, which must be a function, will be used as the constructor.
     * The $load property, which must be a function, f provided, will be used as a static initializer,
     * executed once when the class is *defined*.  The this pointer will be set to the class when
     * this method is executed.</li>
     * <li>The $static property, an object, if provided, will have its properties installed as class variables.</li>
     * <li>The $abstract property, an object or 'true', if provided, will define methods that must be implemented
     * by derivative classes.  If the value is simply true, the object will be marked as abstract (such that
     * it does not necessarily need to provide implementations of abstract methods defined in its base class.</li>
     * <li>The $virtual property, an object, if provided, defines methods that will be placed into the prototype
     * that may be overridden by subclasses.  Attempting to override a property/method of the superclass that
     * is not defined in the virtual block will result in an exception.  Having the default behavior NOT allow
     * for overriding ensures that namespacing between super- and sub-types if all internal variables are instance
     * during Core.extend().</li>
     * </ul>
     * <p>
     * Use of this method enables a class to be derived WITHOUT executing the constructor of the base class
     * in order to create a prototype for the derived class.  This method uses a "shared prototype" architecture,
     * where two objects are created, a "prototype class" and a "constructor class".  These two objects share
     * the same prototype, but the "prototype class" has an empty constructor.  When a class created with
     * this method is derived, the "prototype class" is used to create a prototype for the derivative.
     * <p>
     * This method will return the constructor class, which contains an internal reference to the 
     * prototype class that will be used if the returned class is later derived by this method.
     * 
     * @param {Function} baseClass the base class
     * @param {Object} definition an associative array containing methods and properties of the class
     * @return the constructor class
     */
    extend: function() {
        // Configure baseClass/definition arguments.
        var baseClass = arguments.length == 1 ? null : arguments[0];
        var definition = arguments.length == 1 ? arguments[0] : arguments[1];
        
        // Perform argument error checking.
        if (baseClass) {
            if (typeof(baseClass) != "function") {
                throw new Error("Base class is not a function, cannot derive.");
            }
        }
        if (!definition) {
            throw new Error("Object definition not provided.");
        }
        
        // Create the constructor class.
        var constructorClass;
        if (definition.$construct) {
            // Definition provides constructor, provided constructor function will be used as object.
            constructorClass = definition.$construct;
            
            // Remove property such that it will not later be added to the object prototype.
            delete definition.$construct;
        } else {
            // Definition does not provide constructor.
            if (baseClass) {
                // Base class available: copy constructor function from base class.
                // Note: should function copying not be supported by a future client,
                // it is possible to simply create a new constructor which invokes the base
                // class constructor (using closures and Function.apply()) to achieve the
                // same effect (with a slight performance penalty).
                constructorClass = Core._copyFunction(baseClass);
            } else {
                // No base class: constructor is an empty function.
                constructorClass = Core._createFunction();
            }
        }
        
        // Create virtual property storage.
        constructorClass.$virtual = {};
        
        // Store reference to base class in constructor class.
        constructorClass.$super = baseClass;

        if (baseClass) {
            // Create class with empty constructor that shares prototype of base class.
            var prototypeClass = Core._createFunction();
            prototypeClass.prototype = baseClass.prototype;
            
            // Create new instance of constructor-less prototype for use as prototype of new class.
            constructorClass.prototype = new prototypeClass();
        }
        
        // Assign constructor correctly.
        constructorClass.prototype.constructor = constructorClass;

        // Add abstract properties.
        if (definition.$abstract) {
            constructorClass.$abstract = {};
            if (baseClass && baseClass.$abstract) {
                // Copy abstract properties from base class.
                for (var x in baseClass.$abstract) {
                    constructorClass.$abstract[x] = baseClass.$abstract[x];
                }
            }

            if (definition.$abstract instanceof Object) {
                // Add abstract properties from definition.
                for (var x in definition.$abstract) {
                    constructorClass.$abstract[x] = true;
                    constructorClass.$virtual[x] = true;
                }
            }
            
            // Remove property such that it will not later be added to the object prototype.
            delete definition.$abstract;
        }
        
        // Copy virtual property flags from base class to shared prototype.
        if (baseClass) {
            for (var name in baseClass.$virtual) {
                constructorClass.$virtual[name] = baseClass.$virtual[name];
            }
        }
        
        // Add virtual instance properties from definition to shared prototype.
        if (definition.$virtual) {
            Core._inherit(constructorClass.prototype, definition.$virtual, constructorClass.$virtual);
            for (var name in definition.$virtual) {
                constructorClass.$virtual[name] = true;
            }

            // Remove property such that it will not later be added to the object prototype.
            delete definition.$virtual;
        }
        
        // Add toString and valueOf manually, as they will not be iterated
        // by for-in iteration in Internet Explorer.
        if (definition.hasOwnProperty("toString")) {
            constructorClass.prototype.toString = definition.toString;
        }
        if (definition.hasOwnProperty("valueOf")) {
            constructorClass.prototype.valueOf = definition.valueOf;
        }

        // Remove properties such that they will not later be added to the object prototype.
        delete definition.toString;
        delete definition.valueOf;

        // Add Mixins.
        if (definition.$include) {
            // Reverse order of mixins, such that later-defined mixins will override earlier ones.
            // (Mixins will only be added if they will NOT override an existing method.)
            var mixins = definition.$include.reverse();
            Core._processMixins(constructorClass, mixins);
            
            // Remove property such that it will not later be added to the object prototype.
            delete definition.$include;
        }

        // Store $load static initializer and remove from definition so it is not inherited in static processing.
        var loadMethod = null;
        if (definition.$load) {
            loadMethod = definition.$load;

            // Remove property such that it will not later be added to the object prototype.
            delete definition.$load;
        }
        
        // Process static properties and methods defined in the '$static' object.
        if (definition.$static) {
            Core._inherit(constructorClass, definition.$static);

            // Remove property such that it will not later be added to the object prototype.
            delete definition.$static;
        }

        // Process instance properties and methods.
        Core._inherit(constructorClass.prototype, definition, constructorClass.$virtual);
        
        // If class is concrete, verify all abstract methods are provided.
        if (!constructorClass.$abstract) {
            this._verifyAbstractImpl(constructorClass);
        }
        
        // Invoke static constructors.
        if (loadMethod) {
            // Invoke $load() function with "this" pointer set to class.
            loadMethod.call(constructorClass);
        }
        
        return constructorClass;
    },
    
    /**
     * Retrieves a value from an object hierarchy.
     *
     * Examples: 
     * Given the following object 'o': { a: { b: 4, c: 2 }}
     * Core.get(o, ["a", "b"]) will return 4.
     * Core.get(o, ["a", "c"]) will return 2.
     * Core.get(o, ["a", "d"]) will return null.
     * Core.get(o, ["a"]) will return { b: 4, c: 2 }.
     * Core.get(o, ["b"]) will return null.
     * Core.get(o, ["d"]) will return null.
     *
     * @param object an arbitrary object from which the value should be retrieved
     * @param {Array} path an array of object property names describing the path to retrieve
     * @return the value, if found, or null if it does not exist
     */
    get: function(object, path) {
        for (var i = 0; i < path.length; ++i) {
            object = object[path[i]];
            if (!object) {
                return null;
            }
        }

        return object;
    },
    
    /**
     * Determines if the specified propertyName of the specified object is a virtual
     * property, i.e., that it can be overridden by subclasses.
     */
    _isVirtual: function(virtualProperties, propertyName) {
        switch (propertyName) {
        case "toString":
        case "valueOf":
            return true;
        }
        
        return virtualProperties[propertyName];
    },
    
    /**
     * Installs properties from source object into destination object.
     * <p>
     * In the case where the destination object already has a property defined
     * and the "virtualProperties" argument is provided, the "virtualProperties"
     * collection will be checked to ensure that property is allowed to be
     * overridden.  If "virtualProperties" is omitted, any property may be
     * overridden.
     *
     * @param destination the destination object
     * @param soruce the source object
     * @param virtualProperties (optional) collection of virtual properties from base class.
     * @private  
     */
    _inherit: function(destination, source, virtualProperties) {
        for (var name in source) {
            if (virtualProperties && destination[name] && !this._isVirtual(virtualProperties, name)) {
                // Property exists in destination as is not marked as virtual.
                throw new Error("Cannot override non-virtual property \"" + name + "\".");
            } else {
                destination[name] = source[name];
            }
        }
    },
    
    /**
     * Creates a new function which executes a specific method of an object instance.
     * Any arguments passed to the returned function will be passed to the method.
     * The return value of the method will be returned by the function.
     *
     * CAUTION: When adding and removing methods as listeners, not that two seperately
     * constructed methods will not be treated as equal, even if their instance and method
     * properties are the same.  Failing to heed this warning can result in a memory leak,
     * as listeners would never be removed.
     *
     * @param the object instance
     * @param {Function} method the method to be invoked on the instance
     * @return the return value provided by the method
     */
    method: function(instance, method) {
        return function() {
            return method.apply(instance, arguments);
        };
    },
    
    /**
     * Add properties of mixin objects to destination object.
     * Mixins will be added in order, and any property which is already
     * present in the destination object will not be overridden.
     *
     * @param destination the destination object
     * @param {Array} the mixin objects to add 
     */
    _processMixins: function(destination, mixins) {
        for (var i = 0; i < mixins.length; ++i) {
            for (var mixinProperty in mixins[i]) {
                if (destination.prototype[mixinProperty]) { 
                    // Ignore mixin properties that already exist.
                    continue;
                }
                destination.prototype[mixinProperty] = mixins[i][mixinProperty];
            }
        }
    },
    
    /**
     * Retrieves a value from an object hierarchy.
     *
     * Examples: 
     * Given the following object 'o': { a: { b: 4, c: 2 } }
     * Core.set(o, ["a", "b"], 5) will update the value of 'o' to be: { a: { b: 5, c: 2 } }
     * Core.set(o, ["a", "d"], 7) will update the value of 'o' to be: { a: { b: 4, c: 2, d: 7 } }
     * Core.set(o, ["e"], 9) will update the value of 'o' to be: { a: { b: 4, c: 2 }, e: 9 }
     * Core.set(o, ["f", "g"], 8) will update the value of 'o' to be: { a: { b: 4, c: 2 }, f: { g: 8 } }
     * Core.set(o, ["a"], 10) will update the value of 'o' to be: { a: 10 }
     *
     * @param object an arbitrary object from which the value should be retrieved
     * @param {Array} path an array of object property names describing the path to retrieve
     * @return the value, if found, or null if it does not exist
     */
    set: function(object, path, value) {
        var parentObject = null;
        
        // Find or create container object.
        for (var i = 0; i < path.length - 1; ++i) {
            parentObject = object; 
            object = object[path[i]];
            if (!object) {
                object = {};
                parentObject[path[i]] = object;
            }
        }
        
        // Assign value.
        object[path[path.length - 1]] = value;
    },
    
    /**
     * Verifies that a concrete derivative of an abstract class implements
     * abstract properties present in the base class.
     *
     * @param constructorClass the class to verify
     */
    _verifyAbstractImpl: function(constructorClass) {
         var baseClass = constructorClass.$super;
         if (!baseClass || !baseClass.$abstract || baseClass.$abstract === true) {
             return;
         }
         
         for (var x in baseClass.$abstract) {
             if (!constructorClass.prototype[x]) {
                 throw new Error("Concrete class does not provide implementation of abstract method \"" + x + "\".");
             }
         }
    }
};

/**
 * @class 
 * Namespace for debugging related utilities.
 */
Core.Debug = { 

    /**
     * The DOM element to which console output should be written.
     * @type HTMLElement
     */
    consoleElement: null,
    
    /**
    * Flag indicating whether console output should be displayed as alerts.
    * Enabling is generally not recommended.
    * @type Boolean
    */
    useAlertDialog: false,
    
    /**
     * Writes a message to the debug console.
     * 
     * @param {String} text the message
     */
    consoleWrite: function(text) {
        if (Core.Debug.consoleElement) {
            var entryElement = document.createElement("div");
            entryElement.appendChild(document.createTextNode(text));
            if (Core.Debug.consoleElement.childNodes.length == 0) {
                Core.Debug.consoleElement.appendChild(entryElement);
            } else {
                Core.Debug.consoleElement.insertBefore(entryElement, Core.Debug.consoleElement.firstChild);
            }
        } else if (Core.Debug.useAlertDialog) {
            alert("DEBUG:" + text);
        }
    },
    
    /**
     * Creates a string representation of the state of an object's instance variables.
     *
     * @param object the object to convert to a string
     * @return the string
     * @type String
     */
    toString: function(object) {
        var s = "";
        for (var x in object) {
            if (typeof object[x] != "function") { 
                s += x + ":" + object[x] + "\n";
            }
        }
        return s;
    }
};

/**
 * @class 
 * Arrays namespace.  
 * Non-instantiable object.
 */
Core.Arrays = {

    /**
     * Returns <tt>true</tt> if the first array contains all of the elements
     * in the second array.
     *
     * @param {Array} array1 array to be checked for containment
     * @param {Array} array2 array to be checked for containment in the first array
     * @param {Boolean} optional flag indicating that all elements in array2 are unique
     * @return <tt>true</tt> if the first array contains all of the elements
     *         in the second array
     * @type Boolean
     */
    containsAll: function(array1, array2, unique) {
        if (unique && array1.length < array2.length) {
            return false;
        }
        if (array2.length == 0) {
            return true;
        }
        var found, item;
        for (var i = 0; i < array2.length; ++i) {
            found = false;
            item = array2[i];
            for (var j = 0; j < array1.length; ++j) {
                if (item == array1[j]) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                return false;
            }
        }
        return true;
    },

    /**
     * Returns the index of the specified item within the array, or -1 if it 
     * is not contained in the array.  
     * 
     * @param item the item
     * @return the index of the item, or -1 if it is not present in the array
     * @type Number
     */
    indexOf: function(array, item) {
        for (var i = 0; i < array.length; ++i) {
            if (item == array[i]) {
                return i;
            }
        }
        return -1;
    },
    
    /**
     * Removes the first instance of the specified item from an array.
     * If the item does not exist in the array, no action is taken.
     * Equality is determined using the '==' operator.
     * 
     * @param array the array from which the item should be removed
     * @param item the item to remove
     */
    remove: function(array, item) {
        for (var i = 0; i < array.length; ++i) {
            if (item == array[i]) {
                array.splice(i, 1);
                return;
            }
        }
    },
    
    /**
     * Removes duplicate items from an array.
     * Items retained in the array may not appear in the previous order.
     * 
     * @param array the array from which duplicates are to be removed.
     */
    removeDuplicates: function(array) {
        array.sort();
        var removeCount = 0;
        // Iterate from last element to second element.
        for (var i = array.length - 1; i > 0; --i) {
            // Determine if element is equivalent to previous element.
            if (array[i] == array[i - 1]) {
                // If duplicate, copy last element in array over current element.
                array[i] = array[array.length - 1 - removeCount];
                
                // Increment removeCount (indicating how much the length of the array should be decremented)
                ++removeCount;
            }
        }
        
        if (removeCount > 0) {
            array.length = array.length - removeCount;
        }
    }
};

/**
 * @class Associative array wrapper which periodically recreates the associative array
 *        in order to avoid memory leakage and performance problems on certain browser
 *        platforms, i.e., Internet Explorer 6.
 *        Null values are not permitted as keys.  Setting a key to a null value
 *        will result in the key being removed.
 */
Core.Arrays.LargeMap = Core.extend({
    
    $static: {
        garbageCollectEnabled: false
    },
    
    /**
     * Number of removes since last associative array re-creation.
     * @type Number
     * @private
     */
    _removeCount: 0,
    
    /**
     * Number (integer) of removes between associative array re-creation.
     * @type Number
     */
    garbageCollectionInterval: 250,
    
    /**
     * Associative mapping.
     */
    map: null, 
    
    /**
     * Creates a new LargeMap.
     */
    $construct: function() {
        this.map = {};
    },
    
    /**
     * Performs 'garbage-collection' operations, recreating the array.
     * This operation is necessary due to Internet Explorer memory leak
     * issues.
     */
    _garbageCollect: function() {
        this._removeCount = 0;
        var newMap = {};
        for (var key in this.map) {
            newMap[key] = this.map[key];
        }
        this.map = newMap;
    },
    
    /**
     * Removes the value referenced by the specified key.
     *
     * @param key the key
     */
    remove: function(key) {
        delete this.map[key];
        if (Core.Arrays.LargeMap.garbageCollectEnabled) {
            ++this._removeCount;
            if (this._removeCount >= this.garbageCollectionInterval) {
                this._garbageCollect();
            }
        }
    }
});

/**
 * @class A collection of event listeners.  Provides capability to manage listeners
 *        of multiple types, and fire events to listeners based on type.
 */
Core.ListenerList = Core.extend({

    /**
     * Array containing event types and event listeners.  
     * Even indexes contain event types, and the subsequent odd
     * index contain Functions to invoke.
     * @type Array
     */
    _data: null,
   
    /**
     * Creates a new listener list.
     * 
     * @constructor
     */
    $construct: function() {
        this._data = [];
    },

    /**
     * Adds an event listener.
     * 
     * @param {String} eventType the event type
     * @param {Function} eventTarget the event target
     */
    addListener: function(eventType, eventTarget) {
        this._data.push(eventType, eventTarget);
    },
    
    /**
     * Fires an event.
     * 
     * @param event the event to fire
     * @return true if all event listeners returned values that evaluate to true, 
     *         or false if any event listeners returned values that evaluate to 
     *         false
     * @type Boolean
     */
    fireEvent: function(event) {
        if (event.type == null) {
            throw new Error("Cannot fire event, type property not set.");
        }
        
        var listeners = [];
        for (var i = 0; i < this._data.length; i += 2) {
            if (this._data[i] == event.type) {
                listeners.push(this._data[i + 1]);
            }
        }
        
        var returnValue = true;
        for (var i = 0; i < listeners.length; ++i) {
            returnValue = listeners[i](event) && returnValue; 
        }
        return returnValue;
    },
    
    /**
     * Returns an array containing the types of all listeners
     * in the list.
     * 
     * @return the event types
     * @type Array
     */
    getListenerTypes: function() {
        var types = [];
        for (var i = 0; i < this._data.length; i += 2) {
            types.push(this._data[i]);
        }
        Core.Arrays.removeDuplicates(types);
        return types;
    },
    
    /**
     * Returns an array of all listeners for a specific event type.
     * 
     * @param {String} eventType the event type
     * @return the listeners
     * @type Array
     */
    getListeners: function(eventType) {
        var listeners = [];
        for (var i = 0; i < this._data.length; i += 2) {
            if (this._data[i] == eventType) {
                listeners.push(this._data[i + 1]);
            }
        }
        return listeners;
    },
    
    /**
     * Determines the number of listeners for a specific event type.
     * 
     * @param {String} eventType the event type
     * @return the listener count
     * @type Number
     */
    getListenerCount: function(eventType) {
        var count = 0;
        for (var i = 0; i < this._data.length; i += 2) {
            if (this._data[i] == eventType) {
                ++count;
            }
        }
        return count;
    },
    
    /**
     * Determines if the listeners list has any listeners of a specific type.
     * 
     * @param {String} eventType the event type
     * @return true if any listeners exist
     * @type Boolean
     */
    hasListeners: function(eventType) {
        for (var i = 0; i < this._data.length; i += 2) {
            if (this._data[i] == eventType) {
                return true;
            }
        }
        return false;
    },
    
    /**
     * Determines if any number of listeners are registered to the list.
     * 
     * @return true if the listener list is empty
     * @type Boolean
     */
    isEmpty: function() {
        return this._data.length == 0;
    },
    
    /**
     * Removes an event listener.
     * 
     * @param {String} eventType the event type
     * @param {Function} eventTarget the event target
     */
    removeListener: function(eventType, eventTarget) {
        for (var i = 0; i < this._data.length; i += 2) {
            if (this._data[i] == eventType && eventTarget == this._data[i + 1]) {
                var oldLength = this._data.length;
                this._data.splice(i, 2);
                return;
            }
        }
    },
    
    /**
     * Creates a string representation of the listener list.
     * This should be used for debugging purposes only.
     */
    toString: function() {
        var out = "";
        for (var i = 0; i < this._data.length; i += 2) {
            if (i > 0) {
                out += ", ";
            }
            out += this._data[i] + ":" + this._data[i + 1];
        }
        return out;
    }
});

Core.ResourceBundle = Core.extend({

    $static: {
    
        /**
         * Generates a less specific version of the specified language code.
         * Returns null if no "parent" language code can be determined.
         * This operation is implemented  by removing the subtag (if found)
         * from the specified RFC 1766 language code.  If the language
         * code does not have a subtag, null is returned.
         *
         * @param {String} languageCode an RFC 1766 language code
         * @return a less specific version of the specified language code,
         *         or null if none can be determined
         * @type String 
         */
        getParentLanguageCode : function(languageCode) {
            if (languageCode.indexOf("-") == -1) {
                return null;
            } else {
                return languageCode.substring(0, languageCode.indexOf("-"));
            }
        }
    },

    /**
     * Association between RFC 1766 language codes and resource maps.
     */
    _sourceBundles: null,
    
    _generatedBundles: null,
    
    /**
     * The default resource map that should be utilized in the event that a
     * locale-specific map is not available for a particular language code.
     */
    _defaultBundle: null,

    $construct: function(defaultBundle) {
        this._sourceBundles = {};
        this._generatedBundles = {};
        this._defaultBundle = defaultBundle;
    },
    
    /**
     * Returns a locale-specific map.
     */
    get: function(languageCode) {
        var bundle = languageCode ? this._generatedBundles[languageCode] : this._defaultBundle;
        if (bundle) {
            return bundle;
        }
    
        bundle = {};
        var x;

        // Copy items from exact language bundle into new bundle.
        var sourceBundle = this._sourceBundles[languageCode];
        if (sourceBundle) {
            for (x in sourceBundle) {
                bundle[x] = sourceBundle[x];
            }
        }

        // Copy any missing items found in parent language bundle (if it exists) into new bundle.
        parentLanguageCode = Core.ResourceBundle.getParentLanguageCode(languageCode);
        if (parentLanguageCode) {
            var sourceBundle = this._sourceBundles[parentLanguageCode];
            if (sourceBundle) {
                for (x in sourceBundle) {
                    if (bundle[x] === undefined) {
                        bundle[x] = sourceBundle[x];
                    }
                }
            }
        }

        // Copy any missing items found in default bundle into new bundle.
        for (x in this._defaultBundle) {
            if (bundle[x] === undefined) {
                bundle[x] = this._defaultBundle[x];
            }
        }
        
        this._generatedBundles[languageCode] = bundle;
        return bundle;
    },

    /**
     * Adds a new locale-specific map to the 
     */
    set: function(languageCode, map) {
        this._generatedBundles = {};
        this._sourceBundles[languageCode] = map;
    },
    
    toString: function() {
        out = "ResourceBundle: ";
        for (var x in this._sourceBundles) {
            out += " " + x;
        }
        return out;
    }
});
/**
 * @fileoverview
 * Provides low-level web-client-related APIs.  Features include:
 * <ul>
 *  <li>Provides cross-platform API for accessing web client features that have
 *   inconsistent implementations on various browser platforms.</li>
 *  <li>Provides HTTP Connection object (wrapper for XMLHttpRequest).</li>
 *  <li>Provides HTML DOM manipulation capabilites.</li>
 *  <li>Provides DOM event mangement facility, enabling capturing/bubbling phases
 *   on all browsers, including Internet Explorer 6.</li>
 *  <li>Provides "virtual positioning" capability for Internet Explorer 6 to
 *   render proper top/left/right/bottom CSS positioning.</li>
 *  <li>Provides facilities to convert dimensions (e.g., in/cm/pc) to pixels.</li>
 *  <li>Provides capabilities to measure rendered size of DOM fragments.</li>
 *  <li> Provides capabilities to asynchronously load and install JavaScript modules.</li>
 * </ul>
 * Requires Core.
 */

/**
 * @class
 * Namespace for Web Core.  Non-instantiable object.
 */
WebCore = { 

    /**
     * Flag indicating that a drag-and-drop operation is in process.
     * Setting this flag will prevent text selections within the browser.
     * It must be un-set when the drag operation completes.
     * 
     * @type Boolean
     */
    dragInProgress: false,
    
    /**
     * Initializes the web core.  This method must be executed prior to using any Web Core capabilities.
     */
    init: function() { 
        if (WebCore.initialized) {
            // Already initialized.
            return;
        }
    
        WebCore.Environment._init();
        WebCore.Measure._calculateExtentSizes();
        
        WebCore.initialized = true;
    },
    
    /**
     * Internet Explorer-specific event listener to deny selection.
     * 
     * @param {Event} e the selection event
     * @private
     */
    _selectStartListener: function(e) {
        e = e ? e : window.event;
        if (WebCore.dragInProgress) {
            WebCore.DOM.preventEventDefault(e);
        }
    }
};

/**
 * @class
 * DOM manipulation utility method namespace.
 * Do not instantiate.
 */
WebCore.DOM = {

    /**
     * Temporary storage for the element about to be focused (for clients that require 'delayed' focusing).
     */
    _focusPendingElement: null,

    /**
     * Adds an event listener to an object, using the client's supported event 
     * model.  This method does NOT support method references. 
     *
     * @param {Element} eventSource the event source
     * @param {String} eventType the type of event (the 'on' prefix should NOT be included
     *        in the event type, i.e., for mouse rollover events, "mouseover" would
     *        be specified instead of "onmouseover")
     * @param {Function} eventListener the event listener to be invoked when the event occurs
     * @param {Boolean} useCapture a flag indicating whether the event listener should capture
     *        events in the final phase of propagation (only supported by 
     *        DOM Level 2 event model, not available on Internet Explorer)
     */
    addEventListener: function(eventSource, eventType, eventListener, useCapture) {
        if (eventSource.addEventListener) {
            eventSource.addEventListener(eventType, eventListener, useCapture);
        } else if (eventSource.attachEvent) {
            eventSource.attachEvent("on" + eventType, eventListener);
        }
    },
    
    /**
     * Creates a new XML DOM.
     *
     * @param {String} namespaceUri the unique URI of the namespace of the root element in 
     *        the created document (not supported for
     *        Internet Explorer 6 clients, null may be specified for all clients)
     * @param {String} qualifiedName the name of the root element of the new document (this
     *        element will be created automatically)
     * @type Document
     * @return the created DOM
     */
    createDocument: function(namespaceUri, qualifiedName) {
        if (document.implementation && document.implementation.createDocument) {
            // DOM Level 2 Browsers
            var dom = document.implementation.createDocument(namespaceUri, qualifiedName, null);
            if (!dom.documentElement) {
                dom.appendChild(dom.createElement(qualifiedName));
            }
            return dom;
        } else if (window.ActiveXObject) {
            // Internet Explorer
            var createdDocument = new ActiveXObject("Microsoft.XMLDOM");
            var documentElement = createdDocument.createElement(qualifiedName);
            createdDocument.appendChild(documentElement);
            return createdDocument;
        } else {
            throw new Error("XML DOM creation not supported by browser environment.");
        }
    },
    
    /**
     * Focuses the given DOM element.
     * The focus operation may be placed in the scheduler if the browser requires the focus
     * operation to be performed outside of current JavaScript context (i.e., in the case
     * where the element to be focused was just rendered in this context).
     * 
     * @param {Element} the DOM element to focus
     */
    focusElement: function(element) {
        if (WebCore.Environment.QUIRK_DELAYED_FOCUS_REQUIRED) {
            WebCore.DOM._focusPendingElement = element;
            WebCore.Scheduler.run(Core.method(window, this._focusElementImpl));
        } else {
            this._focusElementImpl(element);
        }
    },
    
    /**
     * Focus element implementation.
     * 
     * @param {Element} the DOM element to focus
     * @private
     */
    _focusElementImpl: function(element) {
        if (!element) {
            element = WebCore.DOM._focusPendingElement;
            WebCore.DOM._focusPendingElement = null;
        }
        if (element && element.focus) {
            try {
                element.focus();
            } catch (ex) {
                // Silently digest IE focus exceptions.
            }
        }
    },
    
    /**
     * Returns the first immediate child element of parentElement with the specified tag name.
     * 
     * @param {Element} parentElement the parent element
     * @param tagName the tag name
     * @return the first child element of parentElement with the specified tag name,
     *         or null if no elements match
     * @type Element
     */
    getChildElementByTagName: function(parentElement, tagName) {
        var element = parentElement.firstChild;
        while (element) {
            if (element.nodeType == 1 && element.nodeName == tagName) {
                return element;
            }
            element = element.nextSibling;
        }
        return null;
    },
    
    /**
     * Returns an array containing all immediate child element of parentElement with the specified tag name.
     * 
     * @param {Element} parentElement the parent element
     * @param tagName the tag name
     * @return the child elements
     * @type Array
     */
    getChildElementsByTagName: function(parentElement, tagName) {
        var elements = [];
        var element = parentElement.firstChild;
        while (element) {
            if (element.nodeType == 1 && element.nodeName == tagName) {
                elements.push(element);
            }
            element = element.nextSibling;
        }
        return elements;
    },
    
    /**
     * Returns the target of an event, using the client's supported event model.
     * On clients which support the W3C DOM Level 2 event specification,
     * the <code>target</code> property of the event is returned.
     * On clients which support only the Internet Explorer event model,
     * the <code>srcElement</code> property of the event is returned.
     *
     * @param {Event} e the event
     * @return the target
     * @type Element
     */
    getEventTarget: function(e) {
        return e.target ? e.target : e.srcElement;
    },
    
    /**
     * Returns the related target of an event, using the client's supported event model.
     * On clients which support the W3C DOM Level 2 event specification,
     * the <code>relatedTarget</code> property of the event is returned.
     * On clients which support only the Internet Explorer event model,
     * the <code>toElement</code> property of the event is returned.
     *
     * @param {Event} e the event
     * @return the target
     * @type Element
     */
    getEventRelatedTarget: function(e) {
        return e.relatedTarget ? e.relatedTarget : e.toElement;
    },
    
    /**
     * Determines if <code>ancestorNode</code> is or is an ancestor of
     * <code>descendantNode</code>.
     *
     * @param {Node} ancestorNode the potential ancestor node
     * @param {Npde} descendantNode the potential descendant node
     * @return true if <code>ancestorNode</code> is or is an ancestor of
     *         <code>descendantNode</code>
     * @type Boolean
     */
    isAncestorOf: function(ancestorNode, descendantNode) {
        var testNode = descendantNode;
        while (testNode !== null) {
            if (testNode == ancestorNode) {
                return true;
            }
            testNode = testNode.parentNode;
        }
        return false;
    },

    /**
     * Prevents the default action of an event from occurring, using the
     * client's supported event model.
     * On clients which support the W3C DOM Level 2 event specification,
     * the preventDefault() method of the event is invoked.
     * On clients which support only the Internet Explorer event model,
     * the 'returnValue' property of the event is set to false.
     *
     * @param {Event} e the event
     */
    preventEventDefault: function(e) {
        if (e.preventDefault) {
            e.preventDefault();
        } else {
            e.returnValue = false;
        }
    },
    
    /**
     * Removes all child nodes from the specified DOM node.
     *
     * @param {Node} node the parent node whose children should be deleted
     */
    removeAllChildren: function(node) {
        while (node.firstChild) {
            node.removeChild(node.firstChild);
        }
    },
    
    /**
     * Removes an event listener from an object, using the client's supported event 
     * model.  This method does NOT support method references.
     *
     * @param {Element} eventSource the event source
     * @param {String} eventType the type of event (the 'on' prefix should NOT be included
     *        in the event type, i.e., for mouse rollover events, "mouseover" would
     *        be specified instead of "onmouseover")
     * @param {Function} eventListener the event listener to be invoked when the event occurs
     * @param {Boolean}useCapture a flag indicating whether the event listener should capture
     *        events in the final phase of propagation (only supported by 
     *        DOM Level 2 event model, not available on Internet Explorer)
     */
    removeEventListener: function(eventSource, eventType, eventListener, useCapture) {
        if (eventSource.removeEventListener) {
            eventSource.removeEventListener(eventType, eventListener, useCapture);
        } else if (eventSource.detachEvent) {
            eventSource.detachEvent("on" + eventType, eventListener);
        }
    },
    
    /**
     * Removes the specified DOM node from the DOM tree. This method employs a workaround for the
     * <code>QUIRK_PERFORMANCE_LARGE_DOM_REMOVE</code> quirk.
     *
     * @param {Node} node the node which should be deleted
     */
    removeNode: function(node) {
        var parentNode = node.parentNode;
        if (!parentNode) {
            // not in DOM tree
            return;
        }
        if (WebCore.Environment.QUIRK_PERFORMANCE_LARGE_DOM_REMOVE) {
            this._removeNodeRecursive(node);
        } else {
            parentNode.removeChild(node);
        }
    },
    
    /**
     * Removes the specified DOM node from the DOM tree in a recursive manner, i.e. all descendants
     * of the given node are removed individually. This alleviates slow performance when removing large
     * DOM trees.
     *
     * @param {Node} node the node which should be deleted
    */
    _removeNodeRecursive: function(node) {
        var childNode = node.firstChild;
        while (childNode) {
            var nextChildNode = childNode.nextSibling;
            this._removeNodeRecursive(childNode);
            childNode = nextChildNode;
        }
        node.parentNode.removeChild(node);
    },
    
    /**
     * Stops an event from propagating ("bubbling") to parent nodes in the DOM, 
     * using the client's supported event model.
     * On clients which support the W3C DOM Level 2 event specification,
     * the stopPropagation() method of the event is invoked.
     * On clients which support only the Internet Explorer event model,
     * the 'cancelBubble' property of the event is set to true.
     *
     * @param {Event} e the event
     */
    stopEventPropagation: function(e) {
        if (e.stopPropagation) {
            e.stopPropagation();
        } else {
            e.cancelBubble = true;
        }
    }
};

/**
 * @class
 * Provides information about the web browser environment.
 * Non-instantiable class.
 */
WebCore.Environment = {

    /**
     * Performs initial analysis of environment.
     * Automatically invoked when WebCore module is initialized.
     */
    _init: function() {
        var ua = navigator.userAgent.toLowerCase();
        this.BROWSER_OPERA = ua.indexOf("opera") != -1;
        this.BROWSER_SAFARI = ua.indexOf("safari") != -1;
        this.BROWSER_KONQUEROR = ua.indexOf("konqueror") != -1;
        this.BROWSER_FIREFOX = ua.indexOf("firefox") != -1;
        
        this.CSS_FLOAT = "cssFloat";
    
        // Note deceptive user agent fields:
        // - Konqueror and Safari UA fields contain "like Gecko"
        // - Opera UA field typically contains "MSIE"
        this.DECEPTIVE_USER_AGENT = this.BROWSER_OPERA || this.BROWSER_SAFARI || this.BROWSER_KONQUEROR;
        
        this.BROWSER_MOZILLA = !this.DECEPTIVE_USER_AGENT && ua.indexOf("gecko") != -1;
        this.BROWSER_INTERNET_EXPLORER = !this.DECEPTIVE_USER_AGENT && ua.indexOf("msie") != -1;
        
        // Retrieve Version Info (as necessary).
        if (this.BROWSER_INTERNET_EXPLORER) {
            this._parseVersionInfo(ua, "msie ");
        } else if (this.BROWSER_FIREFOX) {
            this._parseVersionInfo(ua, "firefox/");
        } else if (this.BROWSER_OPERA) {
            this._parseVersionInfo(ua, "opera/");
        } else if (this.BROWSER_SAFARI) {
            this._parseVersionInfo(ua, "version/");
        } else if (this.BROWSER_MOZILLA) {
            this._parseVersionInfo(ua, "rv:")
        } else if (this.BROWSER_KONQUEROR) {
            this._parseVersionInfo(ua, "konqueror/");
        }
    
        //FIXME Quirk flags not refined yet, some quirk flags from Echo 2.0/1 will/may be deprecated/removed.
        
        // Set IE Quirk Flags
        if (this.BROWSER_MOZILLA) {
            if (this.BROWSER_FIREFOX) {
                if (this.BROWSER_MAJOR_VERSION < 2) {
                    this.QUIRK_DELAYED_FOCUS_REQUIRED = true;
                }
            } else {
                this.QUIRK_PERFORMANCE_LARGE_DOM_REMOVE = true;
                this.QUIRK_DELAYED_FOCUS_REQUIRED = true;
            }
        } else if (this.BROWSER_OPERA) {
            this.NOT_SUPPORTED_RELATIVE_COLUMN_WIDTHS = true;
        }
    },
    
    /**
     * Parses version information from user agent string. The text argument specifies
     * the string that prefixes the version info in the ua string (ie 'version/' for Safari for example).
     * <p>
     * The major version is retrieved by getting the int between text and the first dot. The minor version
     * is retrieved by getting the int between the first dot and the first non-numeric character that appears
     * after the dot, or the end of the ua string (whichever comes first).
     * If the ua string does not supply a minor version, the minor version is assumed to be 0.
     *
     * @private
     * @param ua the lower cased user agent string
     * @param searchString the text that prefixes the version info (version info must be the first appearance of 
     *          this text in the ua string)
     */
    _parseVersionInfo: function(ua, searchString) {
        var ix1 = ua.indexOf(searchString);
        var ix2 = ua.indexOf(".", ix1);
        var ix3 = ua.length;
        
        if (ix2 == -1) {
            ix2 = ua.length;
        } else {
            // search for the first non-number character after the dot
            for (var i = ix2 + 1; i < ua.length; i++) {
                var c = ua.charAt(i);
                if (isNaN(c)) {
                    ix3 = i;
                    break;
                }
            }
        }
        
        this.BROWSER_MAJOR_VERSION = parseInt(ua.substring(ix1 + searchString.length, ix2));
        if (ix2 == ua.length) {
            this.BROWSER_MINOR_VERSION = 0;
        } else {
            this.BROWSER_MINOR_VERSION = parseInt(ua.substring(ix2 + 1, ix3));
        }
    }
};

/**
 * @class
 * EventProcessor namespace.  Non-instantiable object.
 * The static methods in this object provide a standard framework for handling
 * DOM events across incompatible browser platforms.
 * <p>
 * <b>Capturing/Bubbling Listeners:</b>
 * This implementation additionally allows for the registration of capturing and bubbling event 
 * listeners that work even on Internet Explorer platforms, where they are not natively supported.
 * This implementation relies on the fact that all event listeners will be registered
 * through it.
 */
WebCore.EventProcessor = {
    
    /**
     * Provides utilities for restricting selection of DOM elements.  These are necessary as double click and drag
     * events will cause/begin undesired selections.
     */
    Selection: {

        /**
         * Adds a listener to an element that will prevent text selection / highlighting as a result of mouse clicks.
         * The disable() method should be invoked when the element is to be disposed.
         * The event is registered using the event processor, so invoking WebCore.EventProcessor.removeAll() on its
         * element will also dispose it.
         *
         * @param {Element} element the element on which to forbid text selection
         * @see WebCore.EventProcessor.Selection#enable
         */
        disable: function(element) {
            WebCore.EventProcessor.add(element, "mousedown", WebCore.EventProcessor.Selection._disposeEvent, false);
            if (WebCore.Environment.PROPRIETARY_EVENT_SELECT_START_SUPPORTED) {
                WebCore.EventProcessor.add(element, "selectstart", WebCore.EventProcessor.Selection._disposeEvent, false);
            }
        },
        
        /**
         * Selection denial listener implementation.
         * 
         * @param e the selection/click event
         * @private
         */
        _disposeEvent: function(e) {
            WebCore.DOM.preventEventDefault(e);
        },
    
        /**
         * Removes a selection denial listener.
         * 
         * @param element the element from which to remove the selection denial listener
         * @see WebCore.EventProcessor.Selection#enable
         */
        enable: function(element) {
            WebCore.EventProcessor.remove(element, "mousedown", WebCore.EventProcessor.Selection._disposeEvent, false);
            if (WebCore.Environment.PROPRIETARY_EVENT_SELECT_START_SUPPORTED) {
                WebCore.EventProcessor.remove(element, "selectstart", WebCore.EventProcessor.Selection._disposeEvent, false);
            }
        }
    },

    /**
     * Next available sequentially assigned element identifier.
     * Elements are assigned unique identifiers to enable mapping between 
     * elements and lists of registered event listeners.
     *
     * @type Integer
     */
    _nextId: 0,
    
    /**
     * Mapping between element ids and ListenerLists containing listeners to invoke during capturing phase.
     * @type Core.Arrays.LargeMap
     */
    _capturingListenerMap: new Core.Arrays.LargeMap(),
    
    /**
     * Mapping between element ids and ListenerLists containing listeners to invoke during bubbling phase.
     * @type Core.Arrays.LargeMap
     */
    _bubblingListenerMap: new Core.Arrays.LargeMap(),
    
    /**
     * Registers an event handler.
     *
     * @param {Element} element the DOM element on which to add the event handler
     * @param {String} eventType the DOM event type
     * @param {Function} eventTarget the event handler to invoke when the event is fired
     * @param {Boolean} capture true to fire the event during the capturing phase, false to fire the event during
     *        the bubbling phase
     */
    add: function(element, eventType, eventTarget, capture) {
        // Assign event processor element id to element if not present.
        if (!element.__eventProcessorId) {
            element.__eventProcessorId = ++WebCore.EventProcessor._nextId;
        }
    
        var listenerList;
        
        // Determine the Core.ListenerList to which the listener should be added.
        if (element.__eventProcessorId == WebCore.EventProcessor._lastId 
                && capture == WebCore.EventProcessor._lastCapture) {
            // If the 'element' and 'capture' properties are identical to those specified on the prior invocation
            // of this method, the correct listener list is stored in the '_lastListenerList' property. 
            listenerList = WebCore.EventProcessor._lastListenerList; 
        } else {
            // Obtain correct id->ListenerList mapping based on capture parameter.
            var listenerMap = capture ? WebCore.EventProcessor._capturingListenerMap 
                                      : WebCore.EventProcessor._bubblingListenerMap;
            
            // Obtain ListenerList based on element id.                              
            listenerList = listenerMap.map[element.__eventProcessorId];
            if (!listenerList) {
                // Create new ListenerList if none exists.
                listenerList = new Core.ListenerList();
                listenerMap.map[element.__eventProcessorId] = listenerList;
            }
            
            // Cache element's event processor id, capture parameter value, and listener list.
            // If the same element/capture properties are provided in the next call (which commonly occurs),
            // the lookup operation will be unnecessary.
            WebCore.EventProcessor._lastId = element.__eventProcessorId;
            WebCore.EventProcessor._lastCapture = capture;
            WebCore.EventProcessor._lastListenerList = listenerList;
        }
    
        // Register event listener on DOM element.
        if (!listenerList.hasListeners(eventType)) {
            WebCore.DOM.addEventListener(element, eventType, WebCore.EventProcessor._processEvent, false);
        }

        // Add event handler to the ListenerList.
        listenerList.addListener(eventType, eventTarget);
    },
    
    /**
     * Listener method which is invoked when ANY event registered with the event processor occurs.
     * 
     * @param {Event} e 
     */
    _processEvent: function(e) {
        e = e ? e : window.event;
        if (!e.target && e.srcElement) {
            // The Internet Explorer event model stores the target element in the 'srcElement' property of an event.
            // Modify the event such the target is retrievable using the W3C DOM Level 2 specified property 'target'.
            e.target = e.srcElement;
        }
        
        // Establish array containing elements ancestry, with index 0 containing 
        // the element and the last index containing its most distant ancestor.  
        // Only record elements that have ids.
        var elementAncestry = [];
        var targetElement = e.target;
        while (targetElement) {
            if (targetElement.__eventProcessorId) { // Element Node with identifier.
                elementAncestry.push(targetElement);
            }
            targetElement = targetElement.parentNode;
        }
    
        var listenerList;
        
        var propagate = true;
        
        // Fire event to capturing listeners.
        for (var i = elementAncestry.length - 1; i >= 0; --i) {
            listenerList = WebCore.EventProcessor._capturingListenerMap.map[elementAncestry[i].__eventProcessorId];
            if (listenerList) {
                // Set registered target on event.
                e.registeredTarget = elementAncestry[i];
                if (!listenerList.fireEvent(e)) {
                    propagate = false;
                }
            }
            if (!propagate) {
                // Stop propagation if requested.
                break;
            }
        }
        
        if (propagate) {
            // Fire event to bubbling listeners.
            for (var i = 0; i < elementAncestry.length; ++i) {
                listenerList = WebCore.EventProcessor._bubblingListenerMap.map[elementAncestry[i].__eventProcessorId];
                // Set registered target on event.
                e.registeredTarget = elementAncestry[i];
                if (listenerList) {
                    if (!listenerList.fireEvent(e)) {
                        propagate = false;
                    }
                }
                if (!propagate) {
                    // Stop propagation if requested.
                    break;
                }
            }
        }

        // Inform DOM to stop propagation of event, in all cases.
        // Event will otherwise be re-processed by higher-level elements registered with the event processor.
        WebCore.DOM.stopEventPropagation(e);
    },
    
    /**
     * Unregisters an event handler.
     *
     * @param {Element} element the DOM element on which to add the event handler
     * @param {String} eventType the DOM event type
     * @param {Function} eventTarget the function to invoke when the event is fired
     * @param {Boolean} capture true to fire the event during the capturing phase, false to fire the event during
     *        the bubbling phase
     */
    remove: function(element, eventType, eventTarget, capture) {
        WebCore.EventProcessor._lastId = null;
        
        if (!element.__eventProcessorId) {
            return;
        }
    
        // Obtain correct id->ListenerList mapping based on capture parameter.
        var listenerMap = capture ? WebCore.EventProcessor._capturingListenerMap 
                                  : WebCore.EventProcessor._bubblingListenerMap;
    
        // Obtain ListenerList based on element id.                              
        var listenerList = listenerMap.map[element.__eventProcessorId];
        if (listenerList) {
            // Remove event handler from the ListenerList.
            listenerList.removeListener(eventType, eventTarget);
            
            if (listenerList.isEmpty()) {
                listenerMap.remove(element.__eventProcessorId);
            }

            // Unregister event listener on DOM element if all listeners have been removed.
            if (!listenerList.hasListeners(eventType)) {
                WebCore.DOM.removeEventListener(element, eventType, WebCore.EventProcessor._processEvent, false);
            }
        }
    },
    
    /**
     * Unregister all event handlers from a specific element.
     * Use of this operation is recommended when disposing of components, it is
     * more efficient than removing listenerse individually and guarantees proper clean-up.
     * 
     * @param {Element} the element
     */
    removeAll: function(element) {
        if (!element.__eventProcessorId) {
            return;
        }
        WebCore.EventProcessor._removeAllImpl(element, WebCore.EventProcessor._capturingListenerMap);
        WebCore.EventProcessor._removeAllImpl(element, WebCore.EventProcessor._bubblingListenerMap);
    },
    
    /**
     * Implementation method for removeAll().
     * Removes all capturing or bubbling listeners from a specific element
     * 
     * @param {Element} the element
     * @param {Core.Arrays.LargeMap} the map from which the listeners should be removed, either
     *        WebCore.EventProcessor._capturingListenerMap or WebCore.EventProcessor._bubblingListenerMap
     * @private
     */
    _removeAllImpl: function(element, listenerMap) {
        var listenerList = listenerMap.map[element.__eventProcessorId];
        if (!listenerList) {
            return;
        }
    
        var types = listenerList.getListenerTypes();
        for (var i = 0; i < types.length; ++i) {
            WebCore.DOM.removeEventListener(element, types[i], WebCore.EventProcessor._processEvent, false); 
        }
        
        listenerMap.remove(element.__eventProcessorId);
    },
    
    /**
     * toString() implementation for debugging purposes.
     * Displays contents of capturing and bubbling listener maps.
     * 
     * @return string represenation of listener maps
     * @type String
     */
    toString: function() {
        return "Capturing: " + WebCore.EventProcessor._capturingListenerMap + "\n"
                + "Bubbling: " + WebCore.EventProcessor._bubblingListenerMap;
    }
};

/**
 * @class
 * An HTTP connection to the hosting server.  This method provides a cross
 * platform wrapper for XMLHttpRequest and additionally allows method
 * reference-based listener registration.  
 */
WebCore.HttpConnection = Core.extend({

    _url: null,
    
    _contentType: null,
    
    _method: null,
    
    _messageObject: null,
    
    _listenerList: null,
    
    _disposed: false,
    
    _xmlHttpRequest: null,

    /**
     * Creates a new <code>HttpConnection</code>.
     * This method simply configures the connection, the connection
     * will not be opened until <code>connect()</code> is invoked.
     *
     * @param {String} url the target URL
     * @param {String} method the connection method, i.e., GET or POST
     * @param messageObject the message to send (may be a String or XML DOM)
     * @param {String} contentType the request content-type
     * @constructor
     */
    $construct: function(url, method, messageObject, contentType) {
        this._url = url;
        this._contentType = contentType;
        this._method = method;
        this._messageObject = messageObject;
        this._listenerList = new Core.ListenerList();
    },
    
    /**
     * Adds a response listener to be notified when a response is received from the connection.
     * 
     * @param {Function} l the listener to add
     */
    addResponseListener: function(l) {
        this._listenerList.addListener("response", l);
    },
    
    /**
     * Executes the HTTP connection.
     * This method will return before the HTTP connection has received a response.
     */
    connect: function() {
        var usingActiveXObject = false;
        if (window.XMLHttpRequest) {
            this._xmlHttpRequest = new XMLHttpRequest();
        } else if (window.ActiveXObject) {
            usingActiveXObject = true;
            this._xmlHttpRequest = new ActiveXObject("Microsoft.XMLHTTP");
        } else {
            throw "Connect failed: Cannot create XMLHttpRequest.";
        }
    
        var instance = this;
        
        // Create closure around instance.
        this._xmlHttpRequest.onreadystatechange = function() { 
            if (!instance) {
                return;
            }
            try {
                instance._processReadyStateChange();
            } finally {
                if (instance._disposed) {
                    // Release instance reference to allow garbage collection.
                    instance = null;
                }
            }
        };
        
        this._xmlHttpRequest.open(this._method, this._url, true);
    
        if (this._contentType && (usingActiveXObject || this._xmlHttpRequest.setRequestHeader)) {
            this._xmlHttpRequest.setRequestHeader("Content-Type", this._contentType);
        }
        this._xmlHttpRequest.send(this._messageObject ? this._messageObject : null);
    },
    
    /**
     * Disposes of the connection.  This method must be invoked when the connection 
     * will no longer be used/processed.
     */
    dispose: function() {
        this._listenerList = null;
        this._messageObject = null;
        this._xmlHttpRequest = null;
        this._disposed = true;
    },
    
    /**
     * Returns the response status code of the HTTP connection, if available.
     * 
     * @return {Integer} the response status code
     */
    getStatus: function() {
        return this._xmlHttpRequest ? this._xmlHttpRequest.status : null;
    },
    
    /**
     * Returns the response as text.
     * This method may only be invoked from a response handler.
     *
     * @return the response, as text
     */
    getResponseText: function() {
        return this._xmlHttpRequest ? this._xmlHttpRequest.responseText : null;
    },
    
    /**
     * Returns the response as an XML DOM.
     * This method may only be invoked from a response handler.
     *
     * @return the response, as an XML DOM
     */
    getResponseXml: function() {
        return this._xmlHttpRequest ? this._xmlHttpRequest.responseXML : null;
    },
    
    /**
     * Event listener for <code>readystatechange</code> events received from
     * the <code>XMLHttpRequest</code>.
     */
    _processReadyStateChange: function() {
        if (this._disposed) {
            return;
        }
        
        if (this._xmlHttpRequest.readyState == 4) {
            var responseEvent;
            try {
                // 0 included as a valid response code for non-served applications.
                var valid = this._xmlHttpRequest.status == 0 ||  
                        (this._xmlHttpRequest.status >= 200 && this._xmlHttpRequest.status <= 299);
                responseEvent = {type: "response", source: this, valid: valid};
            } catch (ex) {
                responseEvent = {type: "response", source: this, valid: false, exception: ex};
            }
            
            this._listenerList.fireEvent(responseEvent);
            this.dispose();
        }
    },
    
    /**
     * Removes a response listener to be notified when a response is received from the connection.
     * 
     * @param {Function} l the listener to remove
     */
    removeResponseListener: function(l) {
        this._listenerList.removeListener("response", l);
    }
});

/**
 * @class
 * Utilities for dynamically loading additional script libraries.
 * Non-instantiable class. 
 */
WebCore.Library = {

    /**
     * Set of loaded libraries (keys are library urls, value is true when library has been loaded).
     * @private
     */
    _loadedLibraries: { },
    
    /**
     * A representation of a group of libraries to be loaded at the same time.
     * Libraries will be retrieved asynchronously, and then installed once ALL the libraries have
     * been retrieved.  Installation will be done in the order in which the add() method was
     * invoked to add libraries to the group (without regard for the order in which the 
     * HTTP server returns the library code).
     */
    Group: Core.extend({
    
        _listenerList: null,
        
        _libraries: null,
        
        _loadedCount: 0,
        
        _totalCount: 0,
    
        /**
         * Creates a new library group.
         * @constructor 
         */
        $construct: function() {
            this._listenerList = new Core.ListenerList();
            this._libraries = [];
        },
        
        /**
         * Adds a library to the library group.
         * Libraries which have previously been loaded will not be loaded again.
         *
         * @param libraryUrl the URL from which to retrieve the library.
         */
        add: function(libraryUrl) {
            if (WebCore.Library._loadedLibraries[libraryUrl]) {
                // Library already loaded: ignore.
                return;
            }
            
            var libraryItem = new WebCore.Library._Item(this, libraryUrl);
            this._libraries.push(libraryItem);
        },
        
        /**
         * Adds a listener to be notified when all libraries in the group have been loaded.
         *
         * @param {Function} l the listener to add
         */
        addLoadListener: function(l) {
            this._listenerList.addListener("load", l);
        },
        
        /**
         * Notifies listeners of completed library loading.
         * 
         * @private
         */
        _fireLoadEvent: function() {
            this._listenerList.fireEvent({type: "load", source: this});
        },
        
        /**
         * Determines if this library group contains any new (not previously loaded)
         * libraries.
         * 
         * @return true if any new libraries exist
         * @type Boolean
         */
        hasNewLibraries: function() {
            return this._libraries.length > 0;
        },
        
        /**
         * Installs all libraries in the group.
         * This method is invoked once all libraries have been successfully
         * retrieved.  It will invoke any registered load listeners
         * once the libraries have been installed.
         * 
         * @private
         */
        _install: function() {
            for (var i = 0; i < this._libraries.length; ++i) {
                try {
                    this._libraries[i]._install();
                } catch (ex) {
                    throw new Error("Exception installing library \"" + this._libraries[i]._url + "\"; " + ex);
                }
            }
            this._fireLoadEvent();
        },
        
        /**
         * Event listener invoked when a single library has been successfully retrieved.
         * When all libraries have been retrieved, this method will invoke _install().
         * @private
         */
        _notifyRetrieved: function() {
            ++this._loadedCount;
            if (this._loadedCount == this._totalCount) {
                this._install();
            }
        },
        
        /**
         * Initializes library loading.  When this method is invoked
         * the libraries will be asynchronously loaded.  This method
         * will return before the libraries have been loaded.
         * Once this method has been invoked, add() may no longer
         * be invoked.
         */
        load: function() {
            this._totalCount = this._libraries.length;
            for (var i = 0; i < this._libraries.length; ++i) {
                this._libraries[i]._retrieve();
            }
        },
        
        /**
         * Removes a listener from being notified when all libraries in the group have been loaded.
         *
         * @param {Function} l the listener to remove
         */
        removeLoadListener: function(l) {
            this._listenerList.removeListener("load", l);
        }
    }),

    /**
     * @class
     * Representation of a single library to be loaded within a group
     * 
     * @private 
     */    
    _Item: Core.extend({
    
        _url: null,
        
        _group: null,
        
        _content: null,
    
        /**
         * Creates a new library item.
         * 
         * @param {WebCore.Library.Group} group the library group in which the item is contained
         * @param {String} url the URL from which the library may be retrieved
         * @constructor
         */
        $construct: function(group, url) {
            this._url = url;
            this._group = group;
        },
        
        /**
         * Event listener for response from the HttpConnection used to retrive the library.
         * 
         * @param e the event
         * @private
         */
        _retrieveListener: function(e) {
            if (!e.valid) {
                throw new Error("Invalid HTTP response from library request: " + e.source.getStatus());
            }
            this._content = e.source.getResponseText();
            this._group._notifyRetrieved();
        },
        
        /**
         * Installs the library.
         * The library must have been loaded before invoking this method.
         * @private
         */
        _install: function() {
            WebCore.Library._loadedLibraries[this._url] = true;
            if (this._content == null) {
                throw new Error("Attempt to install library when no content has been loaded.");
            }
            
            // Execute content to install library.
            eval(this._content);
        },
        
        /**
         * Asynchronously retrieves the library.
         * This method will invoke the retrieve listener when the library has been completed,
         * it will return before the library has been retrieved.
         */
        _retrieve: function() {
            var conn = new WebCore.HttpConnection(this._url, "GET");
            conn.addResponseListener(Core.method(this, this._retrieveListener));
            conn.connect();
        }
    })
};

/**
 * @class
 * Namespace for measuring-related operations.
 * Do not instantiate.  
 */
WebCore.Measure = { 

    /** Size of one inch in horizontal pixels. */
    _hInch: 96,
    
    /** Size of one inch in vertical pixels. */
    _vInch: 96,
    
    /** Size of one 'ex' in horizontal pixels. */
    _hEx: 7,
    
    /** Size of one 'ex' in vertical pixels. */
    _vEx: 7,
    
    /** Size of one 'em' in horizontal pixels. */
    _hEm: 13.3333,
    
    /** Size of one 'em' in vertical pixels. */
    _vEm: 13.3333,

    /**
     * Converts any non-relative extent value to pixels.
     * 
     * @param {Number} value the value to convert
     * @param {String} the units, one of the following values: in, cm, mm, pt, pc, em, ex
     * @return the pixel value (may have a fractional part)
     * @type Number
     */
    extentToPixels: function(value, units, horizontal) {
        if (!units || units == "px") {
            return value;
        }
        var dpi = horizontal ? WebCore.Measure._hInch : WebCore.Measure._vInch;
        switch (units) {
        case "%":  return 0;
        case "in": return value * (horizontal ? WebCore.Measure._hInch : WebCore.Measure._vInch);
        case "cm": return value * (horizontal ? WebCore.Measure._hInch : WebCore.Measure._vInch) / 2.54;
        case "mm": return value * (horizontal ? WebCore.Measure._hInch : WebCore.Measure._vInch) / 25.4;
        case "pt": return value * (horizontal ? WebCore.Measure._hInch : WebCore.Measure._vInch) / 72;
        case "pc": return value * (horizontal ? WebCore.Measure._hInch : WebCore.Measure._vInch) / 6;
        case "em": return value * (horizontal ? WebCore.Measure._hEm   : WebCore.Measure._vEm);
        case "ex": return value * (horizontal ? WebCore.Measure._hEx   : WebCore.Measure._vEx);
        }
    },

    /**
     * Updates internal measures used in converting length units 
     * (e.g., in, mm, ex, and em) to pixels.
     * Automatically invoked when WebCore module is initialized.
     * @private
     */
    _calculateExtentSizes: function() {
        var containerElement = document.getElementsByTagName("body")[0];
    
        var inchDiv4 = document.createElement("div");
        inchDiv4.style.width = "4in";
        inchDiv4.style.height = "4in";
        containerElement.appendChild(inchDiv4);
        WebCore.Measure._hInch = inchDiv4.offsetWidth / 4;
        WebCore.Measure._vInch = inchDiv4.offsetHeight / 4;
        containerElement.removeChild(inchDiv4);
        
        var emDiv24 = document.createElement("div");
        emDiv24.style.width = "24em";
        emDiv24.style.height = "24em";
        containerElement.appendChild(emDiv24);
        WebCore.Measure._hEm = emDiv24.offsetWidth / 24;
        WebCore.Measure._vEm = emDiv24.offsetHeight / 24;
        containerElement.removeChild(emDiv24);
        
        var exDiv24 = document.createElement("div");
        exDiv24.style.width = "24ex";
        exDiv24.style.height = "24ex";
        containerElement.appendChild(exDiv24);
        WebCore.Measure._hEx = exDiv24.offsetWidth / 24;
        WebCore.Measure._vEx = exDiv24.offsetHeight / 24;
        containerElement.removeChild(exDiv24);
    },
    
    /**
     * Measures the scrollbar offset of an element, including any
     * scroll-bar related offsets of its ancestors.
     * 
     * @param element the elemnt to measure
     * @return the offset data, with 'left' and 'top' properties specifying the offset amounts
     * @type Object
     * @private
     */
    _getScrollOffset: function(element) {
        var valueT = 0, valueL = 0;
        do {
            valueT += element.scrollTop  || 0;
            valueL += element.scrollLeft || 0; 
            element = element.parentNode;
        } while (element);
        return { left: valueL, top: valueT };
    },
    
    /**
     * Measures the cumulative offset of an element.
     * 
     * @param element the element to measure
     * @return the offset data, with 'left' and 'top' properties specifying the offset amounts
     * @type Object
     * @private
     */
    _getCumulativeOffset: function(element) {
        var valueT = 0, valueL = 0;
        do {
            valueT += element.offsetTop  || 0;
            valueL += element.offsetLeft || 0;
            element = element.offsetParent;
        } while (element);
        return { left: valueL, top: valueT };
    },

    /**
     * @class
     * Measures the boundaries of an element,i.e., its left and top position and/or
     * width and height.  If the element is not attached to the rendered DOM hierarchy,
     * the element will be temporarily removed from its hierarchy and placed in an
     * off-screen buffer for measuring.
     */
    Bounds: Core.extend({

        /**
         * The width of the element, in pixels.
         * @type Integer
         */
        width: null,
        
        /**
         * The height of the element, in pixels.
         * @type Integer
         */
        height: null,
        
        /**
         * The top coordinate of the element, in pixels relative to the upper-left corner of the interior of the window.
         * @type Integer
         */
        top: null,
         
        /**
         * The left coordinate of the element, in pixels relative to the upper-left corner of the interior of the window.
         * @type Integer
         */
        left: null,

        /**
         * Creates a new Bounds object to calculate the size and/or position of an element.
         * 
         * @param element the element to measure.
         * @constructor
         */    
        $construct: function(element) {
            var testElement = element;
            while (testElement && testElement != document) {
                testElement = testElement.parentNode;
            }
            var rendered = testElement == document;
            
            // Create off-screen div element for evaluating sizes if necessary.
            if (!WebCore.Measure.Bounds._offscreenDiv) {
                WebCore.Measure.Bounds._offscreenDiv = document.createElement("div");
                WebCore.Measure.Bounds._offscreenDiv.style.cssText 
                        = "position: absolute; top: -1700px; left: -1300px; width: 1600px; height: 1200px;";
                document.body.appendChild(WebCore.Measure.Bounds._offscreenDiv);
            }
        
            var parentNode, nextSibling;
            if (!rendered) {
                // Element must be added to off-screen element for measuring.
                
                // Store parent node and next sibling such that element may be replaced into proper position
                // once off-screen measurement has been completed.
                parentNode = element.parentNode;
                nextSibling = element.nextSibling;
        
                // Remove element from parent.
                if (parentNode) {
                    parentNode.removeChild(element);
                }
                
                // Append element to measuring container DIV.
                WebCore.Measure.Bounds._offscreenDiv.appendChild(element);
            }
            
            // Store  width and height of element.
        
            this.width = element.offsetWidth;
            this.height = element.offsetHeight;
            
            if (!rendered) {
                // Replace off-screen measured element in previous location.
                WebCore.Measure.Bounds._offscreenDiv.removeChild(element);
                if (parentNode) {
                    parentNode.insertBefore(element, nextSibling);
                }
            }
            
            // Determine top and left positions of element if rendered on-screen.
            if (rendered) {
                var cumulativeOffset = WebCore.Measure._getCumulativeOffset(element);
                var scrollOffset = WebCore.Measure._getScrollOffset(element);
        
                this.top = cumulativeOffset.top - scrollOffset.top;
                this.left = cumulativeOffset.left - scrollOffset.left;
            }
        },
        
        /**
         * toString() implementation for debug purposes.
         * 
         * @return a string representation of the object
         * @type String
         */
        toString: function() {
            return (this.left != null ? (this.left + "," + this.top + " : ") : "") + "[" + this.width + "x" + this.height + "]";
        }
    })
};

/**
 * Scheduler namespace.  Non-instantiable object.
 * Provides capability to invoke code at regular intervals, after a delay, 
 * or after the current JavaScript execution context has completed.
 * Provides an object-oriented means of accomplishing this task.
 */
WebCore.Scheduler = {

    /**
     * Interval at which the scheduler should wake to check for queued tasks.
     * @type Number
     */
    INTERVAL: 20,
    
    /**
     * Collection of runnables to execute.
     * @private
     */
    _runnables: [],
    
    /**
     * Executes the scheduler, running any runnables that are due.
     * This method is invoked by the interval/thread.
     */
    _execute: function() {
        var time = new Date().getTime();
        
        for (var i = 0; i < WebCore.Scheduler._runnables.length; ++i) {
            var runnable = WebCore.Scheduler._runnables[i];
            if (!runnable._nextExecution) {
                continue;
            }
            if (runnable._nextExecution < time) {
                try {
                    runnable.run();
                } catch (ex) {
                    runnable._nextExecution = null;
                    throw(ex);
                }
                if (!runnable._nextExecution) {
                    continue;
                }
                if (runnable.timeInterval && runnable.repeat) {
                    runnable._nextExecution = runnable.timeInterval + time;
                } else {
                    runnable._nextExecution = null;
                }
            }
        }
    
        var newRunnables = [];
        for (var i = 0; i < WebCore.Scheduler._runnables.length; ++i) {
            var runnable = WebCore.Scheduler._runnables[i];
            if (runnable._nextExecution) {
                newRunnables.push(runnable);
            }
        }
        WebCore.Scheduler._runnables = newRunnables;
        
        if (WebCore.Scheduler._runnables.length == 0) {
            WebCore.Scheduler._stop();
        }
    },
    
    /**
     * Enqueues a Runnable to be executed by the scheduler.
     * 
     * @param {WebCore.Scheduler.Runnable} the runnable to enqueue
     */
    add: function(runnable) {
        var currentTime = new Date().getTime();
        runnable._nextExecution = runnable.timeInterval ? runnable.timeInterval + currentTime : currentTime;
        WebCore.Scheduler._runnables.push(runnable);
        WebCore.Scheduler._start();
    },
    
    /**
     * Dequeues a Runnable so it will no longer be executed by the scheduler.
     * 
     * @param {WebCore.Scheduler.Runnable} the runnable to dequeue
     */
    remove: function(runnable) {
        runnable._nextExecution = null;
    },
    
    /**
     * Creates a new Runnable that executes the specified method and enqueues it into the scheduler.
     * 
     * @param {Number} time the time interval, in milleseconds, after which the Runnable should be executed
     *        (may be null/undefined to execute task immediately, in such cases repeat must be false)
     * @param {Boolean} repeat a flag indicating whether the task should be repeated
     * @param f a function to invoke, may be null/undefined
     * @return the created Runnable.
     * @type WebCore.Scheduler.Runnable 
     */
    run: function(f, timeInterval, repeat) {
        var runnable = new WebCore.Scheduler.MethodRunnable(f, timeInterval, repeat);
        this.add(runnable);
        return runnable;
    },
    
    /**
     * Starts the scheduler "thread".
     * If the scheduler is already running, no action is taken.
     * @private
     */
    _start: function() {
        if (WebCore.Scheduler._interval != null) {
            return;
        }
        WebCore.Scheduler._interval = window.setInterval(WebCore.Scheduler._execute, WebCore.Scheduler.INTERVAL);
    },
    
    /**
     * Stops the scheduler "thread".
     * If the scheduler is not running, no action is taken.
     * @private
     */
    _stop: function() {
        if (WebCore.Scheduler._interval == null) {
            return;
        }
        window.clearInterval(WebCore.Scheduler._interval);
        WebCore.Scheduler._interval = null;
    }
};

/**
 * @class A runnable task that may be scheduled with the Scheduler.
 */
WebCore.Scheduler.Runnable = Core.extend({
    
    $virtual: {

        /** 
         * Time interval, in milleseconds after which the Runnable should be executed.
         * @type Number
         */
        timeInterval: null,
        
        /**
         * Flag indicating whether task should be repeated.
         * @type Boolean
         */
        repeat: false
    },

    $abstract: {
        
        run: function() { }
    }
});

/**
 * @class A runnable task implemenation that invokes a function at regular intervals.
 */
WebCore.Scheduler.MethodRunnable = Core.extend(WebCore.Scheduler.Runnable, {

    f: null,

    /**
     * Creates a new Runnable.
     *
     * @constructor
     * @param {Number} time the time interval, in milleseconds, after which the Runnable should be executed
     *        (may be null/undefined to execute task immediately, in such cases repeat must be false)
     * @param {Boolean} repeat a flag indicating whether the task should be repeated
     * @param {Function} f a function to invoke, may be null/undefined
     */
    $construct: function(f, timeInterval, repeat) {
        if (!timeInterval && repeat) {
            throw new Error("Cannot create repeating runnable without time delay:" + f);
        }
        this.f = f;
        this.timeInterval = timeInterval;
        this.repeat = !!repeat;
    },

    $virtual: {
        
        /**
         * Default run() implementation. Should be overidden by subclasses.
         */
        run: function() {
            this.f();
        }
    }
});

/**
 * @class
 * Static object/namespace which provides cross-platform CSS positioning 
 * capabilities. Do not instantiate.
 * <p>
 * Internet Explorer 6 is ordinarily handicapped by its lack
 * of support for setting 'left' and 'right' or 'top' and 'bottom' positions
 * simultaneously on a single document element.
 * <p> 
 * To use virtual positioning, simply set the left/right/top/bottom
 * coordinates of an element and invoke redraw().  The redraw() method
 * must be invoked whenever the size of the element should be redrawn,
 * e.g., when the screen or its containing element resizes.
 */
WebCore.VirtualPosition = {

    /** Flag indicating whether virtual positioning is required/enabled. */
    _enabled: false,
    
    /**
     * Enables and initializes the virtual positioning system.
     */
    _init: function() {
        this._enabled = true;
    },
    
    /**
     * Redraws elements registered with the virtual positioning system.
     * Adjusts the style.height and style.width attributes of an element to 
     * simulate its specified top, bottom, left, and right CSS position settings
     * The calculation makes allowances for padding, margin, and border width.
     *
     * @param element the element to redraw
     */
    redraw: function(element) { }
};

/**
 * @fileoverview
 * Application framework main module.
 * Requires Core.
 */

/**
 * @class Namespace for application framework.
 */
EchoApp = { 

    /**
     * Next unique identifier.
     */
    _nextUid: 1,

    /**
     * Generates a unique identifier.  Identifiers are unique for the duration of the existence of this namespace.
     */
    generateUid: function() {
        return this._nextUid++;
    }
};

/**
 * @class 
 * Representation of a single application instance.
 * Derived objects must invoke construtor with root component id.
 */
EchoApp.Application = Core.extend({
    
    $abstract: true,

    /** 
     * Mapping between component ids and component instances.
     * @private 
     * @type Core.Arrays.LargeMap
     */
    _idToComponentMap: null,

    /** 
     * ListenerList instance for application-level events.
     * @private 
     * @type Core.ListenerList 
     */
    _listenerList: null,
        
    /** 
     * Array of modal components.
     * This value is read-only.
     * @type Array 
     */
    _modalComponents: null,

    /** 
     * Displayed style sheet.
     * 
     * @private 
     * @type EchoApp.StyleSheet
     */
    _styleSheet: null,
    
    /** 
     * Currently focused component.
     * @private
     * @type EchoApp.Component
     */
    _focusedComponent: null,
    
    /** 
     * Root component instance.
     * This value is read-only.
     * @type EchoApp.Component 
     */
    rootComponent: null,
    
    /** 
     * UpdateManager instance monitoring changes to the application for redraws. 
     * @type EchoApp.Update.Manager
     */
    updateManager: null,
    
    /**
     * FocusManager instance handling application focus behavior.
     * @type EchoApp.FocusManager
     */
    focusManager: null,
    
    /**
     * Creates a new application instance.  
     * @constructor
     */
    $construct: function() {
        this._idToComponentMap = new Core.Arrays.LargeMap();
        this._listenerList = new Core.ListenerList();
        this.rootComponent = new EchoApp.Component();
        this.rootComponent.componentType = "Root";
        this.rootComponent.register(this);
        this._modalComponents = [];
        this.updateManager = new EchoApp.Update.Manager(this);
        this.focusManager = new EchoApp.FocusManager(this);
    },

    /**
     * Adds a ComponentUpdateListener.
     * 
     * @param {Function} l the listener to add
     */
    addComponentUpdateListener: function(l) {
        this._listenerList.addListener("componentUpdate", l);
    },
    
    /**
     * Adds a FocusListener.  Focus listeners will be invoked when the focused
     * component in the application changes.
     * 
     * @param {Function} l the listener to add
     */
    addFocusListener: function(l) {
        this._listenerList.addListener("focus", l);
    },
    
    /**
     * Disposes of the application.
     * Once invoked, the application will no longer function and cannot be used again.
     * This method will free any resources allocated by the application.
     */ 
    dispose: function() {
        this.updateManager.dispose();
    },
    
    /**
     * Recurisvely determines the current root component of the modal context.
     *
     * @param {EchoApp.Component} searchComponent (optional) the component from which to search
     *        (this paramater is provided when recursively searching, if omitted the sear
     *        will begin at the root component of the application).
     * @return the current modal context root component
     */
    _findModalContextRoot: function(searchComponent) {
        searchComponent = searchComponent ? searchComponent : this.rootComponent;
        for (var i = searchComponent.children.length - 1; i >= 0; --i) {
            var foundComponent = this._findModalContextRoot(searchComponent.children[i]);
            if (foundComponent) {
                return foundComponent;
            }
        }
        
        if (searchComponent.modalSupport && searchComponent.get("modal")) {
            return searchComponent;
        }
        
        return null;
    },

    /**
     * Focuses the previous/next component based on the currently focused component.
     * 
     * @param {Boolean} reverse false to focus the next component, true to focus the
     *        previous component
     */
    focusNext: function(reverse) {
        focusedComponent = this.focusManager.find(null, reverse);
        if (focusedComponent != null) {
            this.setFocusedComponent(focusedComponent);
        }
    },
    
    /**
     * Retrieves the registered component with the specified render id.
     * 
     * @param {String} renderId the render id
     * @return the component
     * @type EchoApp.Component 
     */
    getComponentByRenderId: function(renderId) {
        return this._idToComponentMap.map[renderId];
    },
    
    /**
     * Returns the focused component.
     * 
     * @return the focused component
     * @type {EchoApp.Component}
     */
    getFocusedComponent: function() {
        return this._focusedComponent;
    },
    
    /**
     * Returns the default layout direction of the application.
     *
     * @return the default layout direction
     * @type EchoApp.LayoutDirection 
     */
    getLayoutDirection: function() {
        // FIXME ensure layout direction gets set upon application instantiation
        return this._layoutDirection ? this._layoutDirection : EchoApp.LayoutDirection.LTR;
    },
        
    /**
     * Returns the root component of the modal context.
     *
     * @return the root component of the modal context
     */
    getModalContextRoot: function() {
        if (this._modalComponents.length == 0) {
            return null;
        } else if (this._modalComponents.length == 1) {
            return this._modalComponents[0];
        }
        
        return this._findModalContextRoot();
    },
    
    /**
     * Returns the application style sheet.
     * 
     * @return the application style sheet
     * @type EchoApp.StyleSheet
     */
    getStyleSheet: function() {
        return this._styleSheet;
    },
    
    /**
     * Returns the active state of the application.
     * 
     * @return the active state of the application, a value of 
     *         true indicating the application is ready for user
     *         input, a value of false indicating otherwise
     * @type Boolean
     */
    isActive: function() {
        return true;
    },
    
    /**
     * Notifies the application of an update to a component.
     * 
     * @param {EchoApp.Component} parent the parent component
     * @param {String} propertyName the updated property
     * @param oldValue the previous property value
     * @param newValue the new property value
     */
    notifyComponentUpdate: function(parent, propertyName, oldValue, newValue) {
        if (parent.modalSupport && propertyName == "modal") {
            this._setModal(parent, newValue);
        }
        if (this._listenerList.hasListeners("componentUpdate")) {
            this._listenerList.fireEvent({type: "componentUpdate", parent: parent, propertyName: propertyName, 
                    oldValue: oldValue, newValue: newValue});
        }
        this.updateManager._processComponentUpdate(parent, propertyName, oldValue, newValue);
    },
    
    /**
     * Registers a component with the application.
     * Invoked when a component is added to a hierarchy of 
     * components that is registered with the application.
     * 
     * @param {EchoApp.Component} component the component to register
     * @private
     */
    _registerComponent: function(component) {
        if (this._idToComponentMap.map[component.renderId]) {
            throw new Error("Component already exists with id: " + component.renderId);
        }
        this._idToComponentMap.map[component.renderId] = component;
        if (component.modalSupport && component.get("modal")) {
            this._setModal(component, true);
        }
    },
    
    /**
     * Removes a ComponentUpdateListener.
     * 
     * @param {Function} l the listener to remove
     */
    removeComponentUpdateListener: function(l) {
        this._listenerList.removeListener("componentUpdate", l);
    },
    
    /**
     * Removes a FocusListener.  Focus listeners will be invoked when the focused
     * component in the application changes.
     * 
     * @param {Function} l the listener to remove
     */
    removeFocusListener: function(l) {
        this._listenerList.removeListener("focus", l);
    },
    
    /**
     * Sets the focused component
     * 
     * @param {EchoApp.Component} newValue the new focused component
     */
    setFocusedComponent: function(newValue) {
        // If required, find focusable parent containing 'newValue'.
        while (newValue != null && !newValue.focusable) {
            newValue = newValue.parent;
        }
        
        // Verify new focused component is within modal context.
        if (this._modalComponents.length > 0) {
            var modalContextRoot = this.getModalContextRoot();
            if (!modalContextRoot.isAncestorOf(newValue)) {
                // Reject request to focus component outside of modal context.
                Core.Debug.consoleWrite("not in modal:" + newValue);
                return;
            }
        }
        
        this._focusedComponent = newValue;
        this._listenerList.fireEvent({type: "focus", source: this});
    },
    
    /**
     * Sets the application default layout direction.
     * 
     * @param {EchoApp.LayoutDirection} newValue the new layout direction
     */
    setLayoutDirection: function(newValue) {
        this._layoutDirection = newValue;
    },
    
    /**
     * Informs the application of the modal state of a specific component.
     * When modal components are unregistered, this method must be executed
     * in order to avoid a memory leak.
     */
    _setModal: function(component, modal) {
        Core.Arrays.remove(this._modalComponents, component);
        if (modal) {
            this._modalComponents.push(component);
        }
        
        // Auto-focus first component in modal context if component is currently focused component is not within modal context.
        if (this._modalComponents.length > 0 && this._focusedComponent) {
            var modalContextRoot = this.getModalContextRoot();
            if (!modalContextRoot.isAncestorOf(this._focusedComponent)) {
                if (modalContextRoot.focusable) {
                    this.setFocusedComponent(modalContextRoot);
                } else {
                    this.setFocusedComponent(this.focusManager.findInParent(modalContextRoot, false));
                }
            }
        }
    },
    
    /**
     * Sets the application style sheet.
     * 
     * @param {EchoApp.StyleSheet} newValue the new style sheet
     */
    setStyleSheet: function(newValue) {
        var oldValue = this._styleSheet;
        this._styleSheet = newValue;
    // FIXME updatemanager can't handle this yet.    
    //    this.notifyComponentUpdate(null, "styleSheet", oldValue, newValue);
    },
    
    /**
     * Unregisters a component from the application.
     * This method is invoked when a component is removed from a hierarchy of 
     * components registered with the application.
     * 
     * @param {EchoApp.Component} component the component to remove
     * @private
     */
    _unregisterComponent: function(component) {
        this._idToComponentMap.remove(component.renderId);
        if (component.modalSupport) {
            this._setModal(component, false);
        }
    }
});

/**
 * @class
 * Factory to create new instances of arbitrary components.  This object is 
 * used to instantiate new components during XML deserialization.
 * This is a namespace object, do not instantiate.
 */
EchoApp.ComponentFactory = {
    
    /**
     * Mapping between type names and object constructors.
     * 
     * @type Object
     * @private
     */
    _typeToConstructorMap: {},
    
    /**
     * Creates a new instance of an arbitrary component.
     * 
     * @constructor
     * @param {String} typeName the type name of the component
     * @param {String} renderId the component render id
     * @return a newly instantiated component
     * @type EchoApp.Component
     */
    newInstance: function(typeName, renderId) {
        var typeConstructor = this._typeToConstructorMap[typeName];
        if (!typeConstructor) {
            throw new Error("Type not registered: " + typeName);
        }
        var component = new typeConstructor();
        component.renderId = renderId;
        return component;
    },
    
    /**
     * Returns the component constructor for the specified type.
     *
     * @param {String} typeName the type name
     * @return the component constructor
     * @type Function
     */
    getConstructor: function(typeName) {
        return this._typeToConstructorMap[typeName];
    },
    
    /**
     * Determines the super type of a component, based on the type name of the component.
     *
     * @param {String} typeName the component type
     * @return the parent componetn type
     * @type String
     */
    getSuperType: function(typeName) {
        var typeConstructor = this._typeToConstructorMap[typeName];
        if (!typeConstructor) {
            throw new Error("Type not found: " + typeName + ".");
        }
        if (typeConstructor.$super) {
            return typeConstructor.$super.prototype.componentType;
        } else {
            return null;
        }
    },
    
    /**
     * Registers a type name to a specific constructor.
     * 
     * @param typeName the type name
     * @param typeConstructor the constructor
     */
    registerType: function(typeName, typeConstructor) {
        if (this._typeToConstructorMap[typeName]) {
            throw new Error("Type already registered: " + typeName);
        }
        this._typeToConstructorMap[typeName] = typeConstructor;
    }
};

/**
 * @class 
 * Base class for components.
 * Derived classes should always take renderId as the first parameter and pass it to the super-constructor.
 * A component MUST have its componentType property set before it is used in a hierarchy.  Failing to do so
 * will throw an exception and/or result in indeterminate behavior.
 */
EchoApp.Component = Core.extend({
    
    $static: {

        /**
         * The next automatically assigned client render id.
         * @private
         * @type Number
         */
        _nextRenderId: 0
    },
    
    $load: function() {
        EchoApp.ComponentFactory.registerType("Component", this);
    },

    $abstract: true,
    
    $virtual: {
    
        /**
         * Component type.  This must be set by implementors in order for peer discovery to work properly.
         */
        componentType: "Component",

        /**
         * Returns the child component at the specified index
         * after sorting the children in the order which they 
         * should be focused.  The default implementation simply
         * returns the same value as getComponent().
         * Implementations should override this method when
         * the natural order to focus child components is
         * different than their normal ordering (e.g., when
         * the component at index 1 is positioned above the 
         * component at index 0).
         * 
         * @param index the index of the child (in focus order)
         * @return the child component
         */
        getFocusComponent: function(index) {
            return this.children[index];
        }
    },

    /**
     * The render id.
     * This value should be treated as read-only and immutable.
     * @type String
     */
    renderId: null,
    
    /**
     * The parent component.
     * This value is read-only.
     * @type EchoApp.Component
     */
    parent: null,
    
    /**
     * The registered application.
     * This value is read-only.
     * @type EchoApp.Application
     */
    application: null,
    
    /**
     * Listener list.  Lazily created.
     * @private
     * @type Core.ListenerList
     */
    _listenerList: null,
    
    /**
     * Referenced external style
     * @private
     * @type Object
     */
    _style: null,
    
    /**
     * Assigned style name from application-level style sheet.
     * @private
     * @type String
     */
    _styleName: null,

    /**
     * Enabled state of the component (default true).
     * @private
     * @type Boolean
     */
    _enabled: true,
    
    /**
     * Array of child components.
     * This value is read-only.  Modifying this array will result in undefined behavior.
     * @type Array
     */
    children: null,
    
    /**
     * Internal style used to store properties set directly on component.
     * @private
     * @type Object
     */
    _localStyle: null,
    
    /**
     * Creates a new Component.
     *  
     * @param {String} renderId the render id
     * @param {Object} associative mapping of initial property values (may be null)
     *        By default, all properties will be placed into the local style, except for the following:
     *        <ul>
     *         <li><code>styleName</code> specifies the component stylesheet style name</li>
     *         <li><code>style</code> specifies the referenced component style</li>
     *         <li><code>renderId</code> specifies the render id</li>
     *        </ul>
     * @constructor
     */
    $construct: function(properties) {
        this.children = [];
        this._localStyle = { };
        
        if (properties) {
            for (var name in properties) {
                switch (name) {
                case "style": this._style = properties.style; break;
                case "styleName": this._styleName = properties.styleName; break;
                case "renderId": this._renderId = properties.renderId; break;
                case "children":
                    for (var i = 0; i < properties.children.length; ++i) {
                        this.add(properties.children[i]);
                    }
                    break;
                case "events":
                    for (var eventType in properties.events) {
                        this.addListener(eventType, properties.events[eventType]);
                    }
                    break;
                default:
                    this._localStyle[name] = properties[name];
                    break;
                }
            }
        }
    },

    /**
     * Adds a component as a child.
     * 
     * @param {EchoApp.Component} component the component to add
     * @param {Number} index the (integer) index at which to add it (optional, omission
     *        will cause component to be appended to end)
     */
    add: function(component, index) {
        if (!(component instanceof EchoApp.Component)) {
            throw new Error("Cannot add child: specified component object is not derived from EchoApp.Component. "
                    + "Parent: " + this + ", Child: " + component);
        }
        if (!component.componentType) {
            throw new Error("Cannot add child: specified component object does not have a componentType property. "
                    + "Parent: " + this + ", Child: " + component);
        }
    
        if (component.parent) {
            component.parent.remove(component);
        }
        
        component.parent = this;
            
        if (index == null || index == this.children.length) {
            this.children.push(component);
        } else {
            this.children.splice(index, 0, component);
        }
        
        if (this.application) {
            component.register(this.application);
            this.application.notifyComponentUpdate(this, "children", null, component);
        }
    },
    
    /**
     * Adds an arbitrary event listener.
     * 
     * @param {String} eventType the event type name
     * @param {Function} eventTarget the method to invoke when the event occurs 
     *        (the event will be passed as the single argument)
     */
    addListener: function(eventType, eventTarget) {
        if (this._listenerList == null) {
            this._listenerList = new Core.ListenerList();
        }
        this._listenerList.addListener(eventType, eventTarget);
        if (this.application) {
            this.application.notifyComponentUpdate(this, "listeners", null, eventType);
        }
    },
    
    /**
     * Provides notification of an arbitrary event.
     * Listeners will be notified based on the event's type property.
     * 
     * @param event the event to fire
     */
    fireEvent: function(event) {
        if (this._listenerList == null) {
            return;
        }
        this._listenerList.fireEvent(event);
    },
    
    /**
     * Returns an arbitrary property value.
     * 
     * @param {String} name the name of the property
     * @return the property value
     */
    get: function(name) {
        return this._localStyle[name];
    },
    
    /**
     * Retrieves the child component at the specified index.
     * 
     * @param {Number} index the (integer) index
     * @return the child component
     * @type EchoApp.Component
     */
    getComponent: function(index) {
        return this.children[index];
    },
    
    /**
     * Returns the number of child components
     * 
     * @return the number of child components
     * @type Number
     */
    getComponentCount: function() {
        return this.children.length;
    },
    
    /**
     * Returns an arbitrary indexed property value.
     * 
     * @param {String} name the name of the property
     * @param {Number} index the index to return
     * @return the property value
     */
    getIndex: function(name, index) {
        var valueArray = this._localStyle[name];
        return valueArray ? valueArray[index] : null;
    },
    
    /**
     * Returns the component layout direction.
     * 
     * @return the component layout direction
     * @type {EchoApp.LayoutDirection}
     */
    getLayoutDirection: function() {
        return this._layoutDirection;
    },
    
    /**
     * Retrieves local style property map associations.
     * This method should only be used by a deserialized for
     * the purpose of rapidly loading properties into a new
     * component.
     * 
     * @return the internal style property map associations
     *         (an associative array).
     */
    getLocalStyleData: function() {
        return this._localStyle;
    },
    
    /**
     * Returns the layout direction with which the component should be
     * rendered, based on analyzing the component's layout direction,
     * its parent's, and/or the application's.
     * 
     * @return the rendering layout direction
     * @type EchoApp.LayoutDirection
     */
    getRenderLayoutDirection: function() {
        var component = this;
        while (component) {
            if (component._layoutDirection) {
                return component._layoutDirection;
            }
            component = component.parent;
        }
        if (this.application) {
            return this.application.getLayoutDirection();
        }
        return null;
    },
    
    /**
     * Returns the style assigned to this component, if any.
     * 
     * @return the assigned style
     * @type Object
     */
    getStyle: function() {
        return this._style;
    },
    
    /**
     * Returns the name of the style (from the application's style sheet) 
     * assigned to this component.
     * 
     * @return the style name
     * @type String
     */
    getStyleName: function() {
        return this._styleName;
    },
    
    /**
     * Returns the index of a child component, or -1 if the component
     * is not a child.
     * 
     * @param {EchoApp.Component} component the component
     * @return the index
     * @type Number
     */
    indexOf: function(component) {
        for (var i = 0; i < this.children.length; ++i) {
            if (this.children[i] == component) {
                return i;
            }
        }
        return -1;
    },
    
    /**
     * Determines if the component is active, that is, within the current modal context
     * and ready to receive input.
     * 
     * @return the active state
     * @type Boolean
     */
    isActive: function() {
        // Verify the component and its ancestors are all enabled.
        if (!this.isRenderEnabled()) {
            return false;
        }
        
        // Verify component is registered to an application, and that the application is active.
        if (!this.application || !this.application.isActive()) {
            return false;
        }
        
        // Verify component is in modal context.
        var modalContextRoot = this.application.getModalContextRoot();
        if (modalContextRoot != null && !modalContextRoot.isAncestorOf(this)) {
            return false;
        }
        
        return true;
    },
    
    /**
     * Determines if this component is or is an ancestor of another component.
     * 
     * @param {EchoApp.Component} c the component to test
     * @return true if an ancestor relationship exists
     * @type Boolean
     */
    isAncestorOf: function(c) {
        while (c != null && c != this) {
            c = c.parent;
        }
        return c == this;
    },
    
    /**
     * Determines the enabled state of this component.
     * Use isRenderEnabled() to determine whether a component
     * should be RENDERED as enabled.
     * 
     * @return the enabled state of this specific component
     */
    isEnabled: function() {
        return this._enabled;
    },
    
    /**
     * Determines whether this <code>Component</code> should be rendered with
     * an enabled state.
     * Disabled <code>Component</code>s are not eligible to receive user input.
     * 
     * @return true if the component should be rendered enabled.
     * @type Boolean
     */
    isRenderEnabled: function() {
        var component = this;
        while (component != null) {
            if (!component._enabled) {
                return false;
            }
            component = component.parent;
        }
        return true;
    },
    
    /**
     * Registers / unregisters a component that has been 
     * added/removed to/from a registered hierarchy
     * (a hierarchy that is registered to an application).
     * 
     * @param {EchoApp.Application} application the application 
     *        (null to unregister the component)
     */
    register: function(application) {
        // Sanity check.
        if (application && this.application) {
            throw new Error("Attempt to re-register or change registered application of component.");
        }
    
        if (!application) { // unregistering
            
            if (this.children != null) {
                // Recursively unregister children.
                for (var i = 0; i < this.children.length; ++i) {
                     this.children[i].register(false); // Recursively unregister children.
                }
            }
            
            // Notify application.
            this.application._unregisterComponent(this);
            
            // Change application focus in the event the focused component is being removed.
            // Note that this is performed after deregistration to ensure any removed modal context is cleared.
            if (this.application._focusedComponent == this) {
                this.application.setFocusedComponent(this.parent);
            }
        }
    
        // Link/unlink with application.
        this.application = application;
    
        if (application) { // registering
            
            if (this.renderId == null) {
                this.renderId = "cl_" + ++EchoApp.Component._nextRenderId;
            }
    
            // Notify application.
            this.application._registerComponent(this);
            
            if (this.children != null) {
                // Recursively register children.
                for (var i = 0; i < this.children.length; ++i) {
                     this.children[i].register(application); // Recursively unregister children.
                }
            }
        }
    },
    
    /**
     * Returns the value of a property that should be rendered,
     * based on the value set on this component, in the component's
     * specified style, and/or in the application's stylesheet.
     * 
     * @param {String} name the name of the property
     * @param defaultValue the default value to return if no value is 
     *        specified in an internal property, style, or stylesheet
     * @return the property value
     */
    render: function(name, defaultValue) {
        var value = this._localStyle[name];
        if (value == null) {
            if (this._style != null) {
                value = this._style[name];
            }
            if (value == null && this._styleName && this.application && this.application._styleSheet) {
                var style = this.application._styleSheet.getRenderStyle(this._styleName, this.componentType);
                if (style) {
                    value = style[name];
                }
            }
        }
        return value == null ? defaultValue : value;
    },
    
    /**
     * Returns the value of an indexed property that should be rendered,
     * based on the value set on this component, in the component's
     * specified style, and/or in the application's stylesheet.
     * 
     * @param {String} name the name of the property
     * @param {Number} index the (integer) index of the property
     * @param defaultValue the default value to return if no value is 
     *        specified in an internal property, style, or stylesheet
     */
    renderIndex: function(name, index, defaultValue) {
        var valueArray = this._localStyle[name];
        var value = valueArray ? valueArray[index] : null;
        if (value == null) {
            if (this._style != null) {
                valueArray = this._style[name];
                value = valueArray ? valueArray[index] : null;
            }
            if (value == null && this._styleName && this.application && this.application._styleSheet) {
                var style = this.application._styleSheet.getRenderStyle(this._styleName, this.componentType);
                if (style) {
                    valueArray = this.style[name];
                    value = valueArray ? valueArray[index] : null;
                }
            }
        }
        return value == null ? defaultValue : value;
    },
    
    /**
     * Removes a child component.
     * 
     * @param componentOrIndex 
     *        the index of the component to remove, or the component to remove
     *        (values may be of type EchoApp.Component or Number)
     */
    remove: function(componentOrIndex) {
        var component;
        var index;
        if (typeof componentOrIndex == "number") {
            index = componentOrIndex;
            component = this.children[index];
            if (!component) {
                throw new Error("Index out of bounds: " + index);
            }
        } else {
            component = componentOrIndex;
            index = this.indexOf(component);
            if (index == -1) {
                // Component is not a child: do nothing.
                return;
            }
        }
        
        if (this.application) {
            component.register(null);
        }
        
        this.children.splice(index, 1);
        component.parent = null;
        
        if (this.application) {
            this.application.notifyComponentUpdate(this, "children", component, null);
        }
    },
    
    /**
     * Removes all child components.
     */
    removeAll: function() {
        while (this.children.length > 0) {
            this.remove(this.children.length - 1);
        }
    },
    
    /**
     * Removes an arbitrary event listener.
     * 
     * @param {String} eventType the event type name
     * @param {Function} eventTarget the method to invoke when the event occurs 
     *        (the event will be passed as the single argument)
     */
    removeListener: function(eventType, eventTarget) {
        if (this._listenerList == null) {
            return;
        }
        this._listenerList.removeListener(eventType, eventTarget);
        if (this.application) {
            this.application.notifyComponentUpdate(this, "listeners", eventType, null);
        }
    },
    
    /** 
     * Sets the value of a property in the internal style.
     * 
     * @param {String} name the name of the property
     * @param value the new value of the property
     */
    set: function(name, newValue) {
        var oldValue = this._localStyle[name];
        this._localStyle[name] = newValue;
        if (this._listenerList && this._listenerList.hasListeners("property")) {
            this._listenerList.fireEvent({type: "property", source: this, propertyName: name, 
                    oldValue: oldValue, newValue: newValue});
        }
        if (this.application) {
            this.application.notifyComponentUpdate(this, name, oldValue, newValue);
        }
    },
    
    /**
     * Sets the enabled state of the component.
     * 
     * @param newValue the new enabled state
     */
    setEnabled: function(newValue) {
        var oldValue = this._enabled;
        this._enabled = newValue;
        if (this.application) {
            this.application.notifyComponentUpdate(this, "enabled", oldValue, newValue);
        }
    },
    
    /** 
     * Sets the value of an indexed property in the internal style.
     * 
     * @param {String} name the name of the property
     * @param {Number} index the index of the property
     * @param newValue the new value of the property
     */
    setIndex: function(name, index, newValue) {
        var valueArray = this._localStyle[name];
        var oldValue = null;
        if (valueArray) {
            oldValue = valueArray[index];
        } else {
            valueArray = [];
            this._localStyle[name] = valueArray;
        }
        valueArray[index] = newValue;
        if (this.application) {
            this.application.notifyComponentUpdate(this, name, oldValue, newValue);
        }
        if (this._listenerList && this._listenerList.hasListeners("property")) {
            this._listenerList.fireEvent({type: "property", source: this, propertyName: name, index: index,
                    oldValue: oldValue, newValue: newValue});
        }
    },
    
    /**
     * Sets a component-specific layout direction.
     * 
     * @param {EchoApp.LayoutDirection} newValue the new layout direction
     */
    setLayoutDirection: function(newValue) {
        this._layoutDirection = newValue;
    },
    
    /**
     * Sets the style of the component.
     * 
     * @param {Object} newValue the new style
     */
    setStyle: function(newValue) {
        var oldValue = this._style;
        this._style = newValue;
        if (this.application) {
            this.application.notifyComponentUpdate(this, "style", oldValue, newValue);
        }
    },
    
    /**
     * Sets the name of the style (from the application's style sheet) 
     * assigned to this component.
     * 
     * @param {String} newValue the style name
     */
    setStyleName: function(newValue) {
        var oldValue = this._styleName;
        this._styleName = newValue;
        if (this.application) {
            this.application.notifyComponentUpdate(this, "styleName", oldValue, newValue);
        }
    },
    
    /**
     * Returns a string representation of the component (default implementation).
     * 
     * @param {Boolean} longFormat an optional flag specifying whether all information about
     *        the component should be displayed (e.g., property values)
     * @return a string representation of the component
     * @type String
     */
    toString: function(longFormat) {
        var out = this.renderId + "/" + this.componentType;
        if (longFormat) {
            out += "\n";
            var componentCount = this.getComponentCount();
            out += this.renderId + "/properties:" + this._localStyle + "\n";
            for (var i = 0; i < componentCount; ++i) {
                var component = this.getComponent(i);
                out += this.renderId + "/child:" + component.renderId + "\n";
                out += component.toString(true);
            }
        }
        return out;
    }
});

/**
 * Provides focus management tools for an application.
 */
EchoApp.FocusManager = Core.extend({

    _application: null,

    /**
     * Focus management handler for a specific application instance.
     * One FocusManager is created for each application.
     */
    $construct: function(application) { 
        this._application = application;
    },
    
    /**
     * Searches the component hierarchy for the next component that should
     * be focused (based on the currently focused component).
     * Container components are queried to determine the order in which their
     * children should naturally be focused (certain components, e.g., SplitPanes,
     * will have a child focus order that may be different from the order of their 
     * children).
     * This search is depth first.
     * 
     * Search order (FORWARD):
     * 
     * (Start on Component)
     * First Child, next sibling, parent
     *
     * Search order (REVERSE):
     * Last Child, previous sibling, parent
     * 
     *
     * @return the Component which should be focused
     * @type EchoApp.Component
     */
    find: function(component, reverse) {
        if (!component) {
            component = this._application.getFocusedComponent();
            if (!component) {
                component = this._application.rootComponent;
            }
        }

        /** The component which is currently focused by the application. */
        var originComponent = component;
        
        /** An associative array containing the ids of all previously visited components. */
        var visitedComponents = { };
        
        /** The value of 'component' on the previous iteration. */
        var lastComponent = null;
        
        while (true) {
            /** The candidate next component to be focused */
            var nextComponent = null;

            if ((reverse && component == originComponent) || (lastComponent && lastComponent.parent == component)) {
                // Searching in reverse on origin component (OR) Previously moved up: do not move down.
            } else {
                var componentCount = component.getComponentCount();
                if (componentCount > 0) {
                    // Attempt to move down.
                    // Next component is first child (searching forward) or last child (searching reverse).
                    nextComponent = component.getComponent(reverse ? componentCount - 1 : 0);
                    if (visitedComponents[nextComponent.renderId]) {
                        // Already visited children, cancel the move.
                        nextComponent = null;
                    }
                }
            }
                
            if (nextComponent == null) {
                // Attempt to move to next/previous sibling.
                if (component.parent) {
                    // Get previous sibling.
                    if (reverse) {
                        var componentIndex = component.parent.indexOf(component);
                        if (componentIndex > 0) {
                            nextComponent = component.parent.getComponent(componentIndex - 1);
                        }
                    } else {
                        var componentIndex = component.parent.indexOf(component);
                        if (componentIndex < component.parent.getComponentCount() - 1) {
                            nextComponent = component.parent.getComponent(componentIndex + 1);
                        }
                    }
                }
            }

            if (nextComponent == null) {
                // Attempt to move up.
                nextComponent = component.parent;
            }
            
            if (nextComponent == null) {
                return null;
            }

            lastComponent = component;
            component = nextComponent;
            visitedComponents[component.renderId] = true;
            
            if (component != originComponent && component.isActive() && component.focusable) {
                return component;
            }
        }
    },
    
    /**
     * Finds next (or previous) focusable descendant of a parent component.
     * This method requires that the application's currently focused component
     * be a descendant of the specified parent component.  The search will
     * be limited to descendants of the parent component, i.e., if a suitable descendant
     * component cannot be found, null will be returned.
     *
     * @param parentComponent the parent component to search
     * @param reverse the search direction, false indicating to search forward, true
     *        indicating reverse
     * @param minimumDistance FIXME
     */
    findInParent: function(parentComponent, reverse, minimumDistance) {
        if (!minimumDistance) {
            minimumDistance = 1;
        }
        
        var focusedComponent = this._application.getFocusedComponent();
        
        var focusedIndex = this._getDescendantIndex(parentComponent, focusedComponent);
        if (focusedIndex == -1) {
            return null;
        }
        
        var componentIndex = focusedIndex;
        var component = focusedComponent;
        do {
            component = this.find(component, reverse);
            if (component == null) {
                return null;
            }
            componentIndex = this._getDescendantIndex(parentComponent, component);
        } while (Math.abs(componentIndex - focusedIndex) < minimumDistance && component != focusedComponent);

        if (component == focusedComponent) {
            // Search wrapped, only one focusable component.
            return null;
        }
        
        this._application.setFocusedComponent(component);
        return component;
    },
    
    _getDescendantIndex: function(parentComponent, descendant) {
        while (descendant.parent != parentComponent && descendant.parent != null) {
            descendant = descendant.parent;
        }
        if (descendant.parent == null) {
            return -1;
        }
        return parentComponent.indexOf(descendant);
    }
});

// Fundamental Property Types

/**
 * Describes the layout direction of text and content to provide support 
 * for bidirectional localization.
 */
EchoApp.LayoutDirection = Core.extend({
    
    /**
     * Flag indicating whether layout direction is left-to-right.
     * @type Boolean 
     */
    _ltr: false,
    
    /**
     * LayoutDirection property.  Do not instantiate, use LTR/RTL constants.
     * @constructor
     */
    $construct: function(ltr) {
        this._ltr = ltr;
    },

    /**
     * Determines if the layout direction is left-to-right.
     * 
     * @return true if the layout direction is left-to-right
     * @type Boolean
     */
    isLeftToRight: function() {
        return this._ltr;
    }
});

/**
 * Global instance representing a left-to-right layout direction.
 * @type EchoApp.LayoutDirection
 * @final
 */
EchoApp.LayoutDirection.LTR = new EchoApp.LayoutDirection(true);

/**
 * Global instance representing a right-to-left layout direction.
 * @type EchoApp.LayoutDirection
 * @final
 */
EchoApp.LayoutDirection.RTL = new EchoApp.LayoutDirection(false);

// StyleSheets

/**
 * @class
 * An application style sheet.
 */
EchoApp.StyleSheet = Core.extend({

    _nameToStyleMap: null,

    _renderCache: null,
    
    /**
     * Creates a new style sheet.
     *
     * @param initialValues an optional mapping between style names 
     *        and maps between component types and styles
     */
    $construct: function(initialValues) {
        this._renderCache = { };
        this._nameToStyleMap = { };
        
        if (initialValues) {
            for (var styleName in initialValues) {
                for (var componentType in initialValues[styleName]) {
                     this.setStyle(styleName, componentType, initialValues[styleName][componentType]);
                }
            }
        }
    },
    
    /**
     * Returns the style that should be used for a component.
     * 
     *  @param {String} name the component's style name
     *  @param {String} componentType the type of the component
     *  @return the style
     *  @type Object
     */
    getRenderStyle: function(name, componentType) {
        // Retrieve style from cache.
        var typeToStyleMap = this._renderCache[name];
        if (!typeToStyleMap) {
            return null;
        }
        var style = typeToStyleMap[componentType];
        if (style !== undefined) {
            // If style found in cache, return immediately.
            return style;
        } else {
            return this._loadRenderStyle(name, componentType);
        }
    },
    
    _loadRenderStyle: function(name, componentType) {
        // Retrieve value (type-to-style-map) from name-to-style-map with specified name key.
        var typeToStyleMap = this._nameToStyleMap[name];
        if (typeToStyleMap == null) {
            // No styles available for specified name, mark cache entry as null and return null.
            this._renderCache[name][componentType] = null;
            return null;
        }
        
        // Retrieve style for specific componentType.
        style = typeToStyleMap[componentType];
        if (style == null) {
            var testType = componentType;
            while (style == null) {
                // Search super types of testType to find style until found.
                testType = EchoApp.ComponentFactory.getSuperType(testType);
                if (testType == null) {
                    // No style available for component type, mark cache entry as null and return null.
                    this._renderCache[name][testType] = null;
                    return null;
                }
                style = typeToStyleMap[testType];
            }
        }
        this._renderCache[name][componentType] = style;
        return style;
    },
    
    /**
     * Retrieves a specific style from the style sheet.
     * 
     * @param {String} name the style name
     * @param {String} componentType the component type
     * @return the style
     * @type Object
     */
    getStyle: function(name, componentType) {
        var typeToStyleMap = this._nameToStyleMap[name];
        if (typeToStyleMap == null) {
            return null;
        }
        return typeToStyleMap[componentType];
    },
    
    /**
     * Stores a style in the style sheet.
     * 
     * @param {String} name the style name
     * @param {String} componentType the component type
     * @param {Object} the style
     */
    setStyle: function(name, componentType, style) {
        // Create or clear cache entry for name.
        this._renderCache[name] = {};
        
        var typeToStyleMap = this._nameToStyleMap[name];
        if (typeToStyleMap == null) {
            typeToStyleMap = {};
            this._nameToStyleMap[name] = typeToStyleMap;
        }
        typeToStyleMap[componentType] = style;
    }
});

// Update Management

/**
 * Namespace for update management.  Non-instantiable object.
 * Provides capabilities for storing property changes made to applications and components
 * such that display redraws may be performed efficiently. 
 */
EchoApp.Update = { };

/**
 * @class Representation of an update to a single existing component 
 *        which is currently rendered on the screen.
 */
EchoApp.Update.ComponentUpdate = Core.extend({

    $static: {
    
        /**
         * Data object representing the old and new states of a changed property.
         *
         * @param oldValue the old value of the property
         * @param newValue the new value of the property
         */
        PropertyUpdate: function(oldValue, newValue) {
            this.oldValue = oldValue;
            this.newValue = newValue;
        }
    },
    
    /**
     * The <code>Manager</code> to which this update belongs.
     * @type Array
     */
    _manager: null,
    
    /**
     * The parent component represented in this <code>ServerComponentUpdate</code>.
     * @type EchoApp.Component
     */
    parent: null,
    
    /**
     * Storage for contextual information used by application container to render this update.
     * Object type and content are specified by the application container, this variable is not
     * used by the application module in any capacity.
     */
    renderContext : null,
    
    /**
     * The set of child Component ids added to the <code>parent</code>.
     * @type Array
     */
    _addedChildIds: null,
    
    /**
     * A mapping between property names of the parent component and 
     * <code>PropertyUpdate</code>s.
     * @type Object
     */
    _propertyUpdates: null,
    
    /**
     * The set of child Component ids removed from the <code>parent</code>.
     * @type Array
     */
    _removedChildIds: null,
    
    /**
     * The set of descendant Component ids which are implicitly removed 
     * as they were children of removed children.
     * @type Array
     */
    _removedDescendantIds: null,

    /**
     * The set of child Component ids whose <code>LayoutData</code> 
     * was updated. 
     * @type Array
     */
    _updatedLayoutDataChildIds: null,

    /**
     * Creates a new ComponentUpdate.
     * 
     * @constructor
     * @param parent the updated component
     */
    $construct: function(manager, parent) {
    
        /**
         * The <code>Manager</code> to which this update belongs.
         * @type Array
         */
        this._manager = manager;
        
        /**
         * The parent component represented in this <code>ServerComponentUpdate</code>.
         * @type EchoApp.Component
         */
        this.parent = parent;
    },
    
    /**
     * Records the addition of a child to the parent component.
     * 
     * @param {EchoApp.Component} child the added child
     * @private
     */
    _addChild: function(child) {
        if (!this._addedChildIds) {
            this._addedChildIds = [];
        }
        this._addedChildIds.push(child.renderId);
        this._manager._idMap[child.renderId] = child;
    },
    
    /**
     * Appends removed children and descendants from another update to this
     * update as removed descendants.
     * This method is invoked when a component is removed that is an ancestor
     * of a component that has an update in the update manager.
     * 
     * @private
     * @param {EchoApp.Update.CompoenntUpdate} update the update from which to pull 
     *        removed components/descendants
     */
    _appendRemovedDescendants: function(update) {
        // Append removed descendants.
        if (update._removedDescendantIds != null) {
            if (this._removedDescendantIds == null) {
                this._removedDescendantIds = [];
            }
            for (var x in update._removedDescendantIds) {
                this._removedDescendantIds.push(x);
            }
        }
        
        // Append removed children.
        if (update._removedChildIds != null) {
            if (this._removedDescendantIds == null) {
                this._removedDescendantIds = [];
            }
            for (var x in update._removedChildIds) {
                this._removedDescendantIds.push(x);
            }
        }
        
        if (this._removedDescendantIds != null) {
            Core.Arrays.removeDuplicates(this._removedDescendantIds);
        }
    },
    
    /**
     * Returns an array containing the children added in this update,
     * or null if none were added.
     * 
     * @return the added children
     * @type Array
     */
    getAddedChildren: function() {
        if (!this._addedChildIds) {
            return null;
        }
        var components = new Array(this._addedChildIds.length);
        for (var i = 0; i < this._addedChildIds.length; ++i) {
            components[i] = this._manager._idMap[this._addedChildIds[i]];
        }
        return components;
    },
    
    /**
     * Returns an array containing the children removed in this update,
     * or null if none were removed.
     * 
     * @return the removed children
     * @type Array
     */
    getRemovedChildren: function() {
        if (!this._removedChildIds) {
            return null;
        }
        var components = new Array(this._removedChildIds.length);
        for (var i = 0; i < this._removedChildIds.length; ++i) {
            components[i] = this._manager._removedIdMap[this._removedChildIds[i]];
        }
        return components;
    },
    
    /**
     * Returns an array containing the descendants of any children removed in
     * this update, or null if none were removed.  The removed children
     * themselves are not returned by this method.
     * 
     * @return the removed descendants
     * @type Array
     */
    getRemovedDescendants: function() {
        if (!this._removedDescendantIds) {
            return null;
        }
        var components = new Array(this._removedDescendantIds.length);
        for (var i = 0; i < this._removedDescendantIds.length; ++i) {
            components[i] = this._manager._removedIdMap[this._removedDescendantIds[i]];
        }
        return components;
    },
    
    /**
     * Returns an array containing the children of this component whose
     * LayoutDatas have changed in this update, or null if no such
     * changes were made.
     * 
     * @return the updated layout data children
     * @type Array
     */
    getUpdatedLayoutDataChildren: function() {
        if (!this._updatedLayoutDataChildIds) {
            return null;
        }
        var components = new Array(this._updatedLayoutDataChildIds.length);
        for (var i = 0; i < this._updatedLayoutDataChildIds.length; ++i) {
            components[i] = this._manager._idMap[this._updatedLayoutDataChildIds[i]];
        }
        return components;
    },
    
    /**
     * Determines if any children were added during this update.
     * 
     * @return true if any children were added
     * @type Boolean
     */
    hasAddedChildren: function() {
        return this._addedChildIds != null;
    },
    
    /**
     * Determines if any children were removed during this update.
     * 
     * @return true if any children were removed
     * @type Boolean
     */
    hasRemovedChildren: function() {
        return this._removedChildIds != null;
    },
    
    /**
     * Determines if any children had their LayoutData changed during this update.
     * 
     * @return true if any children had their LayoutData changed
     * @type Boolean
     */
    hasUpdatedLayoutDataChildren: function() {
        return this._updatedLayoutDataChildIds != null;
    },
    
    /**
     * Determines if this update has any changed properties.
     * 
     * @return true if properties are being updated
     * @type Boolean
     */
    hasUpdatedProperties: function() {
        return this._propertyUpdates != null;
    },
    
    /**
     * Returns a <code>PropertyUpdate</code> describing an update to the
     * property with the given <code>name</code>.
     * 
     * @param name the name of the property being updated
     * @return the <code>PropertyUpdate</code>, or null if none exists
     */
    getUpdatedProperty: function(name) {
        if (this._propertyUpdates == null) {
            return null;
        }
        return this._propertyUpdates[name];
    },
    
    /**
     * Returns the names of all properties being updated in this update.
     * 
     * @return the names of all updated properties, if no properties are updated an
     *         empty array is returned
     * @type Array
     */
    getUpdatedPropertyNames: function() {
        if (this._propertyUpdates == null) {
            return [];
        }
        var updatedPropertyNames = [];
        for (var i in this._propertyUpdates) {
            updatedPropertyNames.push(i);
        }
        return updatedPropertyNames;
    },

    /**
     * Determines if the set of updated property names is contained
     * within the specified set.  The provided object should have
     * have keys for the desired property names and  values that evaluate 
     * to true, e.g. to determine if no properties other than "text" and "icon"
     * changed, specify {text: true, icon: true}. 
     */
    isUpdatedPropertySetIn: function(updatedPropertySet) {
        for (var x in this._propertyUpdates) {
            if (!updatedPropertySet[x]) {
                return false;
            }
        }
        return true;
    },
    
    /**
     * Records the removal of a child from the parent component.
     * 
     * @param {EchoApp.Component} child the removed child
     * @private
     */
    _removeChild: function(child) {
        this._manager._removedIdMap[child.renderId] = child;
    
        if (this._addedChildIds) {
            // Remove child from add list if found.
            Core.Arrays.remove(this._addedChildIds, child.renderId);
        }
        
        if (this._updatedLayoutDataChildIds) {
            // Remove child from updated layout data list if found.
            Core.Arrays.remove(this._updatedLayoutDataChildIds, child.renderId);
        }
    
        if (!this._removedChildIds) {
            this._removedChildIds = [];
        }
        
        this._removedChildIds.push(child.renderId);
    
        for (var i = 0; i < child.children.length; ++i) {
            this._removeDescendant(child.children[i]);
        }
    },
    
    /**
     * Records the removal of a descendant of the parent component.
     * All children of a removed compoennt are recorded as removed
     * descendants when the child is removed.
     * This method will recursively invoke itself on children of
     * the specified descendant.
     * 
     * @private
     * @param {EchoApp.Component} descendant the removed descendant 
     */
    _removeDescendant: function(descendant) {
        this._manager._removedIdMap[descendant.renderId] = descendant;
        if (!this._removedDescendantIds) {
            this._removedDescendantIds = [];
        }
        this._removedDescendantIds.push(descendant.renderId);
        for (var i = 0; i < descendant.children.length; ++i) {
            this._removeDescendant(descendant.children[i]);
        }
    },
    
    /**
     * Returns a string representation.
     * 
     * @return a string representation
     * @type String
     */
    toString: function() {
        var s = "ComponentUpdate\n";
        s += "- Parent: " + this.parent + "\n";
        s += "- Adds: " + this._addedChildIds + "\n";
        s += "- Removes: " + this._removedChildIds + "\n";
        s += "- DescendantRemoves: " + this._removedDescendantIds + "\n";
        s += "- Properties: " + this._propertyUpdates + "\n";
        s += "- LayoutDatas: " + this._updatedLayoutDataChildIds + "\n";
        return s;
    },
    
    /**
     * Records the update of the LayoutData of a child component.
     * 
     * @param the child component whose layout data was updated
     * @private
     */
    _updateLayoutData: function(child) {
        this._manager._idMap[child.renderId] = child;
        if (this._updatedLayoutDataChildIds == null) {
            this._updatedLayoutDataChildIds = [];
        }
        this._updatedLayoutDataChildIds.push(child.renderId);
    },
    
    /**
     * Records the update of a property of the parent component.
     * 
     * @param propertyName the name of the property
     * @param oldValue the previous value of the property
     * @param newValue the new value of the property
     * @private
     */
   _updateProperty: function(propertyName, oldValue, newValue) {
        if (this._propertyUpdates == null) {
            this._propertyUpdates = { };
        }
        var propertyUpdate = new EchoApp.Update.ComponentUpdate.PropertyUpdate(oldValue, newValue);
        this._propertyUpdates[propertyName] = propertyUpdate;
    }
});

/**
 * @class Monitors and records updates made to the application between repaints.
 *        Provides API to determine changes to component hierarchy since last update
 *        in order to efficiently repaint the screen.
 */
EchoApp.Update.Manager = Core.extend({
    
    /**
     * Associative mapping between component ids and EchoApp.Update.ComponentUpdate
     * instances.
     * @type Object
     */
    _componentUpdateMap: null,
    
    /**
     * Flag indicating whether a full refresh or incremental update will be performed.
     * @type Boolean
     */
    _fullRefreshRequired: false,
    
    application: null,
    
    /**
     * Flag indicating whether any updates are pending.
     * @type Boolean
     */
    _hasUpdates: true,
    
    _listenerList: null,
    
    /**
     * Associative mapping between component ids and component instances for all
     * updates held in this manager object.
     */
    _idMap: null,
    
    _removedIdMap: null,
    
    /** 
     * The id of the last parent component whose child was analyzed by
     * _isAncestorBeingAdded() that resulted in that method returning false.
     * This id is stored for performance optimization purposes.
     * @type String
     */
    _lastAncestorTestParentId: null,
    
    /**
     * Creates a new Update Manager.
     *
     * @constructor
     * @param {EchoApp.Application} application the supported application
     */
    $construct: function(application) {
        this._componentUpdateMap = { };
        this.application = application;
        this._listenerList = new Core.ListenerList();
        this._idMap = { };
        this._removedIdMap = { };
    },
    
    /**
     * Adds a listener to receive notification of update events.
     * 
     * @param {Function} l the listener to add
     */
    addUpdateListener: function(l) {
        this._listenerList.addListener("update", l);
    },
    
    /**
     * Creates a new ComponentUpdate object (or returns an existing one) for a
     * specific parent component.
     * 
     * @private
     * @param {EchoApp.Component} parent the parent Component
     * @return a ComponentUpdate instance for that Component
     * @type EchoApp.Update.ComponentUpdate 
     */
    _createComponentUpdate: function(parent) {
        this._hasUpdates = true;
        var update = this._componentUpdateMap[parent.renderId];
        if (!update) {
            update = new EchoApp.Update.ComponentUpdate(this, parent);
            this._componentUpdateMap[parent.renderId] = update;
        }
        return update;
    },
    
    /**
     * Permanently disposes of the Update Manager, freeing any resources.
     */
    dispose: function() {
        this.application = null;
    },
    
    /**
     * Notifies update listeners of an event.
     * 
     * @private
     */
    _fireUpdate: function() {
        if (!this._listenerList.isEmpty()) {
            this._listenerList.fireEvent({type: "update", source: this});
        }
    },
    
    /**
     * Returns the current pending updates.  Returns null in the event that that no pending updates exist.
     * 
     * @return an array containing all component updates (as EchoApp.Update.ComponentUpdates)
     * @type Array
     */
    getUpdates: function() {
        var updates = [];
        for (var key in this._componentUpdateMap) {
            updates.push(this._componentUpdateMap[key]);
        }
        return updates;
    },
    
    /**
     * Determines if any updates exist in the Update Manager.
     * 
     * @return true if any updates are present
     * @type Boolean
     */
    hasUpdates: function() {
        return this._hasUpdates;
    },
    
    /**
     * Determines if an ancestor of the specified component is being added.
     * 
     * @private
     * @param {EchoApp.Component} component the component to evaluate
     * @return true if the component or an ancestor of the component is being added
     * @type Boolean
     */
    _isAncestorBeingAdded: function(component) {
        var child = component;
        var parent = component.parent;
        
        var originalParentId = parent ? parent.renderId : null;
        if (originalParentId && this._lastAncestorTestParentId == originalParentId) {
            return false;
        }
        
        while (parent) {
            var update = this._componentUpdateMap[parent.renderId];
            if (update && update._addedChildIds) {
                for (var i = 0; i < update._addedChildIds.length; ++i) {
                    if (update._addedChildIds[i] == child.renderId) {
                        return true;
                    }
                }
            }
            child = parent;
            parent = parent.parent;
        }
        
        this._lastAncestorTestParentId = originalParentId;
        return false;
    },
    
    /**
     * Processes a child addition to a component.
     * 
     * @private
     * @param {EchoApp.Component} parent the parent component
     * @param {EchoApp.Component} child the added child component
     */
    _processComponentAdd: function(parent, child) {
        if (this.fullRefreshRequired) {
            return;
        }
        if (this._isAncestorBeingAdded(child)) {
            return;
        };
        var update = this._createComponentUpdate(parent);
        update._addChild(child);
    },
    
    /**
     * Process a layout data update to a child component.
     * 
     * @private
     * @param {EchoApp.Component} updatedComponent the updated component
     */
    _processComponentLayoutDataUpdate: function(updatedComponent) {
        if (this.fullRefreshRequired) {
            return;
        }
        var parent = updatedComponent.parent;
        if (parent == null || this._isAncestorBeingAdded(parent)) {
            return;
        }
        var update = this._createComponentUpdate(parent);
        update._updateLayoutData(updatedComponent);
    },
    
    /**
     * Processes a child removal from a component.
     * 
     * @private
     * @param {EchoApp.Component} parent the parent component
     * @param {EchoApp.Component} child the removed child component
     */
    _processComponentRemove: function(parent, child) {
        if (this.fullRefreshRequired) {
            return;
        }
        if (this._isAncestorBeingAdded(parent)) {
            return;
        }
        var update = this._createComponentUpdate(parent);
        update._removeChild(child);
        
        var disposedIds = null;
        
        // Search updated components for descendants of removed component.
        // Any found descendants will be removed and added to this update's
        // list of removed components.
        for (var testParentId in this._componentUpdateMap) {
             var testUpdate = this._componentUpdateMap[testParentId];
             if (child.isAncestorOf(testUpdate.parent)) {
                 update._appendRemovedDescendants(testUpdate);
                 if (disposedIds == null) {
                     disposedIds = [];
                 }
                 disposedIds.push(testParentId);
             }
        }
        
        if (disposedIds != null) {
            for (var i = 0; i < disposedIds.length; ++i) {
                delete this._componentUpdateMap[disposedIds[i]];
            }
        }
    },
    
    /**
     * Processes a property update to a component.
     * 
     * @private
     * @component {EchoApp.Component} the updated component
     * @propertyName {String} the updated property name
     * @oldValue the previous value of the property
     * @newValue the new value of the property
     */
    _processComponentPropertyUpdate: function(component, propertyName, oldValue, newValue) {
        if (this.fullRefreshRequired) {
            return;
        }
        if (this._isAncestorBeingAdded(component)) {
            return;
        }
        var update = this._createComponentUpdate(component);
        update._updateProperty(propertyName, oldValue, newValue);
    },
    
    /**
     * Processes component updates received from the application instance.
     */
    _processComponentUpdate: function(parent, propertyName, oldValue, newValue) {
        if (propertyName == "children") {
            if (newValue == null) {
                this._processComponentRemove(parent, oldValue);
            } else {
                this._processComponentAdd(parent, newValue);
            }
        } else if (propertyName == "layoutData") {
            this._processComponentLayoutDataUpdate(parent, oldValue, newValue);
        } else {
            this._processComponentPropertyUpdate(parent, propertyName, oldValue, newValue);
        }
        this._fireUpdate();
    },
    
    /**
     * Purges all updates from the manager.
     * Invoked after the client has repainted the screen.
     */
    purge: function() {
        this.fullRefreshRequired = false;
        this._componentUpdateMap = { };
        this._idMap = { };
        this._removedIdMap = { };
        this._hasUpdates = false;
        this._lastAncestorTestParentId = null;
    },
    
    /**
     * Removes a listener from receiving notification of update events.
     * 
     * @param {Function} l the listener to remove
     */
    removeUpdateListener: function(l) {
        this._listenerList.removeListener("update", l);
    },
    
    /**
     * Returns a string representation.
     * 
     * @return a string representation
     * @type String
     */
    toString: function() {
        var s = "[ UpdateManager ]\n";
        if (this.fullRefreshRequired) {
            s += "fullRefresh";
        } else {
            for (var key in this._componentUpdateMap) {
                s += this._componentUpdateMap[key];
            }
        }
        return s;
    }
});

// Built-in Component Object Definitions

EchoApp.AbstractButton = Core.extend(EchoApp.Component, {

    $abstract: true,
    
    $load: function() {
        EchoApp.ComponentFactory.registerType("AbstractButton", this);
        EchoApp.ComponentFactory.registerType("AB", this);
    },

    componentType: "AbstractButton",
    focusable: true,
    
    $virtual: {
        
        /**
         * Programatically performs a button action.
         */
        doAction: function() {
            this.fireEvent({type: "action", source: this, actionCommand: this.get("actionCommand")});
        }
    }
});

/**
 * @class Button component.
 * @base EchoApp.Component
 */ 
EchoApp.Button = Core.extend(EchoApp.AbstractButton, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("Button", this);
        EchoApp.ComponentFactory.registerType("B", this);
    },

    componentType: "Button"
});

/**
 * @class ToggleButton component.
 * @base EchoApp.Button
 */
EchoApp.ToggleButton = Core.extend(EchoApp.AbstractButton, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("ToggleButton", this);
        EchoApp.ComponentFactory.registerType("TB", this);
    },

    $abstract: true,
    componentType: "ToggleButton"
});

/**
 * @class CheckBox component.
 * @base EchoApp.ToggleButton
 */
EchoApp.CheckBox = Core.extend(EchoApp.ToggleButton, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("CheckBox", this);
        EchoApp.ComponentFactory.registerType("CB", this);
    },

    componentType: "CheckBox"
});

/**
 * @class RadioButton component.
 * @base EchoApp.ToggleButton
 */
EchoApp.RadioButton = Core.extend(EchoApp.ToggleButton, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("RadioButton", this);
        EchoApp.ComponentFactory.registerType("RB", this);
    },

    componentType: "RadioButton"
});

/**
 * @class AbstractListComponent base component.
 * @base EchoApp.Component
 */
EchoApp.AbstractListComponent = Core.extend(EchoApp.Component, {

    $abstract: true,

    $load: function() {
        EchoApp.ComponentFactory.registerType("AbstractListComponent", this);
        EchoApp.ComponentFactory.registerType("LC", this);
    },

    componentType: "AbstractListComponent",
    focusable: true,
    
    $virtual: {
        
        /**
         * Programatically performs a list select action.
         */
        doAction: function() {
            this.fireEvent({type: "action", source: this, actionCommand: this.get("actionCommand")});
        }
    }
});

/**
 * @class ListBox component.
 * @base EchoApp.AbstractListComponent
 */
EchoApp.ListBox = Core.extend(EchoApp.AbstractListComponent, {

    $static: {

        /**
         * Constant for "selectionMode" property indicating single selection.
         */
        SINGLE_SELECTION: 0,
        
        /**
         * Constant for "selectionMode" property indicating multiple selection.
         */
        MULTIPLE_SELECTION: 2
    },

    $load: function() {
        EchoApp.ComponentFactory.registerType("ListBox", this);
        EchoApp.ComponentFactory.registerType("LB", this);
    },

    componentType: "ListBox"
});

/**
 * @class SelectField component.
 * @base EchoApp.AbstractListComponent
 */
EchoApp.SelectField = Core.extend(EchoApp.AbstractListComponent, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("SelectField", this);
        EchoApp.ComponentFactory.registerType("SF", this);
    },

    componentType: "SelectField"
});

/**
 * @class Column component.
 * @base EchoApp.Component
 */
EchoApp.Column = Core.extend(EchoApp.Component, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("Column", this);
        EchoApp.ComponentFactory.registerType("C", this);
    },

    componentType: "Column"
});

/**
 * @class Composite component.
 * @base EchoApp.Component
 */
EchoApp.Composite = Core.extend(EchoApp.Component, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("Composite", this);
        EchoApp.ComponentFactory.registerType("CM", this);
    },

    componentType: "Composite"
});

/**
 * @class Panel component.
 * @base EchoApp.Composite
 */
EchoApp.Panel = Core.extend(EchoApp.Composite, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("Panel", this);
        EchoApp.ComponentFactory.registerType("P", this);
    },

    componentType: "Panel"
});

/**
 * @class ContentPane component.
 * @base EchoApp.Component
 */
EchoApp.ContentPane = Core.extend(EchoApp.Component, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("ContentPane", this);
        EchoApp.ComponentFactory.registerType("CP", this);
    },

    componentType: "ContentPane",
    pane: true
});

/**
 * @class Grid component.
 * @base EchoApp.Component
 */
EchoApp.Grid = Core.extend(EchoApp.Component, {

    $static: {

        /**
         * Constant value for <code>orientation</code> property indicating cells 
         * should be laid out horizontally and then vertically.
         * <code>ORIENTATION_HORIZONTAL</code> is the default orientation setting.
         */
        ORIENTATION_HORIZONTAL: 0,
    
        /**
         * Constant value for <code>orientation</code> property indicating cells 
         * should be laid out vertically and then horizontally. 
         */
        ORIENTATION_VERTICAL: 1,

        /**
         * A constant value for the <code>columnSpan</code> and <code>rowSpan</code>
         * properties of <code>LayoutData</code> objects used by children of a
         * Grid indicating that a cell should fill all remaining cells.  
         * <p>
         * <strong>WARNING</strong>: This value may ONLY be used for spans in the
         * direction of the layout of the <code>Grid</code>, i.e., it may only be 
         * used for column-spans if the orientation is horizontal, and it may only
         * be used for row-spans if the orientation is vertical.
         */
        SPAN_FILL: -1
    },

    $load: function() {
        EchoApp.ComponentFactory.registerType("Grid", this);
        EchoApp.ComponentFactory.registerType("G", this);
    },

    componentType: "Grid"
});

/**
 * @class Label component.
 * @base EchoApp.Component
 */
EchoApp.Label = Core.extend(EchoApp.Component, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("Label", this);
        EchoApp.ComponentFactory.registerType("L", this);
    },

    componentType: "Label"
});

/**
 * @class Row component.
 * @base EchoApp.Component
 */
EchoApp.Row = Core.extend(EchoApp.Component, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("Row", this);
        EchoApp.ComponentFactory.registerType("R", this);
    },

    componentType: "Row"
});

/**
 * @class Abstract base class for text components.
 * @base EchoApp.Component
 */
EchoApp.TextComponent = Core.extend(EchoApp.Component, {

    $abstract: true,

    $load: function() {
        EchoApp.ComponentFactory.registerType("TextComponent", this);
        EchoApp.ComponentFactory.registerType("TC", this);
    },

    componentType: "TextComponent",
    focusable: true
});

/**
 * @class TextArea component.
 * @base EchoApp.Component
 */
EchoApp.TextArea = Core.extend(EchoApp.TextComponent, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("TextArea", this);
        EchoApp.ComponentFactory.registerType("TA", this);
    },

    componentType: "TextArea"
});

/**
 * @class TextField component.
 * @base EchoApp.Component
 */
EchoApp.TextField = Core.extend(EchoApp.TextComponent, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("TextField", this);
        EchoApp.ComponentFactory.registerType("TF", this);
    },

    componentType: "TextField"
});

/**
 * @class PasswordField component.
 * @base EchoApp.Component
 */
EchoApp.PasswordField = Core.extend(EchoApp.TextField, {

    $load: function() {
        EchoApp.ComponentFactory.registerType("PasswordField", this);
        EchoApp.ComponentFactory.registerType("PF", this);
    },
    
    componentType: "PasswordField"
});

/**
 * @fileoverview
 * Module for rendering state of application to DOM.
 * <ul>
 *  <li>Provides capability to process updates in Application UpdateManager,
 *   rendering state changes to the DOM.</li>
 *  <li>Provides component synchronization peer base class.</li>
 *  <li>Provides root component synchronization peer implementation.</li>
 *  <li>Provides rendering utilities for the core properties.</li>
 * </ul>
 */

/**
 * @class Application rendering namespace.  Non-instantiable object.
 */
EchoRender = {

    /**
     * Mapping between component type names and instantiable peer classes.
     * 
     * @type Object
     * @private
     */
    _peers: {},
    
    /**
     * Map containing removed components.  Maps component ids to removed components.
     * Created and destroyed during each render.
     * 
     * @type Object
     * @private
     */
    _disposedComponents: null,
    
    //FIXME.  Scrollbar position tracking code in SplitPane appears to suggest that
    // disposed states are not in good shape....SplitPane is being disposed when
    // parent contentPane is redrawn.
    
    /**
     * An array sorting implemention to organize an array by component depth.
     * @private
     */
    _componentDepthArraySort: function(a, b) {
        return EchoRender._getComponentDepth(a.parent) - EchoRender._getComponentDepth(b.parent);
    },
    
    /**
     * Recursively invokes renderDisplay() method on a sub-hierarchy of the
     * component hierarchy.  If a peer does not provide a renderDisplay() implementation,
     * it is skipped (although its descendants will NOT be skipped).
     * 
     * @param the root component of the sub-hierarchy on which renderDisplay() should be invoked
     * @param includeSelf flag indicating whether renderDisplay() should be invoked on the
     *        specified component (if false, it will only be invoked on child components)
     * @private
     */
    _doRenderDisplay: function(component, includeSelf) {
        if (includeSelf) {
            EchoRender._doRenderDisplayImpl(component);
        } else {
            for (var i = 0; i < component.children.length; ++i) {
                EchoRender._doRenderDisplayImpl(component.children[i]);
            }
        }
    },
    
    /**
     * Recursive work method for _doRenderDisplay().  
     * 
     * @param component the component on which to invoke renderDisplay()
     * @private
     */
    _doRenderDisplayImpl: function(component) {
        if (component.peer) {
            // components that are present on the client, but are not rendered (lazy rendered as in tree), 
            // have no peer installed.
            if (component.peer.renderDisplay) {
                component.peer.renderDisplay();
            }
            
            for (var i = 0; i < component.children.length; ++i) {
                EchoRender._doRenderDisplayImpl(component.children[i]);
            }
        }
    },
    
    /**
     * Returns the depth of a specific component in the hierarchy.
     * The root component is at depth 0, its immediate children are
     * at depth 1, their children are at depth 2, and so on.
     *
     * @param component the component whose depth is to be calculated
     * @return the depth of the component
     * @private
     */
    _getComponentDepth: function(component) {
        var depth = -1;
        while (component != null) {
            component = component.parent;
            ++depth;
        }
        return depth;
    },
    
    getPeerClass: function(component) {
        return EchoRender._peers[component.componentType];
    },
    
    /**
     * Creates a component synchronization peer for a component.
     * The peer will be stored in the "peer" property of the component.
     * The client will be stored in the "client" property of the component.
     * 
     * @param {EchoClient} client the relevant Client
     * @param {EchoApp.Component} component the component
     * @private
     */
    _loadPeer: function(client, component) {
        if (component.peer) {
            return;
    // FIXME. which behavior is correct for this scenario: ignore or fail?    
    //        throw new Error("Peer already installed: " + component);
        }
        
        var peerClass = EchoRender._peers[component.componentType];
        
        if (!peerClass) {
            throw new Error("Peer not found for: " + component.componentType);
        }
        
        component.peer = new peerClass();
        component.peer.component = component;
        component.peer.client = client;
    },
    
    /**
     * Notifies child components that the parent component has been drawn
     * or resized.  At this point the parent component is on the screen
     * (the parent element is part of the DOM hierarchy).
     * Child components (and their descendants) will be notified by having 
     * their renderDisplay() implementations invoked.
     * Note that the parent WILL NOT have its renderDisplay() method
     * invoked.
     * <p>
     * If your component requires virtual positioning (for IE6) you should invoke
     * this method after informing the virtual positioning system to recalculate
     * the size of your component.
     * 
     * @param {EchoApp.Component} parent the component whose size changed
     */
    notifyResize: function(parent) {
        EchoRender._doRenderDisplay(parent, false);
    },
    
    /**
     * Invokes renderDispose() on all removed children and descendants found in the specified update.
     * 
     * @param {EchoApp.Update.ComponentUpdate} update the update
     * @private
     */
    _processDispose: function(update) {
        var components = update.getRemovedDescendants();
        if (components) {
            for (var i = 0; i < components.length; ++i) {
                EchoRender._renderComponentDisposeImpl(update, components[i]);
            }
        }
        components = update.getRemovedChildren();
        if (components) {
            for (var i = 0; i < components.length; ++i) {
                EchoRender._renderComponentDisposeImpl(update, components[i]);
            }
        }
    },
    
    /**
     * Processes all pending updates in the client's application's update manager.
     * 
     * @param {EchoClient} client the client
     */
    processUpdates: function(client) {
        var updateManager = client.application.updateManager;
        
        // Do nothing if no updates exist.
        if (!updateManager.hasUpdates()) {
            return;
        }
        
        // Create map to contain removed components (for peer unloading).
        EchoRender._disposedComponents = {};
        
        // Retrieve updates, sorting by depth in hierarchy.  This will ensure that higher
        // level updates have a chance to execute first, in case they null out lower-level
        // updates if they require re-rendering their descendants.
        var updates = updateManager.getUpdates();
        updates.sort(EchoRender._componentDepthArraySort);
    
        // Load peers for any root components being updated.
        for (var i = 0; i < updates.length; ++i) {
            updates[i].renderContext = {};
        
            var peers = updates[i].parent.peer;
            if (peer == null && updates[i].parent.componentType == "Root") {
                EchoRender._loadPeer(client, updates[i].parent);
            }
        }
    
        // Remove Phase: Invoke renderDispose on all updates.
        for (var i = updates.length - 1; i >= 0; --i) {
            if (updates[i] == null) {
                // Skip removed updates.
                continue;
            }
            var peer = updates[i].parent.peer;
            EchoRender._processDispose(updates[i]);
        }
        
        // Profiling: Mark completion of remove phase. 
        if (EchoClient.profilingTimer) {
            EchoClient.profilingTimer.mark("rem");
        }
        
        // Update Phase: Invoke renderUpdate on all updates.
        for (var i = 0; i < updates.length; ++i) {
            if (updates[i] == null) {
                // The update has been removed, skip it.
                continue;
            }
            
            // Obtain component synchronization peer.
            var peer = updates[i].parent.peer;
            
            // Perform update by invoking peer's renderUpdate() method.
            var fullRender = peer.renderUpdate(updates[i]);
            
            // If the update required re-rendering descendants of the updated component,
            // null-out any pending updates to descandant components.
            if (fullRender) {
                for (var j = i + 1; j < updates.length; ++j) {
                    if (updates[j] != null && updates[i].parent.isAncestorOf(updates[j].parent)) {
                        updates[j] = null;
                    }
                }
            }
    
            //FIXME ....moved after loop, ensure this is okay (evaluate use of dispose).
            // Set disposed set of peer to false.
            EchoRender._setPeerDisposedState(updates[i].parent, false);
        }
        
        // Profiling: Mark completion of update phase.
        if (EchoClient.profilingTimer) {
            EchoClient.profilingTimer.mark("up");
        }
        
        // Display Phase: Invoke renderDisplay on all updates.
        for (var i = 0; i < updates.length; ++i) {
            if (updates[i] == null) {
                // Skip removed updates.
                continue;
            }
            //FIXME. this does needless work....resizing twice is quite possible.
            // if property updates are present.
            if (updates[i].renderContext.displayRequired) {
                for (var j = 0; j < updates[i].renderContext.displayRequired.length; ++j) {
                    Core.Debug.consoleWrite("PartialRenderUpdate:" + updates[i].renderContext.displayRequired[j]);
                    EchoRender._doRenderDisplay(updates[i].renderContext.displayRequired[j], true);
                }
            } else {
                EchoRender._doRenderDisplay(updates[i].parent, true);
            }
        }
    
        // Profiling: Mark completion of display phase.
        if (EchoClient.profilingTimer) {
            EchoClient.profilingTimer.mark("disp");
        }
    
        // Unload peers for truly removed components, destroy mapping.
        for (var componentId in EchoRender._disposedComponents) {
            var component = EchoRender._disposedComponents[componentId];
            EchoRender._unloadPeer(component);
        }
        EchoRender._disposedComponents = null;
        
        // Inform UpdateManager that all updates have been completed.
        updateManager.purge();
        
        // Perform focus update if required.
        if (client.focusUpdateRequired) {
            var focusedComponent = client.application.getFocusedComponent();
            if (focusedComponent && focusedComponent.peer && focusedComponent.peer.renderFocus) {
                focusedComponent.peer.renderFocus();
                client.focusUpdateRequired = false;
            }
        }
    },
    
    /**
     * Registers a component type name with an instantiable peer class.
     * Components of the specified type name will be assigned new instasnces of the peer class
     * when rendered for the first time.
     * 
     * @param {String} componentName the component type name
     * @param {Function} peerObject the peer class object
     */
    registerPeer: function(componentName, peerObject) {
        if (this._peers[componentName]) {
            throw new Error("Peer already registered: " + componentName);
        }
        this._peers[componentName] = peerObject;
    },
    
    /**
     * Renders a new component inside of a DOM element.
     * This method should be called by container components in order to render their children.
     * 
     * @param {EchoApp.Update.ComponentUpdate} update the revelant ComponentUpdate
     * @param {EchoApp.Component} component the component to add
     * @param {Element} parentElement the DOM element to which the rendered component should be added
     */
    renderComponentAdd: function(update, component, parentElement) {
        if (!component.parent || !component.parent.peer || !component.parent.peer.client) {
            throw new Error("Cannot find reference to the Client with which this component should be associated: "
                    + "cannot load peer.  This is due to the component's parent's peer not being associated with a Client. "
                    + "Component = " + component);
        }
    
        EchoRender._loadPeer(component.parent.peer.client, component);
        EchoRender._setPeerDisposedState(component, false);
        component.peer.renderAdd(update, parentElement);
    },
    
    /**
     * Manually invokes renderDisplay on a component (and its descendants) that was added to the
     * hierarchy outside of processUpdates().  This method is only used in special cases,
     * e.g., by in the case of Application Rendered Components that need to render children.
     * 
     * @param parent the parent component of the sub-hierarchy on which renderDisplay() should
     *        be invoked (note that renderDisplay WILL be invoked on the parent as well 
     *        as its descendants)
     */
    renderComponentDisplay: function(parent) {
        this._doRenderDisplay(parent, true);
    },
    
    /**
     * Disposes of a component and its descendants.
     * This method should be invoked by any peer that will be updating a component in such
     * a fashion that it will be destroying the rendering of its children and re-rendering them.
     * It is not necessary to invoke this method on components that may not contain children.
     *
     * @param update the <code>ComponentUpdate</code> for which this change is being performed
     * @param component the <code>Component</code> to be disposed
     */
    renderComponentDispose: function(update, component) {
        this._renderComponentDisposeImpl(update, component);
    },
    
    /**
     * Recursive implementation of renderComponentDispose.  Invokes
     * renderDispose() on all child peers, sets disposed state on each.
     * 
     * @param update the <code>ComponentUpdate</code> for which this change is being performed
     * @param component the <code>Component</code> to be disposed
     * @private
     */
    _renderComponentDisposeImpl: function(update, component) {
        if (!component.peer || component.peer.disposed) {
            return;
        }
        EchoRender._setPeerDisposedState(component, true);
    
        component.peer.renderDispose(update);
        for (var i = 0; i < component.children.length; ++i) {
            EchoRender._renderComponentDisposeImpl(update, component.children[i]);
        }
    },
    
    /**
     * Sets the peer disposed state of a component.
     * The peer disposed state indicates whether the renderDispose()
     * method of the component has been executed since it was last rendered.
     * 
     * @param {EchoApp.Component} component the component
     * @param {Boolean} disposed the disposed state, true indicating the component has
     *        been disposed
     * @private
     */
    _setPeerDisposedState: function(component, disposed) {
        if (disposed) {
            component.peer.disposed = true;
            EchoRender._disposedComponents[component.renderId] = component;
        } else {
            component.peer.disposed = false;
            delete EchoRender._disposedComponents[component.renderId];
        }
    },
    
    // FIXME. Ensure this is properly invoked and no peers are being leaked.
    /**
     * Destroys a component synchronization peer for a specific compoennt.
     * The peer will be removed from the "peer" property of the component.
     * The client will be removed from the "client" property of the component.
     * The peer to component association will be removed.
     * 
     * @param {EchoApp.Component} component the component
     * @private
     */
    _unloadPeer: function(component) {
        component.peer.client = null;
        component.peer.component = null;
        component.peer = null;
    }
};

/**
 * @class
 * Component synchronization peer. 
 */
EchoRender.ComponentSync = Core.extend({ 

    $static: {
        FOCUS_PERMIT_ARROW_UP: 0x1,
        FOCUS_PERMIT_ARROW_DOWN: 0x2, 
        FOCUS_PERMIT_ARROW_LEFT: 0x4,
        FOCUS_PERMIT_ARROW_RIGHT: 0x8, 
        FOCUS_PERMIT_ARROW_ALL: 0xf
    },

    /**
     * The client supported by this peer.
     * @type EchoClient
     */
    client: null,

    /**
     * The component instance supported by this peer.  
     * Each peer instance will support a single component instance.
     * @type EchoApp.Component
     */
    component: null,

    /**
     * Creates a new copmonent synchronization peer.
     * @constructor
     */
    $construct: function() { },
    
    $abstract: {

        /**
         * Renders the component to the DOM.
         * The supplied update will refer to a ancestor component of the supported component
         * being updated.
         *
         * @param {EchoApp.Update.ComponentUpdate} update the update being rendered
         */
        renderAdd: function(update, parentElement) {
            throw new Error("Operation \"renderAdd\" not supported (Component: " + this.component + ").");
        },

        /**
         * Invoked when the rendered component is about to be removed from the DOM.
         * This method should dispose of any client resources in use by the component, e.g.,
         * unregistering event listeners and removing any DOM elements that are not children of
         * the parent component's DOM element.
         * The DOM should NOT be modified to remove the element(s) representing this component
         * for performance as well as aesthetic reasons (e.g., in the case where a parent component
         * might be using an animated transition effect to remove the component.
         * The supplied update will refer to a ancestor component of the supported component
         * being updated.
         *
         * A component may be re-added to the screen after being disposed, e.g., in the case
         * where a parent component does not possess a 'partial update' capability and removes
         * a child component hierarchy and then re-renders it.  A synchronization peer should
         * allow for the fact that its renderAdd() method may be invoked at some point in time
         * after renderDispose() has been invoked.
         *        
         * @param {EchoApp.Update.ComponentUpdate} update the update being rendered
         */
        renderDispose: function(update) {
            throw new Error("Operation \"renderDispose\" not supported (Component: " + this.component + ").");
        },
        
        /**
         * Renders an update to a component, e.g., children added/removed, properties updated.
         * The supplied update will refer specifically to an update of the supported component.
         *
         * @param {EchoApp.Update.ComponentUpdate} update the update being rendered
         * @return true if this invocation has re-rendered all child components, false otherwise
         */
        renderUpdate: function(update) {
            throw new Error("Operation \"renderUpdate\" not supported (Component: " + this.component + ").");
        }
    },
    
    $virtual: {
    
        getFocusFlags: null,
        
        /**
         * Invoked when component is rendered focused.
         */
        renderFocus: null,
        
        /**
         * Optional method.  Invoked when the component has been added to the hierarchy and first appears
         * on screen, and when ancestors of the component (or the containing window) have
         * resized.         
         */
        renderDisplay: null
    }
});

/**
 * Creates a new root component synchronization peer.
 * The root component is not managed by the server, but rather is an existing
 * element within which the Echo application is rendered.
 * This is a very special case in that there is no renderAdd() method.
 * 
 * @constructor
 * @class Root component synchronization peer.
 */
EchoRender.RootSync = Core.extend(EchoRender.ComponentSync, { 

    renderAdd: function(update, parentElement) {
        throw new Error("Unsupported operation: renderAdd().");
    },
    
    renderDispose: function(update) { },
    
    renderUpdate: function(update) {
        var fullRender = false;
        if (update.hasAddedChildren() || update.hasRemovedChildren()) {
            WebCore.DOM.removeAllChildren(this.client.domainElement);
            for (var i = 0; i < update.parent.children.length; ++i) {
                EchoRender.renderComponentAdd(update, update.parent.children[i], this.client.domainElement);
            }
            fullRender = true;
        }
        
        if (update.hasUpdatedProperties()) {
            var titleUpdate = update.getUpdatedProperty("title");
            if (titleUpdate) {
                document.title = titleUpdate.newValue;
            }
        }
        
        return fullRender;
    }
});

/**
 * @class
 * Namespace for utility objects.
 */
EchoRender.Util = {
    
    // FIXME abstract this somehow so it works with FreeClient too
    TRANSPARENT_IMAGE: "?sid=Echo.TransparentImage"
};

EchoRender.registerPeer("Root", EchoRender.RootSync);
/**
 * @fileoverview
 * <ul> 
 *  <li>Provides core property renderers.</li>
 *  <li>Provides property rendering utilities.</li>
 *  <li>Provides TriCellTable rendering utility (used by buttons and labels).</li>
 *  <li>Provides a floating pane z-index management system.</li> 
 * </ul>
 */

EchoAppRender = { 

    getEffectProperty: function(component, defaultPropertyName, effectPropertyName, effectState,
            defaultDefaultPropertyValue, effectDefaultPropertyValue) {
        var property;
        if (effectState) {
            property = component.render(effectPropertyName, effectDefaultPropertyValue);
        }
        if (!property) {
            property = component.render(defaultPropertyName, defaultDefaultPropertyValue);
        }
        return property;
    }
};

EchoAppRender.Alignment = { 

    _HORIZONTALS: { left: true, center: true, right: true, leading: true, trailing: true },
    _VERTICALS: { top: true, middle: true, bottom: true },

    getRenderedHorizontal: function(alignment, layoutDirectionProvider) {
        if (alignment == null) {
            return null;
        }
    
        var layoutDirection = layoutDirectionProvider ? 
                layoutDirectionProvider.getRenderLayoutDirection() : EchoApp.LayoutDirection.LTR;
         
        var horizontal = typeof(alignment) == "object" ? alignment.horizontal : alignment; 
                
        switch (horizontal) {
        case "leading":
            return layoutDirection.isLeftToRight() ? "left" : "right";
        case "trailing":
            return layoutDirection.isLeftToRight() ? "right" : "left";
        default:
            return horizontal in this._HORIZONTALS ? horizontal : null;
        }
    },
    
    getHorizontal: function(alignment) {
        if (alignment == null) {
            return null;
        }
        if (typeof(alignment == "string")) {
            return alignment in this._HORIZONTALS ? alignment : null;
        } else {
            return alignment.horizontal;
        }
    },

    getVertical: function(alignment) {
        if (alignment == null) {
            return null;
        }
        if (typeof(alignment == "string")) {
            return alignment in this._VERTICALS ? alignment : null;
        } else {
            return alignment.vertical;
        }
    },

    render: function(alignment, element, renderToElement, layoutDirectionProvider) {
        if (alignment == null) {
            return;
        }
        
        var horizontal = EchoAppRender.Alignment.getRenderedHorizontal(alignment, layoutDirectionProvider);
        var vertical = typeof(alignment) == "object" ? alignment.vertical : alignment;
    
        var horizontalValue;
        switch (horizontal) {
        case "left":   horizontalValue = "left";   break;
        case "center": horizontalValue = "center"; break;
        case "right":  horizontalValue = "right";  break;
        default:       horizontalValue = "";       break;
        }
        var verticalValue;
        switch (vertical) {
        case "top":    verticalValue = "top";      break;
        case "middle": verticalValue = "middle";   break;
        case "bottom": verticalValue = "bottom";   break;
        default:       verticalValue = "";         break;
        }
        
        if (renderToElement) {
            element.align = horizontalValue;
            element.vAlign = verticalValue;
        } else {
            element.style.textAlign = horizontalValue;
            element.style.verticalAlign = verticalValue;
        }
    }
};

EchoAppRender.Border = {

    /**
     * Regular expression to validate/parse a CSS border expression, e.g., "1px solid #abcdef".
     * Supports omission of any term, or empty strings.
     */
    _PARSER_PX: new RegExp("^(-?\\d+px)?(?:^|$|(?= )) ?(none|hidden|dotted|dashed|solid|" 
            + "double|groove|ridge|inset|outset)?(?:^|$|(?= )) ?(#[0-9a-fA-F]{6})?$"),

    /**
     * Regular expression to validate/parse a pixel-based CSS border expression, e.g., "1px solid #abcdef".
     * Supports omission of any term, or empty strings.
     */
    _PARSER: new RegExp("^(-?\\d+(?:px|pt|pc|cm|mm|in|em|ex))?(?:^|$|(?= )) ?(none|hidden|dotted|dashed|solid|"
            + "double|groove|ridge|inset|outset)?(?:^|$|(?= )) ?(#[0-9a-fA-F]{6})?$"),
            
    _TEST_EXTENT_PX: /^(-?\d+px*)$/,
    
    compose: function(size, style, color) {
        if (typeof size == "number") {
            size += "px";
        }
        var out = [];
        if (size) {
            out.push(size);
        }
        if (style) {
            out.push(style);
        }
        if (color) {
            out.push(color);
        }
        return out.join(" ");
    },
    
    parse: function(border) {
        if (!border) {
            return { };
        }
        if (typeof(border) == "string") {
            var parts = this._PARSER.exec(border);
            return { size: parts[1], style: parts[2], color: parts[3] } 
        } else {
            // FIXME support multisided borders.
            return { }
        }
    },

    render: function(border, element, styleName) {
        if (!border) {
            return;
        }
        styleName = styleName ? styleName : "border";
        if (typeof(border) == "string") {
            if (this._PARSER_PX.test(border)) {
                element.style[styleName] = border;
            } else {
                var elements = this._PARSER.exec(border);
                if (elements == null) {
                    throw new Error("Invalid border: \"" + border + "\"");
                }
            }
        } else {
            // FIXME. impl multisided rendering.
            if (border.top) {
                if (border.right) {
                    if (border.bottom) {
                        if (border.left) {
                        }
                    }
                }
            }
        }
    },
    
    renderClear: function(border, element) {
        if (border) {
            this.render(border, element);
        } else {
            element.style.border = "";
        }
    },

    getPixelSize: function(border, sideName) {
        if (!border) {
            return 0;
        }
        
        if (typeof(border) == "string") {
            var extent = this._PARSER.exec(border)[1];
            if (extent == null) {
                return 0;
            } else if (this._TEST_EXTENT_PX.test(extent)) {
                return parseInt(extent);
            } else {
                return EchoAppRender.Extent.toPixels(extent);
            }
        } else if (typeof(border) == "object") {
            // Retrieve value for indivudal side.
            // Specified side is queried first, followed by alternatives.
            while (true) {
                var side = this.getPixelSize(border[sideName]);
                if (side == null) {
                    switch (sideName) {
                    case "left": 
                        // If left side specified but value null, try again with right.
                        sideName = "right"; 
                        continue;
                    case "right":
                    case "bottom": 
                        // If bottom or right side specified, try again with top.
                        sideName = "top";
                        continue; 
                    }
                }
                return side;
            }
        }
    }
};

EchoAppRender.Color = {

    /**
     * Adjusts the value of the color's RGB values by the
     * specified amounts, returning a new Color.
     * The original color is unchanged.
     * 
     * @param color the color to adjust (a 24 bit hex value, e.g., #1a2b3c)
     * @param r the amount to adjust the red value of the color (-255 to 255)
     * @param g the amount to adjust the green value of the color (-255 to 255)
     * @param b the amount to adjust the blue value of the color (-255 to 255)
     * @return the adjusted color (a 24 bit hex value)
     */
    adjust: function(value, r, g, b) {
        var colorInt = parseInt(value.substring(1), 16);
        var red = parseInt(colorInt / 0x10000) + r;
        if (red < 0) {
            red = 0;
        } else if (red > 255) {
            red = 255;
        }
        var green = parseInt(colorInt / 0x100) % 0x100 + g;
        if (green < 0) {
            green = 0;
        } else if (green > 255) {
            green = 255;
        }
        var blue = colorInt % 0x100 + b;
        if (blue < 0) {
            blue = 0;
        } else if (blue > 255) {
            blue = 255;
        }
        return "#" + (red < 16 ? "0" : "") + red.toString(16)
                + (green < 16 ? "0" : "") + green.toString(16)
                + (blue < 16 ? "0" : "") + blue.toString(16); 
    },

    render: function(color, element, styleProperty) {
        if (color) {
            element.style[styleProperty] = color;
        }
    },
    
    renderClear: function(color, element, styleProperty) {
        element.style[styleProperty] = color ? color : "";
    },
    
    renderFB: function(component, element) { 
        var color;
        if (color = component.render("foreground")) {
            element.style.color = color;
        }
        if (color = component.render("background")) {
            element.style.backgroundColor = color;
        }
    }
};

EchoAppRender.Extent = { 

    /**
     * Regular expression to parse an extent value, e.g., "12px" into its value and unit components.
     */
    _PARSER: /^(-?\d+(?:\.\d+)?)(.+)?$/,

    _FORMATTED_PIXEL_TEST: /^(-?\d+px *)$/,
    
    isPercent: function(extent) {
        if (extent == null || typeof(extent) == "number") {
            return false;
        } else {
            var parts = this._PARSER.exec(arguments[0]);
            if (!parts) {
                throw new Error("Invalid Extent: " + arguments[0]);
            }
            return parts[2] == "%";
        }
    },

    toCssValue: function(extent, horizontal) {
        switch(typeof(extent)) {
            case "number":
                return extent + "px";
                break;
            case "string":
                if (this._FORMATTED_PIXEL_TEST.test(extent)) {
                    return extent;
                } else {
                    return this.toPixels(extent, horizontal) + "px";
                }
                break;
        }
        return "";
    },

    toPixels: function(extent, horizontal) {
        if (extent == null) {
            return 0;
        } else if (typeof(extent) == "number") {
            return extent;
        } else {
            var parts = this._PARSER.exec(extent);
            if (!parts) {
                throw new Error("Invalid Extent: " + extent);
            }
            var value = parseFloat(parts[1]);
            var units = parts[2] ? parts[2] : "px";
            return WebCore.Measure.extentToPixels(value, units, horizontal);
        }
    }
};

EchoAppRender.FillImage = { 

    _REPEAT_VALUES: {
        "0": "no-repeat",
        "x": "repeat-x",
        "y": "repeat-y",
        "xy": "repeat",
        "no-repeat": "no-repeat",
        "repeat-x": "repeat-x",
        "repeat-y": "repeat-y",
        "repeat": "repeat"
    },

    FLAG_ENABLE_IE_PNG_ALPHA_FILTER: 0x1,
    
    render: function(fillImage, element, flags) {
        if (fillImage == null) {
            // No image specified, do nothing.
            return;
        }
        
        var isObject = typeof(fillImage) == "object";
        var url = isObject ? fillImage.url : fillImage;

        if (WebCore.Environment.PROPRIETARY_IE_PNG_ALPHA_FILTER_REQUIRED &&
                flags && (flags & this.FLAG_ENABLE_IE_PNG_ALPHA_FILTER)) {
            // IE6 PNG workaround required.
            element.style.filter = "progid:DXImageTransform.Microsoft.AlphaImageLoader(src='" 
                + url + "', sizingMethod='scale')";
        } else {
            // IE6 PNG workaround not required.
            element.style.backgroundImage = "url(" + url + ")";
        }
        
        if (isObject) {
            if (this._REPEAT_VALUES[fillImage.repeat]) {
                element.style.backgroundRepeat = this._REPEAT_VALUES[fillImage.repeat]; 
            }
            
            if (fillImage.x || fillImage.y) {
                var x, y;
                if (EchoAppRender.Extent.isPercent(fillImage.x)) {
                    x = fillImage.x;
                } else {
                    x = EchoAppRender.Extent.toPixels(fillImage.x, true) + "px";
                }
                if (EchoAppRender.Extent.isPercent(fillImage.y)) {
                    y = fillImage.y;
                } else {
                    y = EchoAppRender.Extent.toPixels(fillImage.y, false) + "px";
                }
                element.style.backgroundPosition = x + " " + y;
            }
        }
        
    },
    
    renderClear: function(fillImage, element, flags) {
        if (fillImage) {
            this.render(fillImage, element, flags);
        } else {
            element.style.backgroundImage = "";
            element.style.backgroundPosition = "";
            element.style.backgroundRepeat = "";
        }
    }
};

EchoAppRender.Font = { 

    render: function(font, element) {
        if (!font) {
            return;
        }
        if (font.typeface) {
            if (font.typeface instanceof Array) {
                element.style.fontFamily = font.typeface.join(",");
            } else {
                element.style.fontFamily = font.typeface;
            }
        }
        if (font.size) {
            element.style.fontSize = EchoAppRender.Extent.toCssValue(font.size);
        }

        if (font.bold) {
            element.style.fontWeight = "bold";
        }
        if (font.italic) {
            element.style.fontStyle = "italic";
        }
        if (font.underline) {
            element.style.textDecoration = "underline";
        } else if (font.overline) {
            element.style.textDecoration = "overline";
        } else if (font.lineThrough) {
            element.style.textDecoration = "line-through";
        }
    },
    
    renderClear: function(font, element) {
        if (font) {
            this.render(font, element);
        } else {
            element.style.fontFamily = "";
            element.style.fontSize = "";
            element.style.fontWeight = "";
            element.style.fontStyle = "";
            element.style.textDecoration = "";
        }
    }
};

EchoAppRender.ImageReference = {

    getUrl: function(imageReference) {
        return imageReference ? (typeof(imageReference) == "string" ? imageReference : imageReference.url) : null;
    },

    renderImg: function(imageReference, imgElement) {
        if (!imageReference) {
            return;
        }
        
        if (typeof(imageReference) == "string") {
            imgElement.src = imageReference;
        } else {
            imgElement.src = imageReference.url;
            if (imageReference.width) {
                imgElement.style.width = EchoAppRender.Extent.toCssValue(imageReference.width, true);
            }
            if (imageReference.height) {
                imgElement.style.height = EchoAppRender.Extent.toCssValue(imageReference.height, false);
            }
        }
    }
};

EchoAppRender.Insets = {

    /**
     * Regular expression to test extents which are entirely presented in pixels
     * and may thus be directly added to CSS.
     */
    _FORMATTED_PIXEL_INSETS: /^(-?\d+px *){1,4}$/,

    _ZERO: { top: 0, right: 0, bottom: 0, left: 0 },
    
    /**
     * Mapping between number of inset values provided and arrays which represent the
     * inset value index for the top, right, bottom, and left value. 
     */
    _INDEX_MAPS: {
        1: [0, 0, 0, 0], 
        2: [0, 1, 0, 1], 
        3: [0, 1, 2, 1], 
        4: [0, 1, 2, 3] 
    },

    render: function(insets, element, styleAttribute) {
        switch(typeof(insets)) {
            case "number":
                element.style[styleAttribute] = insets + "px";
                break;
            case "string":
                if (this._FORMATTED_PIXEL_INSETS.test(insets)) {
                    element.style[styleAttribute] = insets;
                } else {
                    var pixelInsets = this.toPixels(insets);
                    element.style[styleAttribute] = pixelInsets.top + "px " + pixelInsets.right + "px "
                            + pixelInsets.bottom + "px " + pixelInsets.left + "px";
                }
                break;
        }
    },
    
    toCssValue: function(insets) {
        switch(typeof(insets)) {
            case "number":
                return insets + "px";
                break;
            case "string":
                if (this._FORMATTED_PIXEL_INSETS.test(insets)) {
                    return insets;
                } else {
                    var pixelInsets = this.toPixels(insets);
                    return pixelInsets.top + "px " + pixelInsets.right + "px "
                            + pixelInsets.bottom + "px " + pixelInsets.left + "px";
                }
                break;
        }
        return "";
    },
    
    toPixels: function(insets) {
        if (insets == null) {
            return this._ZERO;
        } else if (typeof(insets) == "number") {
            return { top: insets, right: insets, bottom: insets, left: insets };
        }
        
        insets = insets.split(" ");
        var map = this._INDEX_MAPS[insets.length];
        return {
            top: EchoAppRender.Extent.toPixels(insets[map[0]], false),
            right: EchoAppRender.Extent.toPixels(insets[map[1]], true),
            bottom: EchoAppRender.Extent.toPixels(insets[map[2]], false),
            left: EchoAppRender.Extent.toPixels(insets[map[3]], true)
        };
    }
};

/**
 * @class Manages floating windows, e.g., window panes in a content pane.
 * Provides listener facility to receive notifications when the panes are raised or lowered,
 * such that floating panes may adjust their z-indices appropriately for correct display.
 * Registered listeners will be notified when one or more z-indices have changed.
 */
EchoAppRender.FloatingPaneManager = Core.extend({

    /**
     * Creates a new Floating Pane Manager.
     */
    $construct: function() {
        this._floatingPanes = null;
        this._listeners = null;
    },
    
    /**
     * Adds a floating pane to be managed, or, if the floating pane already exists,
     * raises it to the top.
     * The floating pane will be placed above all others, at the highest z-index.
     * 
     * @param {String} renderId the id of the floating pane
     * @return the initial z-index of the added floating pane
     */
    add: function(renderId) {
        if (!this._floatingPanes) {
            this._floatingPanes = [];
        }
        Core.Arrays.remove(this._floatingPanes, renderId);
        this._floatingPanes.push(renderId);
        this._fireZIndexEvent();
        return this._floatingPanes.length;
    },
    
    /**
     * Adds a z-index listener.  
     * 
     * @param {Function} the listener to add
     */
    addZIndexListener: function(l) {
        if (!this._listeners) {
            this._listeners = new Core.ListenerList();
        }
        this._listeners.addListener("zIndex", l);
    },
    
    /**
     * Notifies listeners of a z-index change.
     * @private
     */
    _fireZIndexEvent: function() {
        if (this._listeners) {
            this._listeners.fireEvent({type: "zIndex", source: this});
        }
    },
    
    /**
     * Returns the z-index of the floating pane with the specified id.
     * -1 is returned if the pane is not registered.
     * 
     * @param {String} renderId the id of the floating pane
     * @return the z-index
     */
    getIndex: function(renderId) {
        if (this._floatingPanes) {
            var index = Core.Arrays.indexOf(this._floatingPanes, renderId);
            return index == -1 ? -1 : index + 1;
        } else {
            return -1;
        }
    },
    
    /**
     * Removes a floating pane from being managed.
     * 
     * @param {String} renderId the id of the floating pane
     */
    remove: function(renderId) {
        if (!this._floatingPanes) {
            return;
        }
        Core.Arrays.remove(this._floatingPanes, renderId);
        this._fireZIndexEvent();
    },
    
    /**
     * Removes a z-index listener.
     * 
     * @param {Function} the listener to remove
     */
    removeZIndexListener: function(l) {
        if (!this._listeners) {
            return;
        }
        this._listeners.removeListener("zIndex", l);
    }
});

EchoAppRender.TriCellTable = Core.extend({

    $static: {
        
        INVERTED: 1,
        VERTICAL: 2,
        
        LEADING_TRAILING: 0,
        TRAILING_LEADING: 1, // INVERTED
        TOP_BOTTOM: 2,       // VERTICAL
        BOTTOM_TOP: 3,       // VERTICAL | INVERTED

        //FIXME. verify this method will work with  RTL settings (not tested)
        getOrientation: function(component, propertyName) {
            var position = component.render(propertyName);
            var orientation;
            if (position) {
                switch (EchoAppRender.Alignment.getRenderedHorizontal(position, component)) {
                case "leading":  orientation = this.LEADING_TRAILING; break;
                case "trailing": orientation = this.TRAILING_LEADING; break;
                case "left":     orientation = this.LEADING_TRAILING; break;
                case "right":    orientation = this.TRAILING_LEADING; break;
                default:
                    switch (EchoAppRender.Alignment.getVertical(position, component)) {
                    case "top":    orientation = this.TOP_BOTTOM;       break;
                    case "bottom": orientation = this.BOTTOM_TOP;       break;
                    default:       orientation = this.TRAILING_LEADING; break;
                    }
                }
            } else {
                orientation = this.TRAILING_LEADING;
            }
            return orientation;
        },
        
        /**
         * @private
         */
        _createTablePrototype: function() {
            var tableElement = document.createElement("table");
            tableElement.style.borderCollapse = "collapse";
            tableElement.style.padding = "0px";
            
            tbodyElement = document.createElement("tbody");
            tableElement.appendChild(tbodyElement);
            
            return tableElement;
        }
    },
    
    $load: function() {
        this._tablePrototype = this._createTablePrototype(); 
    },
    
    /**
     * The rendered TABLE element.
     * @type Element
     */
    tableElement: null,
    
    /**
     * The rendered TBODY element.
     * @type Element
     */
    tbodyElement: null,

    /**
     * Creates a new <code>TriCellTable</code>
     * 
     * @param orientation0_1 the orientation of element 0 with respect to element 1, one of 
     *        the following values:
     *        <ul>
     *        <ul>
     *        <li>LEADING_TRAILING (element 0 is leading element 1)</li>
     *        <li>TRAILING_LEADING (element 1 is leading element 0)</li>
     *        <li>TOP_BOTTOM (element 0 is above element 1)</li>
     *        <li>BOTTOM_TOP (element 1 is above element 0)</li>
     *        </ul>
     * @param margin0_1 the margin size between element 0 and element 1
     * @param orientation01_2 (omitted for two-cell tables)
     *        the orientation of Elements 0 and 1 with 
     *        respect to Element 2, one of the following values:
     *        <ul>
     *        <li>LEADING_TRAILING (elements 0 and 1 are leading element 2)</li>
     *        <li>TRAILING_LEADING (element 2 is trailing elements 0 and 1)</li>
     *        <li>TOP_BOTTOM (elements 0 and 1 are above element 2)</li>
     *        <li>BOTTOM_TOP (element 2 is above elements 0 and 1)</li>
     *        </ul>
     * @param margin01_2 (omitted for two-cell tables)
     *        The margin size between the combination
     *        of elements 0 and 1 and element 2.
     */
    $construct: function(orientation0_1, margin0_1, orientation01_2, margin01_2) {
        this.tableElement = EchoAppRender.TriCellTable._tablePrototype.cloneNode(true);
        this.tbodyElement = this.tableElement.firstChild;
        
        if (orientation01_2 == null) {
            this.configure2(orientation0_1, margin0_1);
        } else {
            this.configure3(orientation0_1, margin0_1, orientation01_2, margin01_2);
        }
    },
    
    addColumn: function(trElement, tdElement) {
        if (tdElement != null) {
            trElement.appendChild(tdElement);
        }
    },
    
    addRow: function(tdElement) {
        if (tdElement == null) {
            return;
        }
        var trElement = document.createElement("tr");
        trElement.appendChild(tdElement);
        this.tbodyElement.appendChild(trElement);
    },
    
    addSpacer: function(parentElement, size, vertical) {
        var divElement = document.createElement("div");
        divElement.style.width = vertical ? "1px" : size + "px";
        divElement.style.height = vertical ? size + "px" : "1px";
        parentElement.appendChild(divElement);
    },
    
    /**
     * @param id the id of 
     */
    configure2: function(orientation0_1, margin0_1) {
        this.tdElements = [document.createElement("td"), document.createElement("td")];
        this.tdElements[0].style.padding = "0px";
        this.tdElements[1].style.padding = "0px";
        this.marginTdElements = new Array(1);
        
        if (margin0_1) {
            this.marginTdElements[0] = document.createElement("td");
            this.marginTdElements[0].style.padding = "0px";
            if ((orientation0_1 & EchoAppRender.TriCellTable.VERTICAL) == 0) {
                this.marginTdElements[0].style.width = margin0_1 + "px";
                this.addSpacer(this.marginTdElements[0], margin0_1, false);
            } else {
                this.marginTdElements[0].style.height = margin0_1 + "px";
                this.addSpacer(this.marginTdElements[0], margin0_1, true);
            }
        }
        
        if (orientation0_1 & EchoAppRender.TriCellTable.VERTICAL) {
            // Vertically oriented.
            if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                // Inverted (bottom to top).
                this.addRow(this.tdElements[1]);
                this.addRow(this.marginTdElements[0]);
                this.addRow(this.tdElements[0]);
            } else {
                // Normal (top to bottom).
                this.addRow(this.tdElements[0]);
                this.addRow(this.marginTdElements[0]);
                this.addRow(this.tdElements[1]);
            }
        } else {
            // Horizontally oriented.
            var trElement = document.createElement("tr");
            if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                // Trailing to leading.
                this.addColumn(trElement, this.tdElements[1]);
                this.addColumn(trElement, this.marginTdElements[0]);
                this.addColumn(trElement, this.tdElements[0]);
            } else {
                // Leading to trailing.
                this.addColumn(trElement, this.tdElements[0]);
                this.addColumn(trElement, this.marginTdElements[0]);
                this.addColumn(trElement, this.tdElements[1]);
            }
            this.tbodyElement.appendChild(trElement);
        }
    },
    
    configure3: function(orientation0_1, margin0_1, orientation01_2, margin01_2) {
        this.tdElements = new Array(3);
        for (var i = 0; i < 3; ++i) {
            this.tdElements[i] = document.createElement("td");
            this.tdElements[i].style.padding = "0px";
        }
        this.marginTdElements = new Array(2);
        
        if (margin0_1 || margin01_2 != null) {
            if (margin0_1 && margin0_1 > 0) {
                this.marginTdElements[0] = document.createElement("td");
                if (orientation0_1 & EchoAppRender.TriCellTable.VERTICAL) {
                    this.marginTdElements[0].style.height = margin0_1 + "px";
                    this.addSpacer(this.marginTdElements[0], margin0_1, true);
                } else {
                    this.marginTdElements[0].style.width = margin0_1 + "px";
                    this.addSpacer(this.marginTdElements[0], margin0_1, false);
                }
            }
            if (margin01_2 != null && margin01_2 > 0) {
                this.marginTdElements[1] = document.createElement("td");
                if (orientation0_1 & EchoAppRender.TriCellTable.VERTICAL) {
                    this.marginTdElements[1].style.height = margin01_2 + "px";
                    this.addSpacer(this.marginTdElements[1], margin01_2, true);
                } else {
                    this.marginTdElements[1].style.width = margin01_2 + "px";
                    this.addSpacer(this.marginTdElements[1], margin01_2, false);
                }
            }
        }
        
        if (orientation0_1 & EchoAppRender.TriCellTable.VERTICAL) {
            // Vertically oriented 0/1.
            if (orientation01_2 & EchoAppRender.TriCellTable.VERTICAL) {
                // Vertically oriented 01/2
                
                if (orientation01_2 & EchoAppRender.TriCellTable.INVERTED) {
                    // 2 before 01: render #2 and margin at beginning of TABLE.
                    this.addRow(this.tdElements[2]);
                    this.addRow(this.marginTdElements[1]);
                }
                
                // Render 01
                if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                    // Inverted (bottom to top)
                    this.addRow(this.tdElements[1]);
                    this.addRow(this.marginTdElements[0]);
                    this.addRow(this.tdElements[0]);
                } else {
                    // Normal (top to bottom)
                    this.addRow(this.tdElements[0]);
                    this.addRow(this.marginTdElements[0]);
                    this.addRow(this.tdElements[1]);
                }
    
                if (!(orientation01_2 & EchoAppRender.TriCellTable.INVERTED)) {
                    // 01 before 2: render #2 and margin at end of TABLE.
                    this.addRow(this.marginTdElements[1]);
                    this.addRow(this.tdElements[2]);
                }
            } else {
                // Horizontally oriented 01/2
                
                // Determine and apply row span based on presence of margin between 0 and 1.
                var rows = (margin0_1 && margin0_1 > 0) ? 3 : 2;
                this.tdElements[2].rowSpan = rows;
                if (this.marginTdElements[1]) {
                    this.marginTdElements[1].rowSpan = rows;
                }
                
                var trElement = document.createElement("tr");
                if (orientation01_2 & EchoAppRender.TriCellTable.INVERTED) {
                    this.addColumn(trElement, this.tdElements[2]);
                    this.addColumn(trElement, this.marginTdElements[1]);
                    if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                        this.addColumn(trElement, this.tdElements[1]);
                    } else {
                        this.addColumn(trElement, this.tdElements[0]);
                    }
                } else {
                    if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                        this.addColumn(trElement, this.tdElements[1]);
                    } else {
                        this.addColumn(trElement, this.tdElements[0]);
                    }
                    this.addColumn(trElement, this.marginTdElements[1]);
                    this.addColumn(trElement, this.tdElements[2]);
                }
                this.tbodyElement.appendChild(trElement);
                
                this.addRow(this.marginTdElements[0]);
                if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                    this.addRow(this.tdElements[0]);
                } else {
                    this.addRow(this.tdElements[1]);
                }
            }
        } else {
            // horizontally oriented 0/1
            if (orientation01_2 & EchoAppRender.TriCellTable.VERTICAL) {
                // vertically oriented 01/2
    
                // determine and apply column span based on presence of margin between 0 and 1
                var columns = margin0_1 ? 3 : 2;
                this.tdElements[2].setAttribute("colspan", columns);
                if (this.marginTdElements[1] != null) {
                    this.marginTdElements[1].setAttribute("colspan", Integer.toString(columns));
                }
                
                if (orientation01_2 & EchoAppRender.TriCellTable.INVERTED) {
                    // 2 before 01: render #2 and margin at beginning of TR.
                    this.addRow(this.tdElements[2]);
                    this.addRow(this.marginTdElements[1]);
                }
                
                // Render 01
                trElement = document.createElement("tr");
                if ((orientation0_1 & EchoAppRender.TriCellTable.INVERTED) == 0) {
                    // normal (left to right)
                    this.addColumn(trElement, this.tdElements[0]);
                    this.addColumn(trElement, this.marginTdElements[0]);
                    this.addColumn(trElement, this.tdElements[1]);
                } else {
                    // inverted (right to left)
                    this.addColumn(trElement, this.tdElements[1]);
                    this.addColumn(trElement, this.marginTdElements[0]);
                    this.addColumn(trElement, this.tdElements[0]);
                }
                this.tbodyElement.appendChild(trElement);
                
                if (!(orientation01_2 & EchoAppRender.TriCellTable.INVERTED)) {
                    // 01 before 2: render margin and #2 at end of TR.
                    this.addRow(this.marginTdElements[1]);
                    this.addRow(this.tdElements[2]);
                }
            } else {
                // horizontally oriented 01/2
                trElement = document.createElement("tr");
                if (orientation01_2 & EchoAppRender.TriCellTable.INVERTED) {
                    // 2 before 01: render #2 and margin at beginning of TR.
                    this.addColumn(trElement, this.tdElements[2]);
                    this.addColumn(trElement, this.marginTdElements[1]);
                }
                
                // Render 01
                if (orientation0_1 & EchoAppRender.TriCellTable.INVERTED) {
                    // inverted (right to left)
                    this.addColumn(trElement, this.tdElements[1]);
                    this.addColumn(trElement, this.marginTdElements[0]);
                    this.addColumn(trElement, this.tdElements[0]);
                } else {
                    // normal (left to right)
                    this.addColumn(trElement, this.tdElements[0]);
                    this.addColumn(trElement, this.marginTdElements[0]);
                    this.addColumn(trElement, this.tdElements[1]);
                }
                
                if (!(orientation01_2 & EchoAppRender.TriCellTable.INVERTED)) {
                    this.addColumn(trElement, this.marginTdElements[1]);
                    this.addColumn(trElement, this.tdElements[2]);
                }
                
                this.tbodyElement.appendChild(trElement);        
            }
        }
    }
});
/**
 * Abstract base class for Echo clients.
 */
EchoClient = Core.extend({
    
    $static: {
    
        /**
         * Flag for verifyInput() flags parameter, indicating that input is 
         * a property update.
         */
        FLAG_INPUT_PROPERTY: 0x1,

        /**
         * Global array containing all active client instances in the current browser window.
         */
        _activeClients: [],

        /**
         * Global listener to respond to resizing of browser window.
         * Invokes _windowResizeListener() method on all active clients.
         * 
         * @param e the DOM resize event
         */
        _globalWindowResizeListener: function(e) {
            for (var i = 0; i < EchoClient._activeClients.length; ++i) {
                EchoClient._activeClients[i]._windowResizeListener(e);
            }
        },
        
        _resources: {},

        _listeners: new Core.ListenerList(),
        
        addResourceChangeListener: function(l) {
           this._listeners.addListener("resourceChange", l);
        },
    
        getResource: function(path) {
            return Core.get(this._resources, path);
        },
        
        removeResourceChangeListener: function(l) {
           this._listeners.removeListener("resourceChange", l);
        },
        
        setResource: function(path, newValue) {
            Core.set(this._resources, path, newValue); 
            this._listeners.fireEvent({ type: "resourceChange", source: this, path: path });
        }
    },
    
    $load: function() {
        // Register resize listener on containing window one time.
        WebCore.DOM.addEventListener(window, "resize", this._globalWindowResizeListener, false);
    },
    
    /**
     * The root DOM element in which the application is contained.
     */
    domainElement: null,
    
    /**
     * The application being managed by this client.
     */
    application: null,
    
    /**
     * Flag indicating whether a focus update is required once the renderer has completed its next update cycle.
     */
    focusUpdateRequired: false,
    
    /**
     * Id of last issued input restriction id (incremented to deliver unique identifiers). 
     * @type Integer
     * @private
     */
    _lastInputRestrictionId: 0,
    
    /**
     * Number of currently registered input restrictions.
     * @type Integer
     * @private
     */
    _inputRestrictionCount: 0,
    
    /**
     * Id (String) map containing input restrictions.
     * Values are booleans, true indicating property updates are NOT restricted, and false
     * indicated all updates are restricted.
     * @type Object
     * @private
     */
    _inputRescriptionMap: null,
    
    _keyPressListener: null,
    
    _applicationFocusListener: null,
    
    /**
     * The parent client.
     */
    parent: null,

    $construct: function() { 
        this._inputRestrictionMap = { };
        this._keyPressListener = Core.method(this, this._processKeyPress);
        this._applicationFocusListener = Core.method(this, this._processApplicationFocus);
    },
    
    $virtual: {

        /**
         * Returns a default named image.
         * May return null if the client does not provide a default image for the specified name.
         * Default implementation delegates to parent client
         * (if one is present) or otherwise returns null.
         * 
         * @param {String} imageName the image name 
         */
        getDefaultImage: function(imageName) {
            if (this.parent) {
                return this.parent.getDefaultImage(imageName);
            } else {
                return null;
            }
        },
        
        /**
         * Returns the URL of a service based on the serviceId.
         * Default implementation delegates to parent client
         * (if one is present) or otherwise returns null.
         * 
         * @param {String} serviceId the serviceId
         * @return the full URL
         * @type String
         */
        getServiceUrl: function(serviceId) {
            if (this.parent) {
                return this.parent.getServiceUrl(serviceId);
            } else {
                return null;
            }
        },

        /**
         * Determines if the specified component and containing application is ready to receive input.
         * This method should be overridden by client implementations as needed, returning the value
         * from this implementation if the client has no other reason to disallow input.
         * 
         * @param component optional parameter indicating the component to query (if omitted, only the
         *        applications readiness state will be investigated)
         * @return true if the application/component are ready to receive inputs
         */
        verifyInput: function(component, flags) {
            // Check for input restrictions.
            if (this._inputRestrictionCount != 0) {
                if (!flags & EchoClient.FLAG_INPUT_PROPERTY) {
                    // Input is not a property update, automatically return false if any input restrictions pressent.
                    return false;
                }
                for (var x in this._inputRestrictionMap) {
                    if (this._inputRestrictionMap[x] === false) {
                        // Input restriction set to false, indicating no updates, not even property updates.
                        return false;
                    }
                }
            }
        
            if (component) {
                return component.isActive();
            } else {
                return this.application.isActive();
            }
        },
        
        /**
         * Default dispose implementation.
         * Invokes configure(null, null) to deconfigure the client. 
         */
        dispose: function() {
            this.configure(null, null);
        }
    },
    
    /**
     * Configures/Deconfigures the client.  This method must be invoked
     * with the supported application/containing domain element before
     * the client is used, and invoked with null values before it is
     * disposed (in order to clean up resources).
     * 
     * @param application the application the client will support (if configuring)
     *        or null (if deconfiguring)
     * @param domainElement the DOM element into which the client will be rendered (if configuring),
     *        or null (if deconfiguring)
     */
    configure: function(application, domainElement) {
        if (this.application) {
            Core.Arrays.remove(EchoClient._activeClients, this);
            WebCore.EventProcessor.remove(this.domainElement, 
                    WebCore.Environment.QUIRK_IE_KEY_DOWN_EVENT_REPEAT ? "keydown" : "keypress", this._keyPressListener, false);
            this.application.removeFocusListener(this._applicationFocusListener);
        }
        
        this.application = application;
        this.domainElement = domainElement;
    
        if (this.application) {
            this.application.addFocusListener(this._applicationFocusListener);
            WebCore.EventProcessor.add(this.domainElement, 
                    WebCore.Environment.QUIRK_IE_KEY_DOWN_EVENT_REPEAT ? "keydown" : "keypress", this._keyPressListener, false);
            EchoClient._activeClients.push(this);
        }
    },
    
    /**
     * Registers a new input restriction.  Input will be restricted until this and all other
     * input restrictiosn are removed.
     *
     * @param {Boolean} allowPropertyUpdates flag indicating whether property updates should be
     *        allowed (if true) or whether all input should be restricted (if false)
     * @return a handle identifier for the input restriction, which will be used to unregister
     *         the restriction by invoking removeInputRestriction()
     */
    createInputRestriction: function(allowPropertyUpdates) {
        var id = (++this._lastInputRestrictionId).toString();
        ++this._inputRestrictionCount;
        this._inputRestrictionMap[id] = allowPropertyUpdates;
        return id;
    },
    
    /**
     * Listener for application change of component focus:
     * invokes focus() method on focused component's peer.
     * 
     * @param {Event} e the event
     */
    _processApplicationFocus: function(e) {
        var focusedComponent = this.application.getFocusedComponent();
        if (focusedComponent && focusedComponent.peer && focusedComponent.peer.renderFocus) {
            focusedComponent.peer.renderFocus();
            this.focusUpdateRequired = false;
        } else {
            this.focusUpdateRequired = true;
        }
    },
    
    /**
     * Root KeyDown event handler.
     * Specifically processes tab key events for focus management.
     * 
     * @param {Event} e the event
     */
    _processKeyPress: function(e) {
        if (e.keyCode == 9) { // Tab
            this.application.focusNext(e.shiftKey);
            WebCore.DOM.preventEventDefault(e);
            return false; // Stop propagation.
        }
        return true; // Allow propagation.
    },
    
    /**
     * Removes an input restriction.
     *
     * @param id the id of the input restriction to remove
     */
    removeInputRestriction: function(id) {
        if (this._inputRestrictionMap[id] === undefined) {
            return;
        }
        delete this._inputRestrictionMap[id];
        --this._inputRestrictionCount;
    },
    
    /**
     * Instance listener to respond to resizing of browser window.
     * 
     * @param e the DOM resize event
     */
    _windowResizeListener: function(e) {
        EchoRender.notifyResize(this.application.rootComponent);
    }
});

/**
 * @class Provides a tool for measuring performance of the Echo3 client engine.
 */
EchoClient.Timer = Core.extend({

    _times: null,
    
    _labels: null,
    
    /**
     * Creates a new debug timer.
     * 
     * @constructor
     */
    $construct: function() {
        this._times = [new Date().getTime()];
        this._labels = ["Start"];
    },
    
    /**
     * Marks the time required to complete a task.  This method should be invoked
     * when a task is completed with the 'label' specifying a description of the task.
     * 
     * @param {String} label a description of the completed task.
     */
    mark: function(label) {
        this._times.push(new Date().getTime());
        this._labels.push(label);
    },
    
    /**
     * Returns a String representation of the timer results, showing how long
     * each task required to complete (and included a total time).
     * 
     * @return the timer results
     * @type String
     */
    toString: function() {
        var out = "";
        for (var i = 1; i < this._times.length; ++i) {
            var time = this._times[i] - this._times[i - 1];
            out += this._labels[i] + ":" + time + " ";
        }
        out += "TOT:" + (this._times[this._times.length - 1] - this._times[0]) + "ms";
        return out;
    }
});

/**
 * @fileoverview
 * Freestanding Client Implementation.
 * Provides capapbility to develop server-independent applications.
 * Requires Core, WebCore, Application, Render, Serial, Client.
 */
 
/** 
 * @class 
 * FreeClient implementation.
 * The init() and dispose() lifecycle methods must be called before the client is used,
 * and when the client will no longer be used, respectively.
 */ 
EchoFreeClient = Core.extend(EchoClient, {

    _processUpdateRef: null,

    /**
     * Creates a new FreeClient.
     *
     * @param {EchoApp.Application} application the application the client operate on.
     * @param {Element} domainElement the HTML 
     */
    $construct: function(application, domainElement) {
        EchoClient.call(this);
        this._processUpdateRef = Core.method(this, this._processUpdate);;
        this.configure(application, domainElement);
    },

    /**
     * Disposes of the FreeClient.
     * This method must be invoked when the client will no longer be used, in order to clean up resources.
     */
    dispose: function() {
        WebCore.Scheduler.remove(this._autoUpdate);
        this.application.updateManager.removeUpdateListener(this._processUpdateRef);
        this._autoUpdate = null;
        EchoRender.renderComponentDispose(null, this.application.rootComponent);
        EchoClient.prototype.dispose.call(this);
    },
    
    _processUpdate: function(e) {
        //FIXME implement or remove
    },
    
    /**
     * Initializes the FreeClient.
     * This method must be invoked before the client is initially used.
     */
    init: function() {
        WebCore.init();
        this._autoUpdate = new EchoFreeClient.AutoUpdate(this);
        this.application.updateManager.addUpdateListener(this._processUpdateRef);
        WebCore.Scheduler.add(this._autoUpdate);
    }
});

/**
 * @class WebCore.Scheduler.Runnable to automatically update client when application state has changed.
 */
EchoFreeClient.AutoUpdate = Core.extend(WebCore.Scheduler.Runnable, {

    timeInterval: 10,
    
    repeat: true,
    
    _client: null,

    /**
     * Creates a new automatic render update runnable.
     * 
     * @param client the supported client
     */
    $construct: function(client) {
        this._client = client;
    },
    
    /**
     * Runnable run() implementation.
     */
    run: function() {
        EchoRender.processUpdates(this._client);
    }
});

/**
 * Component rendering peer: Button
 */
EchoAppRender.ButtonSync = Core.extend(EchoRender.ComponentSync, {

    $static: {
        _createPrototypeButton: function() {
            var divElement = document.createElement("div");
            divElement.tabIndex = "0";
            divElement.style.outlineStyle = "none";
            divElement.style.cursor = "pointer";
            divElement.style.overflow = "hidden";
            return divElement;
        },

        _defaultIconTextMargin: 5
    },
    
    $load: function() {
        this._prototypeButton = this._createPrototypeButton();
        EchoRender.registerPeer("Button", this);
    },
    
    _prototypeButton: null,
    
    _processRolloverExitRef: null,
    
    _processInitEventRef: null,
    
    $construct: function() { 
        this._processInitEventRef = Core.method(this, this._processInitEvent);
    },
    
    $virtual: {
        
        doAction: function() {
            this.component.doAction();
        },
        
        renderContent: function() {
            var text = this.component.render("text");
            var icon = EchoAppRender.getEffectProperty(this.component, "icon", "disabledIcon", !this._enabled);
            if (text) {
                if (icon) {
                    // Text and icon.
                    var iconTextMargin = this.component.render("iconTextMargin", 
                            EchoAppRender.ButtonSync._defaultIconTextMargin);
                    var orientation = EchoAppRender.TriCellTable.getOrientation(this.component, "textPosition");
                    var tct = new EchoAppRender.TriCellTable(orientation, 
                            EchoAppRender.Extent.toPixels(iconTextMargin));
                    this._renderButtonText(tct.tdElements[0], text);
                    this._iconElement = this._renderButtonIcon(tct.tdElements[1], icon);
                    this._divElement.appendChild(tct.tableElement);
                } else {
                    // Text only.
                    this._renderButtonText(this._divElement, text);
                }
            } else if (icon) {
                // Icon only.
                this._iconElement = this._renderButtonIcon(this._divElement, icon);
            }
        }
    },
    
    /**
     * Registers listners on the button.  This method is invoked lazily, i.e., the first time the button
     * is focused or moused over.  The initial focus/mouseover listeners are removed by this method.
     * This strategy is used for performance reasons due to the fact that many buttons may be present 
     * on the screen, and each button has many event listeners.
     */
    _addEventListeners: function() {
        this._processRolloverExitRef = Core.method(this, this._processRolloverExit);
    
        // Remove initialization listeners.
        WebCore.EventProcessor.remove(this._divElement, "focus", this._processInitEventRef);
        WebCore.EventProcessor.remove(this._divElement, "mouseover", this._processInitEventRef);
        
        WebCore.EventProcessor.add(this._divElement, "click", Core.method(this, this._processClick), false);
        WebCore.EventProcessor.add(this._divElement, "keypress", Core.method(this, this._processKeyPress), false);
        if (this.component.render("rolloverEnabled")) {
            var mouseEnterLeaveSupport = WebCore.Environment.PROPRIETARY_EVENT_MOUSE_ENTER_LEAVE_SUPPORTED;
            var enterEvent = mouseEnterLeaveSupport ? "mouseenter" : "mouseover";
            var exitEvent = mouseEnterLeaveSupport ? "mouseleave" : "mouseout";
            WebCore.EventProcessor.add(this._divElement, enterEvent, 
                    Core.method(this, this._processRolloverEnter), false);
            WebCore.EventProcessor.add(this._divElement, exitEvent, 
                    Core.method(this, this._processRolloverExit), false);
        }
        if (this.component.render("pressedEnabled")) {
            WebCore.EventProcessor.add(this._divElement, "mousedown", Core.method(this, this._processPress), false);
            WebCore.EventProcessor.add(this._divElement, "mouseup", Core.method(this, this._processRelease), false);
        }
        WebCore.EventProcessor.add(this._divElement, "focus", Core.method(this, this._processFocus), false);
        WebCore.EventProcessor.add(this._divElement, "blur", Core.method(this, this._processBlur), false);
        
        WebCore.EventProcessor.Selection.disable(this._divElement);
    },
    
    getFocusFlags: function() {
        return EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_ALL;
    },
    
    _processBlur: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return;
        }
        this._setFocusState(false);
    },
    
    _processClick: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return;
        }
        this.component.application.setFocusedComponent(this.component);
        this.doAction();
    },
    
    _processFocus: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return;
        }
        this._setFocusState(true);
    },
    
    /**
     * Initial focus/mouseover listener.
     * This listener is invoked the FIRST TIME the button is focused or moused over.
     * It invokes the addListeners() method to lazily add the full listener set to the button.
     */
    _processInitEvent: function(e) {
        this._addEventListeners();
        switch (e.type) {
        case "focus":
            this._processFocus(e);
            break;
        case "mouseover":
            if (this.component.render("rolloverEnabled")) {
                this._processRolloverEnter(e);
            }
            break;
        }
    },
    
    _processKeyPress: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return true;
        }
        if (e.keyCode == 13) { // FIXME This will fail in IE (I think)
            this.doAction();
            return false;
        } else {
            return true;
        }
    },
    
    _processPress: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return;
        }
        WebCore.DOM.preventEventDefault(e);
        this._setPressedState(true);
    },
    
    _processRelease: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return;
        }
        this._setPressedState(false);
    },
    
    _processRolloverEnter: function(e) {
        if (!this.client.verifyInput(this.component) || WebCore.dragInProgress) {
            return;
        }
        this.component.application.addFocusListener(this._processRolloverExitRef);
        this._setRolloverState(true);
    },
    
    _processRolloverExit: function(e) {
        if (!this.client.verifyInput(this.component)) {
            return;
        }
        if (this._processRolloverExitRef) {
            this.component.application.removeFocusListener(this._processRolloverExitRef);
        }
        this._setRolloverState(false);
    },
    
    renderAdd: function(update, parentElement) {
        this._enabled = this.component.isRenderEnabled();
        
        this._divElement = EchoAppRender.ButtonSync._prototypeButton.cloneNode(false); 
        this._divElement.id = this.component.renderId;
    
        if (this._enabled) {
            EchoAppRender.Color.render(this.component.render("foreground"), this._divElement, "color");
            EchoAppRender.Color.render(this.component.render("background"), this._divElement, "backgroundColor");
            EchoAppRender.Border.render(this.component.render("border"), this._divElement);
            EchoAppRender.Font.render(this.component.render("font"), this._divElement);
            EchoAppRender.FillImage.render(this.component.render("backgroundImage"), this._divElement);
        } else {
            EchoAppRender.Color.render(EchoAppRender.getEffectProperty(this.component, "foreground", "disabledForeground", true), 
                    this._divElement, "color");
            EchoAppRender.Color.render(EchoAppRender.getEffectProperty(this.component, "background", "disabledBackground", true), 
                    this._divElement, "backgroundColor");
            EchoAppRender.Border.render(EchoAppRender.getEffectProperty(this.component, "border", "disabledBorder", true), 
                    this._divElement);
            EchoAppRender.Font.render(EchoAppRender.getEffectProperty(this.component, "font", "disabledFont", true), 
                    this._divElement);
            EchoAppRender.FillImage.render(EchoAppRender.getEffectProperty(this.component, 
                    "backgroundImage", "disabledBackgroundImage", true), this._divElement);
        }
        
        EchoAppRender.Insets.render(this.component.render("insets"), this._divElement, "padding");
        EchoAppRender.Alignment.render(this.component.render("alignment"), this._divElement, true, this.component);
        
        var toolTipText = this.component.render("toolTipText");
        if (toolTipText) {
            this._divElement.title = toolTipText;
        }
        var width = this.component.render("width");
        if (width) {
            this._divElement.style.width = EchoAppRender.Extent.toCssValue(width);
        }
        var height = this.component.render("height");
        if (height) {
            this._divElement.style.height = EchoAppRender.Extent.toCssValue(height);
            this._divElement.style.overflow = "hidden";
        }
        
        this.renderContent();
        
        if (this._enabled) {
            // Add event listeners for focus and mouseover.  When invoked, these listeners will register the full gamut
            // of button event listeners.  There may be a large number of such listeners depending on how many effects
            // are enabled, and as such we do this lazily for performance reasons.
            WebCore.EventProcessor.add(this._divElement, "focus", this._processInitEventRef, false);
            WebCore.EventProcessor.add(this._divElement, "mouseover", this._processInitEventRef, false);
        }
    
        parentElement.appendChild(this._divElement);
    },
    
    _renderButtonText: function(element, text) {
        var textAlignment = this.component.render("textAlignment") 
        if (textAlignment) {
            EchoAppRender.Alignment.render(textAlignment, element, true, this.component);
        }
        element.appendChild(document.createTextNode(text));
        if (!this.component.render("lineWrap", true)) {
            element.style.whiteSpace = "nowrap";
        }
    },
    
    _renderButtonIcon: function(element, icon) {
        var alignment = this.component.render("alignment") 
        if (alignment) {
            EchoAppRender.Alignment.render(alignment, element, true, this.component);
        }
        var imgElement = document.createElement("img");
        EchoAppRender.ImageReference.renderImg(icon, imgElement);
        element.appendChild(imgElement);
        return imgElement;
    },
    
    renderDispose: function(update) {
        if (this._processRolloverExitRef) {
            this.client.application.removeFocusListener(this._processRolloverExitRef);
        }
        WebCore.EventProcessor.removeAll(this._divElement);
        this._iconElement = null;
    },

    renderFocus: function() {
        WebCore.DOM.focusElement(this._divElement);
        this._setFocusState(true);
    },
    
    renderUpdate: function(update) {
        var element = this._divElement;
        var containerElement = element.parentNode;
        this.renderDispose(update);
        containerElement.removeChild(element);
        this.renderAdd(update, containerElement);
        return false; // Child elements not supported: safe to return false.
    },
    
    _setFocusState: function(focusState) {
        if (!this.component.render("focusedEnabled")) {
            // Render default focus aesthetic.
            var background = this.component.render("background");
            if (background != null) {
                var newBackground = focusState ? EchoAppRender.Color.adjust(background, 0x20, 0x20, 0x20) : background;
                EchoAppRender.Color.render(newBackground, this._divElement, "backgroundColor");
            }
            return;
        } else {
            var foreground = EchoAppRender.getEffectProperty(this.component, "foreground", "focusedForeground", focusState);
            var background = EchoAppRender.getEffectProperty(this.component, "background", "focusedBackground", focusState);
            var backgroundImage = EchoAppRender.getEffectProperty(
                    this.component, "backgroundImage", "focusedBackgroundImage", focusState);
            var font = EchoAppRender.getEffectProperty(this.component, "font", "focusedFont", focusState);
            var border = EchoAppRender.getEffectProperty(this.component, "border", "focusedBorder", focusState);
            
            EchoAppRender.Color.renderClear(foreground, this._divElement, "color");
            EchoAppRender.Color.renderClear(background, this._divElement, "backgroundColor");
            EchoAppRender.FillImage.renderClear(backgroundImage, this._divElement, "backgroundColor");
            EchoAppRender.Font.renderClear(font, this._divElement);
            EchoAppRender.Border.renderClear(border, this._divElement);
        
            if (this._iconElement) {
                var iconUrl = EchoAppRender.ImageReference.getUrl(
                        EchoAppRender.getEffectProperty(this.component, "icon", "focusedIcon", focusState));
                if (iconUrl != this._iconElement.src) {
                    this._iconElement.src = iconUrl;
                }
            }
        }
    },
    
    _setPressedState: function(pressedState) {
        var foreground = EchoAppRender.getEffectProperty(this.component, "foreground", "pressedForeground", pressedState);
        var background = EchoAppRender.getEffectProperty(this.component, "background", "pressedBackground", pressedState);
        var backgroundImage = EchoAppRender.getEffectProperty(
                this.component, "backgroundImage", "pressedBackgroundImage", pressedState);
        var font = EchoAppRender.getEffectProperty(this.component, "font", "pressedFont", pressedState);
        var border = EchoAppRender.getEffectProperty(this.component, "border", "pressedBorder", pressedState);
        
        EchoAppRender.Color.renderClear(foreground, this._divElement, "color");
        EchoAppRender.Color.renderClear(background, this._divElement, "backgroundColor");
        EchoAppRender.FillImage.renderClear(backgroundImage, this._divElement, "backgroundColor");
        EchoAppRender.Font.renderClear(font, this._divElement);
        EchoAppRender.Border.renderClear(border, this._divElement);
        
        if (this._iconElement) {
            var iconUrl = EchoAppRender.ImageReference.getUrl(
                    EchoAppRender.getEffectProperty(this.component, "icon", "pressedIcon", pressedState));
            if (iconUrl != this._iconElement.src) {
                this._iconElement.src = iconUrl;
            }
        }
    },
    
    _setRolloverState: function(rolloverState) {
        var foreground = EchoAppRender.getEffectProperty(this.component, "foreground", "rolloverForeground", rolloverState);
        var background = EchoAppRender.getEffectProperty(this.component, "background", "rolloverBackground", rolloverState);
        var backgroundImage = EchoAppRender.getEffectProperty(
                this.component, "backgroundImage", "rolloverBackgroundImage", rolloverState);
        var font = EchoAppRender.getEffectProperty(this.component, "font", "rolloverFont", rolloverState);
        var border = EchoAppRender.getEffectProperty(this.component, "border", "rolloverBorder", rolloverState);
        
        EchoAppRender.Color.renderClear(foreground, this._divElement, "color");
        EchoAppRender.Color.renderClear(background, this._divElement, "backgroundColor");
        EchoAppRender.FillImage.renderClear(backgroundImage, this._divElement, "backgroundColor");
        EchoAppRender.Font.renderClear(font, this._divElement);
        EchoAppRender.Border.renderClear(border, this._divElement);
    
        if (this._iconElement) {
            var iconUrl = EchoAppRender.ImageReference.getUrl(
                    EchoAppRender.getEffectProperty(this.component, "icon", "rolloverIcon", rolloverState));
            if (iconUrl != this._iconElement.src) {
                this._iconElement.src = iconUrl;
            }
        }
    }
});

/**
 * Component rendering peer: ToggleButton
 */
EchoAppRender.ToggleButtonSync = Core.extend(EchoAppRender.ButtonSync, {
    
    $load: function() {
        EchoRender.registerPeer("ToggleButton", this);
    },
    
    _selected: false,

    _stateElement: null,
    
    $abstract: {
        createStateElement: function() { },
    
        updateStateElement: function() { }
    },
    
    $virtual: {
        doAction: function() {
            this.setSelected(!this._selected);
            EchoAppRender.ButtonSync.prototype.doAction.call(this);
        }
    },
    
    renderAdd: function(update, parentElement) {
        this._selected = this.component.render("selected");
        
        EchoAppRender.ButtonSync.prototype.renderAdd.call(this, update, parentElement);
    },
    
    getStateIcon: function() {
        var icon;
        if (this._selected) {
            icon = EchoAppRender.getEffectProperty(this.component, "selectedStateIcon", "disabledSelectedStateIcon", 
                    !this._enabled);
        }
        if (!icon) {
            icon = EchoAppRender.getEffectProperty(this.component, "stateIcon", "disabledStateIcon", !this._enabled);
        }
        return icon;
    },
    
    renderContent: function() {
        var text = this.component.render("text");
        var icon = this.component.render("icon");
        this._stateElement = this.createStateElement();
        
        var entityCount = (text ? 1 : 0) + (icon ? 1 : 0) + (this._stateElement ? 1 : 0);
        if (entityCount == 1) {
            if (text) {
                this._renderButtonText(this._divElement, text);
            } else if (icon) {
                this._iconElement = this._renderButtonIcon(this._divElement, icon);
            } else {
                this._divElement.appendChild(this._stateElement);
            }
        } else if (entityCount == 2) {
            var orientation = EchoAppRender.TriCellTable.getOrientation(this.component, "textPosition");
            var margin;
            if (this._stateElement) {
                margin = this.component.render("stateMargin", EchoAppRender.ButtonSync._defaultIconTextMargin);
            } else {
                margin = this.component.render("iconTextMargin", EchoAppRender.ButtonSync._defaultIconTextMargin);
            }
            var tct = new EchoAppRender.TriCellTable(orientation, EchoAppRender.Extent.toPixels(margin));
            if (text) {
                this._renderButtonText(tct.tdElements[0], text);
                if (icon) {
                    this._iconElement = this._renderButtonIcon(tct.tdElements[1], icon);
                } else {
                    tct.tdElements[1].appendChild(this._stateElement);
                }
            } else {
                this._iconElement = this._renderButtonIcon(tct.tdElements[0], icon);
                tct.tdElements[1].appendChild(this._stateElement);
            }
            this._divElement.appendChild(tct.tableElement);
        } else if (entityCount == 3) {
            var orientation = EchoAppRender.TriCellTable.getOrientation(this.component, "textPosition");
            var margin = this.component.render("iconTextMargin", EchoAppRender.ButtonSync._defaultIconTextMargin);
            var stateOrientation = EchoAppRender.TriCellTable.TRAILING_LEADING;
            var stateMargin = this.component.render("stateMargin", EchoAppRender.ButtonSync._defaultIconTextMargin);
            var tct = new EchoAppRender.TriCellTable(orientation, 
                EchoAppRender.Extent.toPixels(margin), stateOrientation, EchoAppRender.Extent.toPixels(stateMargin));
            this._renderButtonText(tct.tdElements[0], text);
            this._iconElement = this._renderButtonIcon(tct.tdElements[1], icon);
            tct.tdElements[2].appendChild(this._stateElement);
            this._divElement.appendChild(tct.tableElement);
        }
    },
    
    renderDispose: function(update) {
        EchoAppRender.ButtonSync.prototype.renderDispose.call(this, update);
        if (this._stateElement) {
            WebCore.EventProcessor.removeAll(this._stateElement);
            this._stateElement = null;
        }
    },
    
    /**
     * Selects or deselects this button.
     * 
     * @param newState {Boolean} the new selection state
     */
    setSelected: function(newState) {
        if (this._selected == newState) {
            return;
        }
        this._selected = newState;
        this.component.set("selected", newState);
        
        this.updateStateElement();
    }
});

/**
 * Component rendering peer: CheckBox
 */
EchoAppRender.CheckBoxSync = Core.extend(EchoAppRender.ToggleButtonSync, {
    
    $load: function() {
        EchoRender.registerPeer("CheckBox", this);
    },
    
    createStateElement: function() {
        var stateIcon = this.getStateIcon();
        var stateElement;
        if (stateIcon) {
            stateElement = document.createElement("img");
            EchoAppRender.ImageReference.renderImg(stateIcon, stateElement);
        } else {
            stateElement = document.createElement("input");
            stateElement.type = "checkbox";
            stateElement.defaultChecked = this._selected ? true : false;
            WebCore.EventProcessor.add(stateElement, "change", Core.method(this, this._processStateChange), false);
        }
        return stateElement;
    },
    
    _processStateChange: function(e) {
        this.updateStateElement();
    },
        
    updateStateElement: function() {
        var stateIcon = this.getStateIcon();
        if (stateIcon) {
            this._stateElement.src = EchoAppRender.ImageReference.getUrl(stateIcon);
        } else {
            this._stateElement.checked = this._selected ? true : false;
        }
    }
});

/**
 * Component rendering peer: RadioButton
 */
EchoAppRender.RadioButtonSync = Core.extend(EchoAppRender.ToggleButtonSync, {

    $static: {
    
        _nextNameId: 0,
        
        /**
         * Contains mappings from RadioButton render ids to EchoAppRender.RadioButtonSync.Group objects.
         * 
         * @type Core.Arrays.LargeMap
         */
        _groups: new Core.Arrays.LargeMap()
    },

    $load: function() {
        EchoRender.registerPeer("RadioButton", this);;
    },
    
    _group: null,

    $construct: function() {
        EchoAppRender.ToggleButtonSync.call(this);
    },

    doAction: function() {
        if (this._group) {
            this._group.deselect();
        }
        EchoAppRender.ToggleButtonSync.prototype.doAction.call(this);
    },
    
    renderAdd: function(update, parentElement) {
        var groupId = this.component.render("group");
        if (groupId != null) {
            var group = EchoAppRender.RadioButtonSync._groups.map[groupId];
            if (!group) {
                group = new EchoAppRender.RadioButtonSync.Group(groupId);
                EchoAppRender.RadioButtonSync._groups.map[groupId] = group;
            }
            group.add(this);
            this._group = group;
        }
        EchoAppRender.ToggleButtonSync.prototype.renderAdd.call(this, update, parentElement);
    },
    
    createStateElement: function() {
        var stateIcon = this.getStateIcon();
        var stateElement;
        if (stateIcon) {
            stateElement = document.createElement("img");
            EchoAppRender.ImageReference.renderImg(stateIcon, stateElement);
        } else {
            stateElement = document.createElement("input");
            stateElement.type = "radio";
            stateElement.name = "__echo_" + EchoAppRender.RadioButtonSync._nextNameId++;
            stateElement.defaultChecked = this._selected ? true : false;
            WebCore.EventProcessor.add(stateElement, "change", Core.method(this, this._processStateChange), false);
        }
        return stateElement;
    },
    
    _processStateChange: function(e) {
        this.updateStateElement();
    },
    
    renderDispose: function(update) {
        EchoAppRender.ToggleButtonSync.prototype.renderDispose.call(this, update);
        if (this._group) {
            this._group.remove(this);
            if (this._group.size() == 0) {
                EchoAppRender.RadioButtonSync._groups.remove(this._group.id);
            }
            this._group = null;
        }
    },
    
    updateStateElement: function() {
        var stateIcon = this.getStateIcon();
        if (stateIcon) {
            this._stateElement.src = EchoAppRender.ImageReference.getUrl(stateIcon);
        } else {
            this._stateElement.checked = this._selected ? true : false;
        }
    }
});

EchoAppRender.RadioButtonSync.Group = Core.extend({

    id: null,
    
    _buttons: null,

    /**
     * Creates a RadioButton group.
     * 
     * @param id {String} the id
     */
    $construct: function(id) {
        this.id = id;
        this._buttons = [];
    },
    
    /**
     * Adds the specified button to this button group.
     *
     * @param button {EchoRender.ComponentSync.ToggleButton} the button
     */
    add: function(button) {
        this._buttons.push(button);
    },
    
    /**
     * Deselects all buttons in this button group.
     */
    deselect: function() {
        for (var i = 0; i < this._buttons.length; ++i) {
            this._buttons[i].setSelected(false);
        }
    },
    
    /**
     * Removes the specified button from this button group.
     * 
     * @param button {EchoRender.ComponentSync.ToggleButton} the button
     */
    remove: function(button) {
        // Find index of button in array.
        var buttonIndex = -1;
        for (var i = 0; i < this._buttons.length; ++i) {
            if (this._buttons[i] == button) {
                buttonIndex = i;
                break;
            }
        }
        
        if (buttonIndex == -1) {
            // Button does not exist in group.
            throw new Error("No such button: " + button.component.renderId);
        }
        
        if (this._buttons.length == 1) {
            // Array will now be empty.
            this._buttons = [];
        } else {
            // Buttons remain, remove button from button group.
            this._buttons[buttonIndex] = this._buttons[this._buttons.length - 1];
            this._buttons.length = this._buttons.length - 1;
        }
    },

    /**
     * Gets the amount of buttons contained by this button group.
     * 
     * @return the number of buttons.
     * @type {Number}
     */
    size: function() {
        return this._buttons.length;
    }
});
/**
 * Component rendering peer: Column
 */
EchoAppRender.ColumnSync = Core.extend(EchoRender.ComponentSync, {

    $load: function() {
        EchoRender.registerPeer("Column", this);
    },
    
    _processKeyPress: function(e) {
        switch (e.keyCode) {
        case 38:
        case 40:
            var focusPrevious = e.keyCode == 38;
            var focusedComponent = this.component.application.getFocusedComponent();
            if (focusedComponent && focusedComponent.peer && focusedComponent.peer.getFocusFlags) {
                var focusFlags = focusedComponent.peer.getFocusFlags();
                if ((focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_UP)
                        || (!focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_DOWN)) {
                    var focusChild = this.component.application.focusManager.findInParent(this.component, focusPrevious);
                    if (focusChild) {
                        this.component.application.setFocusedComponent(focusChild);
                        WebCore.DOM.preventEventDefault(e);
                        return false;
                    }
                }
            }
            break;
        }
        return true;
    },
    
    renderAdd: function(update, parentElement) {
        this._divElement = document.createElement("div");
        this._divElement.id = this.component.renderId;
        this._divElement.style.outlineStyle = "none";
        this._divElement.tabIndex = "-1";
    
        EchoAppRender.Border.render(this.component.render("border"), this._divElement);
        EchoAppRender.Color.renderFB(this.component, this._divElement);
        EchoAppRender.Font.render(this.component.render("font"), this._divElement);
        EchoAppRender.Insets.render(this.component.render("insets"), this._divElement, "padding");
    
        this._cellSpacing = EchoAppRender.Extent.toPixels(this.component.render("cellSpacing"), false);
        if (this._cellSpacing) {
            this._spacingPrototype = document.createElement("div");
            this._spacingPrototype.style.height = this._cellSpacing + "px";
            this._spacingPrototype.style.fontSize = "1px";
            this._spacingPrototype.style.lineHeight = "0px";
        }
        
        this._childIdToElementMap = {};
        
        var componentCount = this.component.getComponentCount();
        for (var i = 0; i < componentCount; ++i) {
            var child = this.component.getComponent(i);
            this._renderAddChild(update, child);
        }
        
        WebCore.EventProcessor.add(this._divElement, 
                WebCore.Environment.QUIRK_IE_KEY_DOWN_EVENT_REPEAT ? "keydown" : "keypress",
                Core.method(this, this._processKeyPress), false);
        
        parentElement.appendChild(this._divElement);
    },
    
    _renderAddChild: function(update, child, index) {
        var divElement = document.createElement("div");
        this._childIdToElementMap[child.renderId] = divElement;
        EchoRender.renderComponentAdd(update, child, divElement);
    
        var layoutData = child.render("layoutData");
        if (layoutData) {
            EchoAppRender.Color.render(layoutData.background, divElement, "backgroundColor");
            EchoAppRender.FillImage.render(layoutData.backgroundImage, divElement);
            EchoAppRender.Insets.render(layoutData.insets, divElement, "padding");
            EchoAppRender.Alignment.render(layoutData.alignment, divElement, true, this.component);
            if (layoutData.height) {
                divElement.style.height = EchoAppRender.Extent.toPixels(layoutData.height, false) + "px";
            }
        }
        
        if (index != null) {
            var currentChildCount;
            if (this._divElement.childNodes.length >= 3 && this._cellSpacing) {
                currentChildCount = (this._divElement.childNodes.length + 1) / 2;
            } else {
                currentChildCount = this._divElement.childNodes.length;
            }
            if (index == currentChildCount) {
                index = null;
            }
        }
        if (index == null) {
            // Full render or append-at-end scenario
            
            // Render spacing div first if index != 0 and cell spacing enabled.
            if (this._cellSpacing && this._divElement.firstChild) {
                this._divElement.appendChild(this._spacingPrototype.cloneNode(false));
            }
    
            // Render child div second.
            this._divElement.appendChild(divElement);
        } else {
            // Partial render insert at arbitrary location scenario (but not at end)
            var insertionIndex = this._cellSpacing ? index * 2 : index;
            var beforeElement = this._divElement.childNodes[insertionIndex];
            
            // Render child div first.
            this._divElement.insertBefore(divElement, beforeElement);
            
            // Then render spacing div if required.
            if (this._cellSpacing) {
                this._divElement.insertBefore(this._spacingPrototype.cloneNode(false), beforeElement);
            }
        }
    },
    
    _renderRemoveChild: function(update, child) {
        var childElement = this._childIdToElementMap[child.renderId];
    
        if (this._cellSpacing) {
            // If cell spacing is enabled, remove a spacing element, either before or after the removed child.
            // In the case of a single child existing in the column, no spacing element will be removed.
            if (childElement.previousSibling) {
                this._divElement.removeChild(childElement.previousSibling);
            } else if (childElement.nextSibling) {
                this._divElement.removeChild(childElement.nextSibling);
            }
        }
        this._divElement.removeChild(childElement);
    
        delete this._childIdToElementMap[child.renderId];
    },
    
    renderDispose: function(update) {
        WebCore.EventProcessor.removeAll(this._divElement);
        this._divElement = null;
        this._childIdToElementMap = null;
        this._spacingPrototype = null;
    },
    
    renderUpdate: function(update) {
        var fullRender = false;
        if (update.hasUpdatedProperties() || update.hasUpdatedLayoutDataChildren()) {
            // Full render
            fullRender = true;
        } else {
            var removedChildren = update.getRemovedChildren();
            if (removedChildren) {
                // Remove children.
                for (var i = 0; i < removedChildren.length; ++i) {
                    this._renderRemoveChild(update, removedChildren[i]);
                }
            }
            var addedChildren = update.getAddedChildren();
            if (addedChildren) {
                // Add children.
                for (var i = 0; i < addedChildren.length; ++i) {
                    this._renderAddChild(update, addedChildren[i], this.component.indexOf(addedChildren[i])); 
                }
            }
        }
        if (fullRender) {
            var element = this._divElement;
            var containerElement = element.parentNode;
            EchoRender.renderComponentDispose(update, update.parent);
            containerElement.removeChild(element);
            this.renderAdd(update, containerElement);
        }
        
        return fullRender;
    }
});
/**
 * Component rendering peer: Composite
 */
EchoAppRender.CompositeSync = Core.extend(EchoRender.ComponentSync, {

    $load: function() {
        EchoRender.registerPeer("Composite", this);
    },

    $virtual: {
        renderStyle: function(element) {
            EchoAppRender.Color.renderFB(this.component, element);
            EchoAppRender.Font.render(this.component.render("font"), element);
        }
    },
    
    renderAdd: function(update, parentElement) {
        this._divElement = document.createElement("div");
        this._divElement.id = this.component.renderId;
        
        var componentCount = this.component.getComponentCount();
        if (componentCount > 0) {
            this.renderStyle(this._divElement);
            for (var i = 0; i < componentCount; ++i) {
                var child = this.component.getComponent(i);
                EchoRender.renderComponentAdd(update, child, this._divElement);
            }
        }
        
        parentElement.appendChild(this._divElement);
    },
    
    renderDispose: function(update) { 
        this._divElement = null;
    },
    
    renderUpdate: function(update) {
        var element = this._divElement;
        var containerElement = element.parentNode;
        EchoRender.renderComponentDispose(update, update.parent);
        containerElement.removeChild(element);
        this.renderAdd(update, containerElement);
        return true;
    }
});

EchoRender.ComponentSync.Panel = Core.extend(EchoAppRender.CompositeSync, {
    $load: function() {
        EchoRender.registerPeer("Panel", this);
    },

    renderStyle: function(element) {
        EchoAppRender.CompositeSync.prototype.renderStyle.call(this, element);
        EchoAppRender.Border.render(this.component.render("border"), element);
        EchoAppRender.Insets.render(this.component.render("insets"), element, "padding");
    }
});
/**
 * Component rendering peer: ContentPane
 */
EchoAppRender.ContentPaneSync = Core.extend(EchoRender.ComponentSync, {

    $static: {
        DEFAULT_BACKGROUND: "#ffffff"
    },

    $load: function() {
        EchoRender.registerPeer("ContentPane", this);
    },

    $construct: function() {
        this._floatingPaneManager = null;
    },
    
    _processZIndexChanged: function(e) {
        for (var i = 0; i < this.component.children.length; ++i) {
            if (!this.component.children[i].floatingPane) {
                continue;
            }
            var index = this._floatingPaneManager.getIndex(this.component.children[i].renderId);
            var childElement = this._childIdToElementMap[this.component.children[i].renderId];
            if (childElement) {
                childElement.style.zIndex = index;
            }
        }
    },
    
    raise: function(child) {
        if (!this._floatingPaneManager) {
            this._floatingPaneManager = new EchoAppRender.FloatingPaneManager();
            this._floatingPaneManager.addZIndexListener(Core.method(this, this._processZIndexChanged));
        }
        this._floatingPaneManager.add(child.renderId);
    },
    
    renderAdd: function(update, parentElement) {
        this._divElement = document.createElement("div");
        this._divElement.id = this.component.renderId;
        this._divElement.style.position = "absolute";
        this._divElement.style.width = "100%";
        this._divElement.style.height = "100%";
        this._divElement.style.overflow = "hidden";
        this._divElement.style.zIndex = "0";
        EchoAppRender.Font.render(this.component.render("font"), this._divElement);
        EchoAppRender.Color.render(this.component.render("background", EchoAppRender.ContentPaneSync.DEFAULT_BACKGROUND),
                this._divElement, "backgroundColor");
        EchoAppRender.Color.render(this.component.render("foreground"), this._divElement, "color");
        EchoAppRender.FillImage.render(this.component.render("backgroundImage"), this._divElement); 
    
        this._childIdToElementMap = {};
        
        var componentCount = this.component.getComponentCount();
        for (var i = 0; i < componentCount; ++i) {
            var child = this.component.getComponent(i);
            this._renderAddChild(update, child);
        }
    
        // Store values of horizontal/vertical scroll such that 
        // renderDisplay() will adjust scrollbars appropriately after rendering.
        this._pendingScrollX = this.component.render("horizontalScroll");
        this._pendingScrollY = this.component.render("verticalScroll");
        
        parentElement.appendChild(this._divElement);
    },
    
    _renderAddChild: function(update, child) {
        var divElement = document.createElement("div");
        this._childIdToElementMap[child.renderId] = divElement;
        divElement.style.position = "absolute";
        if (child.floatingPane) {
            divElement.style.zIndex = "1";
        } else {
            var insets = this.component.render("insets", 0);
            var pixelInsets = EchoAppRender.Insets.toPixels(insets);
            divElement.style.zIndex = "0";
            divElement.style.left = pixelInsets.left + "px";
            divElement.style.top = pixelInsets.top + "px";
            divElement.style.bottom = pixelInsets.bottom + "px";
            divElement.style.right = pixelInsets.right + "px";
            divElement.style.overflow = "auto";
        }
        EchoRender.renderComponentAdd(update, child, divElement);
        this._divElement.appendChild(divElement);
        
        if (child.floatingPane) {
            this.raise(child);
        }
    },
    
    renderDispose: function(update) { 
        this._childIdToElementMap = null;
        this._divElement = null;
    },
    
    _renderRemoveChild: function(update, child) {
        if (child.floatingPane && this._floatingPaneManager) {
            this._floatingPaneManager.remove(child.renderId);
        }
        
        var divElement = this._childIdToElementMap[child.renderId];
        divElement.parentNode.removeChild(divElement);
        delete this._childIdToElementMap[child.renderId];
    },
    
    renderDisplay: function() {
        var child = this._divElement.firstChild;
        while (child) {
            WebCore.VirtualPosition.redraw(child);
            child = child.nextSibling;
        }
    
        // If a scrollbar adjustment has been requested by renderAdd, perform it.
        if (this._pendingScrollX || this._pendingScrollY) {
            var componentCount = this.component.getComponentCount();
            for (var i = 0; i < componentCount; ++i) {
                var child = this.component.getComponent(i);
                if (!child.floatingPane) {
                    var contentElement = this._childIdToElementMap[child.renderId];
                    if (this._pendingScrollX) {
                        contentElement.scrollLeft = parseInt(this._pendingScrollX < 0) ? 1000000 : this._pendingScrollX;
                        this._pendingScrollX = null;
                    }
                    if (this._pendingScrollY) {
                        contentElement.scrollTop = parseInt(this._pendingScrollY) < 0 ? 1000000 : this._pendingScrollY;
                        this._pendingScrollY = null;
                    }
                    break;
                }
            }
        }
    },
    
    renderUpdate: function(update) {
        var fullRender = false;
        if (update.hasUpdatedProperties() || update.hasUpdatedLayoutDataChildren()) {
            // Full render
            fullRender = true;
        } else {
            var removedChildren = update.getRemovedChildren();
            if (removedChildren) {
                // Remove children.
                for (var i = 0; i < removedChildren.length; ++i) {
                    this._renderRemoveChild(update, removedChildren[i]);
                }
            }
            var addedChildren = update.getAddedChildren();

            // FIXME experimental, nonfinal API
            update.renderContext.displayRequired = [];
            
            if (addedChildren) {
                // Add children.
                for (var i = 0; i < addedChildren.length; ++i) {
                    this._renderAddChild(update, addedChildren[i], this.component.indexOf(addedChildren[i]));
                    update.renderContext.displayRequired.push(addedChildren[i]); 
                }
            }
        }
        if (fullRender) {
            var element = this._divElement;
            var containerElement = element.parentNode;
            EchoRender.renderComponentDispose(update, update.parent);
            containerElement.removeChild(element);
            this.renderAdd(update, containerElement);
        }
        
        return fullRender;
    }
});
/**
 * Component rendering peer: Grid
 */
EchoAppRender.GridSync = Core.extend(EchoRender.ComponentSync, {

    $static: {

        _createPrototypeTable: function() {
            var tableElement = document.createElement("table");
            tableElement.style.outlineStyle = "none";
            tableElement.tabIndex = "-1";
            tableElement.style.borderCollapse = "collapse";
            
            var colGroupElement = document.createElement("colgroup");
            tableElement.appendChild(colGroupElement);
        
            var tbodyElement = document.createElement("tbody");
            tableElement.appendChild(tbodyElement);
            
            return tableElement;
        },
        
        Processor: Core.extend({
        
            $static: {
            
                Cell: Core.extend({
                    $construct: function(component, index, xSpan, ySpan) {
                        this.component = component;
                        this.index = index;
                        this.xSpan = xSpan;
                        this.ySpan = ySpan;
                    }
                })
            },
        
            $construct: function(grid) {
                this.grid = grid;
                this.cellArrays = [];
                this.horizontalOrientation = grid.render("orientation") != EchoApp.Grid.ORIENTATION_VERTICAL;
                
                var cells = this.createCells();
                if (cells == null) {
                    // Special case: empty Grid.
                    this.gridXSize = 0;
                    this.gridYSize = 0;
                    return;
                }
            
                this.renderCellMatrix(cells);
                
                this.calculateExtents();
                
                this.reduceY();
                this.reduceX();
            },
            
            calculateExtents: function() {
                var xProperty = this.horizontalOrientation ? "columnWidth" : "rowHeight";
                var yProperty = this.horizontalOrientation ? "rowHeight" : "columnWidth";
                
                this.xExtents = [];
                for (var i = 0; i < this.gridXSize; ++i) {
                    this.xExtents.push(this.grid.renderIndex(xProperty, i));
                }
            
                this.yExtents = [];
                for (var i = 0; i < this.gridYSize; ++i) {
                    this.yExtents.push(this.grid.renderIndex(yProperty, i));
                }
            },
            
            createCells: function() {
                var childCount = this.grid.getComponentCount();
                if (childCount == 0) {
                    // Abort if Grid is empty.
                    return null;
                }
                
                var cells = [];
                for (var i = 0; i < childCount; ++i) {
                    var child = this.grid.getComponent(i);
                    var layoutData = child.render("layoutData");
                    if (layoutData) {
                        var xSpan = this.horizontalOrientation ? layoutData.columnSpan : layoutData.rowSpan; 
                        var ySpan = this.horizontalOrientation ? layoutData.rowSpan : layoutData.columnSpan; 
                        cells.push(new EchoAppRender.GridSync.Processor.Cell(child, i, xSpan ? xSpan : 1, ySpan ? ySpan : 1));
                    } else {
                        cells.push(new EchoAppRender.GridSync.Processor.Cell(child, i, 1, 1));
                    }
                }
                return cells;
            },
            
            /**
             * Returns an array representing the cells at the specified y-index.
             * If no array currently exists, one is created.
             * 
             * @param {Integer} y the y-index
             * @return the array of cells.
             * @type {Array}
             */
            _getCellArray: function(y) {
                while (y >= this.cellArrays.length) {
                    this.cellArrays.push(new Array(this.gridXSize));
                }
                return this.cellArrays[y]; 
            },
            
            /**
             * Returns the number of columns that should be rendered.
             * 
             * @return the number of rendered columns
             * @type Integer
             */
            getColumnCount: function() {
                return this.horizontalOrientation ? this.gridXSize : this.gridYSize;
            },
            
            /**
             * Returns the cell that should be rendered at the
             * specified position.
             * 
             * @param {Integer} column the column index
             * @param {Integer} row the row index
             * @return the cell
             * @type EchoAppRender.GridSync.Processor.Cell
             */
            getCell: function(column, row) {
                if (this.horizontalOrientation) {
                    return this.cellArrays[row][column];
                } else {
                    return this.cellArrays[column][row];
                }
            },
            
            /**
             * Returns the number of rows that should be rendered.
             * 
             * @return the number of rendered rows
             * @type Integer
             */
            getRowCount: function() {
                return this.horizontalOrientation ? this.gridYSize : this.gridXSize;
            },
            
            /**
             * Remove duplicates from the x-axis where all cells simply
             * "span over" a given x-axis coordinate. 
             */
            reduceX: function() {
                // Determine duplicate cell sets on x-axis.
                var xRemoves = [];
                var x = 1;
                var length = this.cellArrays[0].length;
                while (x < length) {
                    var y = 0;
                    var identical = true;
                    while (y < this.cellArrays.length) {
                        if (this.cellArrays[y][x] != this.cellArrays[y][x - 1]) {
                            identical = false;
                            break;
                        }
                        ++y;
                    }
                    if (identical) {
                        xRemoves[x] = true;
                    }
                    ++x;
                }
                
                // If no reductions are necessary on the x-axis, do nothing.
                if (xRemoves.length == 0) {
                    return;
                }
                
                for (var removedX = this.gridXSize - 1; removedX >= 0; --removedX) {
                    if (!xRemoves[removedX]) {
                        continue;
                    }
                    
                    for (var y = 0; y < this.gridYSize; ++y) {
                        if (y == 0 || this.cellArrays[y][removedX - 1] != this.cellArrays[y - 1][removedX - 1]) {
                            // Reduce x-span, taking care not to reduce it multiple times if cell has a y-span.
                            if (this.cellArrays[y][removedX - 1] != null) {
                                --this.cellArrays[y][removedX - 1].xSpan;
                            }
                        }
                        for (x = removedX; x < this.gridXSize - 1; ++x) {
                            this.cellArrays[y][x] = this.cellArrays[y][x + 1];
                        }
                    }
                    
                    //FIXME. Add extent-size recalc.
                    
                    --this.gridXSize;
                }
            },
            
            /**
             * Remove duplicates from the y-axis where all cells simply
             * "span over" a given y-axis coordinate. 
             */
            reduceY: function() {
                var yRemoves = [];
                var y = 1;
                
                var size = this.cellArrays.length;
                var previousCellArray;
                var currentCellArray = this.cellArrays[0];
                
                while (y < size) {
                    previousCellArray = currentCellArray;
                    currentCellArray = this.cellArrays[y];
                    
                    var x = 0;
                    var identical = true;
                    
                    while (x < currentCellArray.length) {
                        if (currentCellArray[x] != previousCellArray[x]) {
                            identical = false;
                            break;
                        }
                        ++x;
                    }
                    if (identical) {
                        yRemoves[y] = true;
                    }
                    
                    ++y;
                }
                
                // If no reductions are necessary on the y-axis, do nothing.
                if (yRemoves.length == 0) {
                    return;
                }
                
                for (var removedY = this.gridYSize - 1; removedY >= 0; --removedY) {
                    if (!yRemoves[removedY]) {
                        continue;
                    }
                    
                    // Shorten the y-spans of the cell array that will be retained to 
                    // reflect the fact that a cell array is being removed.
                    var retainedCellArray = this.cellArrays[removedY - 1];
                    for (var x = 0; x < this.gridXSize; ++x) {
                        if (x == 0 || retainedCellArray[x] != retainedCellArray[x - 1]) {
                            // Reduce y-span, taking care not to reduce it multiple times if cell has an x-span.
                            if (retainedCellArray[x] != null) {
                                --retainedCellArray[x].ySpan;
                            }
                        }
                    }
                    
                    // Remove the duplicate cell array.
                    this.cellArrays.splice(removedY, 1);
                    
                    //FIXME. insert code here for extent adjustment.
                    
                    // Decrement the grid size to reflect cell array removal.
                    --this.gridYSize;
                }
            },
            
            renderCellMatrix: function(cells) {
                this.gridXSize = parseInt(this.grid.render("size", 2));
                var x = 0, y = 0;
                var yCells = this._getCellArray(y);
                
                for (var componentIndex = 0; componentIndex < cells.length; ++componentIndex) {
                    
                    // Set x-span to fill remaining size in the event SPAN_FILL has been specified or if the cell would
                    // otherwise extend past the specified size.
                    if (cells[componentIndex].xSpan == EchoApp.Grid.SPAN_FILL || cells[componentIndex].xSpan > this.gridXSize - x) {
                        cells[componentIndex].xSpan = this.gridXSize - x;
                    }
                    
                    // Set x-span of any cell INCORRECTLY set to negative value to 1 (note that SPAN_FILL has already been handled).
                    if (cells[componentIndex].xSpan < 1) {
                        cells[componentIndex].xSpan = 1;
                    }
                    // Set y-span of any cell INCORRECTLY set to negative value (or more likely SPAN_FILL) to 1.
                    if (cells[componentIndex].ySpan < 1) {
                        cells[componentIndex].ySpan = 1;
                    }
                    
                    if (cells[componentIndex].xSpan != 1 || cells[componentIndex].ySpan != 1) {
                        // Scan to ensure no y-spans are blocking this x-span.
                        // If a y-span is blocking, shorten the x-span to not
                        // interfere.
                        for (var xIndex = 1; xIndex < cells[componentIndex].xSpan; ++xIndex) {
                            if (yCells[x + xIndex] != null) {
                                // Blocking component found.
                                cells[componentIndex].xSpan = xIndex;
                                break;
                            }
                        }
                        for (var yIndex = 0; yIndex < cells[componentIndex].ySpan; ++yIndex) {
                            var yIndexCells = this._getCellArray(y + yIndex);
                            for (var xIndex = 0; xIndex < cells[componentIndex].xSpan; ++xIndex) {
                                yIndexCells[x + xIndex] = cells[componentIndex];
                            }
                        }
                    }
                    yCells[x] = cells[componentIndex];
            
                    if (componentIndex < cells.length - 1) {
                        // Move rendering cursor.
                        var nextRenderPointFound = false;
                        while (!nextRenderPointFound) {
                            if (x < this.gridXSize - 1) {
                                ++x;
                            } else {
                                // Move cursor to next line.
                                x = 0;
                                ++y;
                                yCells = this._getCellArray(y);
                                
                            }
                            nextRenderPointFound = yCells[x] == null;
                        }
                    }
                }
            
                // Store actual 'y' dimension.
                this.gridYSize = this.cellArrays.length;
            }
        })
    },
    
    $load: function() {
        this._prototypeTable = this._createPrototypeTable();
        EchoRender.registerPeer("Grid", this);
    },
    
    _columnCount: null,
    
    _rowCount: null,
    
    _processKeyPress: function(e) { 
        switch (e.keyCode) {
        case 37:
        case 39:
            var focusPrevious = e.keyCode == 37;
            var focusedComponent = this.component.application.getFocusedComponent();
            if (focusedComponent && focusedComponent.peer && focusedComponent.peer.getFocusFlags) {
                var focusFlags = focusedComponent.peer.getFocusFlags();
                if ((focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_LEFT)
                        || (!focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_RIGHT)) {
                    var focusChild = this.component.application.focusManager.findInParent(this.component, focusPrevious);
                    if (focusChild) {
                        this.component.application.setFocusedComponent(focusChild);
                        WebCore.DOM.preventEventDefault(e);
                        return false;
                    }
                }
            }
            break;
        case 38:
        case 40:
            var focusPrevious = e.keyCode == 38;
            var focusedComponent = this.component.application.getFocusedComponent();
            if (focusedComponent && focusedComponent.peer && focusedComponent.peer.getFocusFlags) {
                var focusFlags = focusedComponent.peer.getFocusFlags();
                if ((focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_UP)
                        || (!focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_DOWN)) {
                    var focusChild = this.component.application.focusManager.findInParent(this.component, focusPrevious,
                            this._columnCount);
                    if (focusChild) {
                        this.component.application.setFocusedComponent(focusChild);
                        WebCore.DOM.preventEventDefault(e);
                        return false;
                    }
                }
            }
            break;
        }
        return true;
    },

    renderAdd: function(update, parentElement) {
        var gridProcessor = new EchoAppRender.GridSync.Processor(this.component);
        
        this._columnCount = gridProcessor.getColumnCount();
        this._rowCount = gridProcessor.getRowCount();
        
        var defaultInsets = this.component.render("insets", "0");
        var defaultBorder = this.component.render("border", "");
    
        this._tableElement = EchoAppRender.GridSync._prototypeTable.cloneNode(true);
        this._tableElement.id = this.component.renderId;
        
        EchoAppRender.Color.renderFB(this.component, this._tableElement);
        EchoAppRender.Border.render(defaultBorder, this._tableElement);
        EchoAppRender.Font.render(this.component.render("font"), this._tableElement);
        EchoAppRender.Insets.render(this.component.render("insets"), this._tableElement, "padding");
    
        var width = this.component.render("width");
        
        if (width && WebCore.Environment.QUIRK_IE_TABLE_PERCENT_WIDTH_SCROLLBAR_ERROR && EchoAppRender.Extent.isPercent(width)) {
            this._renderPercentWidthByMeasure = parseInt(width);
            width = null;
        }
        
        if (width) {
            if (EchoAppRender.Extent.isPercent(width)) {
                this._tableElement.style.width = width;
            } else {
                this._tableElement.style.width = EchoAppRender.Extent.toCssValue(width, true);
            }
        }
        
        var height = this.component.render("height");
        if (height) {
            if (EchoAppRender.Extent.isPercent(height)) {
                this._tableElement.style.height = height;
            } else {
                this._tableElement.style.height = EchoAppRender.Extent.toCssValue(height, false);
            }
        }
        
        var colGroupElement = this._tableElement.firstChild;
        for (var columnIndex = 0; columnIndex < this._columnCount; ++columnIndex) {
            var colElement = document.createElement("col");
            var width = gridProcessor.xExtents[columnIndex];
            if (width != null) {
                if (EchoAppRender.Extent.isPercent(width)) {
                    colElement.width = width.toString();
                } else {
                    colElement.width = EchoAppRender.Extent.toCssValue(width, true);
                }
            }
            colGroupElement.appendChild(colElement);
        }
        
        var tbodyElement = colGroupElement.nextSibling;
        
        var size = parseInt(this.component.render("size", 2));
        
        var trElement;
        var height;
        var renderedComponentIds = {};
        
        var xSpan, ySpan;
        if (gridProcessor.horizontalOrientation) {
            xSpan = "colSpan";
            ySpan = "rowSpan"; 
        } else {
            xSpan = "rowSpan";
            ySpan = "colSpan"; 
        }
        
        var tdPrototype = document.createElement("td");
        EchoAppRender.Border.render(defaultBorder, tdPrototype);
        tdPrototype.style.padding = defaultInsets.toString();
        tdPrototype.style.overflow = "hidden";
        
        for (var rowIndex = 0; rowIndex < this._rowCount; ++rowIndex) {
            trElement = document.createElement("tr");
            height = gridProcessor.yExtents[rowIndex];
            if (height) {
                trElement.style.height = EchoAppRender.Extent.toCssValue(height, false);
            }
            tbodyElement.appendChild(trElement);
            
            for (var columnIndex = 0; columnIndex < this._columnCount; ++columnIndex) {
                var cell = gridProcessor.getCell(columnIndex, rowIndex);
                if (cell == null) {
                    var tdElement = document.createElement("td");
                    trElement.appendChild(tdElement);
                    continue;
                }
                if (renderedComponentIds[cell.component.renderId]) {
                    // Cell already rendered.
                    continue;
                }
                renderedComponentIds[cell.component.renderId] = true;
                
                var tdElement = tdPrototype.cloneNode(false);
                
                if (cell.xSpan > 1) {
                    tdElement.setAttribute(xSpan, cell.xSpan);
                }
                if (cell.ySpan > 1) {
                    tdElement.setAttribute(ySpan, cell.ySpan);
                }
                
                var layoutData = cell.component.render("layoutData");
                if (layoutData) {
                    EchoAppRender.Insets.render(layoutData.insets, tdElement, "padding");
                    EchoAppRender.Alignment.render(layoutData.alignment, tdElement, true, this.component);
                    EchoAppRender.FillImage.render(layoutData.backgroundImage, tdElement);
                    EchoAppRender.Color.render(layoutData.background, tdElement, "backgroundColor");
                }
                
                EchoRender.renderComponentAdd(update, cell.component, tdElement);
    
                trElement.appendChild(tdElement);
            }
        }
        
        WebCore.EventProcessor.add(this._tableElement, 
                WebCore.Environment.QUIRK_IE_KEY_DOWN_EVENT_REPEAT ? "keydown" : "keypress",
                Core.method(this, this._processKeyPress), false);

        parentElement.appendChild(this._tableElement);
    },
    
    renderDispose: function(update) {
        WebCore.EventProcessor.removeAll(this._tableElement);
        this._tableElement = null;
    },
    
    renderDisplay: function() {
        if (this._renderPercentWidthByMeasure) {
            this._tableElement.style.width = "";
            var percentWidth = (this._tableElement.parentNode.offsetWidth * this._renderPercentWidthByMeasure) / 100;
            this._tableElement.style.width = percentWidth + "px";
        }
    },
    
    renderUpdate: function(update) {
        var element = this._tableElement;
        var containerElement = element.parentNode;
        EchoRender.renderComponentDispose(update, update.parent);
        containerElement.removeChild(element);
        this.renderAdd(update, containerElement);
        return true;
    }
});
/**
 * Component rendering peer: Label
 */
EchoAppRender.LabelSync = Core.extend(EchoRender.ComponentSync, { 

    $static: {
    
       _defaultIconTextMargin: 5
    },
    
    $load: function() {
        EchoRender.registerPeer("Label", this);
    },
    
    /**
     * Formats the whitespace in the given text for use in HTML.
     * 
     * @param text {String} the text to format
     * @param parentElement the element to append the text to
     */
    _formatWhitespace: function(text, parentElement) {
        // switch between spaces and non-breaking spaces to preserve line wrapping
        text = text.replace(/\t/g, " \u00a0 \u00a0");
        text = text.replace(/ {2}/g, " \u00a0");
        var lines = text.split('\n');
        for (var i = 0; i < lines.length; i++) {
            var line = lines[i];
            if (i > 0) {
                parentElement.appendChild(document.createElement("br"));
            }
            if (line.length > 0) {
                parentElement.appendChild(document.createTextNode(line));
            }
        }
    },
    
    renderAdd: function(update, parentElement) {
        this._containerElement = parentElement;
        var icon = this.component.render("icon");
        var text = this.component.render("text");
        var foreground = this.component.render("foreground");
        var background = this.component.render("background");
    
        if (text) {
            var lineWrap = this.component.render("lineWrap", true);
            var formatWhitespace = this.component.render("formatWhitespace", false)
                    && (text.indexOf(' ') != -1 || text.indexOf('\n') != -1 || text.indexOf('\t') != -1);
            
            if (icon) {
                // Text and icon.
                var iconTextMargin = this.component.render("iconTextMargin", 
                        EchoAppRender.LabelSync._defaultIconTextMargin);
                var orientation = EchoAppRender.TriCellTable.getOrientation(this.component, "textPosition");
                var tct = new EchoAppRender.TriCellTable(orientation, EchoAppRender.Extent.toPixels(iconTextMargin));
                var imgElement = document.createElement("img");
                EchoAppRender.ImageReference.renderImg(icon, imgElement);
                if (formatWhitespace) {
                    this._formatWhitespace(text, tct.tdElements[0]);
                } else {
                    tct.tdElements[0].appendChild(document.createTextNode(text));
                }
                if (!lineWrap) {
                    tct.tdElements[0].style.whiteSpace = "nowrap";
                }
                tct.tdElements[1].appendChild(imgElement);
                this._labelNode = tct.tableElement;
                this._labelNode.id = this.component.renderId;
                EchoAppRender.Font.render(this.component.render("font"), this._labelNode);
                EchoAppRender.Color.renderFB(this.component, this._labelNode);
            } else {
                // Text without icon.
                var font = this.component.render("font");
                if (!font && lineWrap && !foreground && !background && !formatWhitespace) {
                    this._labelNode = document.createTextNode(text);
                } else {
                    this._labelNode = document.createElement("span");
                    this._labelNode.id = this.component.renderId;
                    if (formatWhitespace) {
                        this._formatWhitespace(text, this._labelNode);
                    } else {
                        this._labelNode.appendChild(document.createTextNode(text));
                    }
                    if (!lineWrap) {
                        this._labelNode.style.whiteSpace = "nowrap";
                    }
                    EchoAppRender.Font.render(font, this._labelNode);
                    EchoAppRender.Color.renderFB(this.component, this._labelNode);
                }
            }
        } else if (icon) {
            var imgElement = document.createElement("img");
            EchoAppRender.ImageReference.renderImg(icon, imgElement);
            this._labelNode = document.createElement("span");
            this._labelNode.id = this.component.renderId;
            this._labelNode.appendChild(imgElement);
            EchoAppRender.Color.renderFB(this.component, this._labelNode); // should be BG only.
        } else {
            // Neither icon nor text, render blank.
            this._labelNode = null;
        }
    
        if (this._labelNode) {
            parentElement.appendChild(this._labelNode);
        }
    },
    
    renderDispose: function(update) {
        this._containerElement = null;
        this._labelNode = null;
    },
    
    renderUpdate: function(update) {
        if (this._labelNode) {
            this._labelNode.parentNode.removeChild(this._labelNode);
        }
        // Note: this.renderDispose() is not invoked (it does nothing).
        this.renderAdd(update, this._containerElement);
        return false; // Child elements not supported: safe to return false.
    }
});
/**
 * Abstract base class for SELECT-element based list components.
 */
EchoAppRender.ListComponentSync = Core.extend(EchoRender.ComponentSync, {

    $static: {
        DEFAULT_DIV_HEIGHT: "6em",
        DEFAULT_DIV_BORDER: "1px solid",
        DEFAULT_SELECTED_BACKGROUND: "#0a246a",
        DEFAULT_SELECTED_FOREGROUND: "#ffffff"
    },

    $abstract: true,

    _hasRenderedSelectedItems: false,
    
    _multipleSelect: false,

    /**
     * Flag indicating that component will be rendered as a DHTML-based ListBox.
     * This form of rendering is necessary on Internet Explorer 6 browsers due to unstable
     * code in this web browser when using listbox-style SELECT elements.
     */
    _alternateRender: false,
    
    _mainElement: null,
    
    _divElement: null,
    
    /**
     * Processes a click event.
     * This event handler is registered only in the case of the "alternate" DHTML-based rendered
     * listbox for IE6, i.e., the _alternateRender flag will be true. 
     */
    _processClick: function(e) {
        if (!this.client.verifyInput(this.component)) {
            WebCore.DOM.preventEventDefault(e);
            this._renderSelection();
            return;
        }
        
        var child = this._divElement.firstChild;
        var i = 0;
        while (child) {
            if (child == e.target) {
                break;
            }
            child = child.nextSibling;
            ++i;
        }
        if (child == null) {
            return;
        }
        
        if (this._multipleSelect && e.ctrlKey) {
            // Multiple selection and user has pressed ctrl key to select multiple items.
            var selection = this.component.get("selection");
            if (!selection) {
                selection = [];
            }
            var arrayIndex = Core.Arrays.indexOf(selection, i); 
            if (arrayIndex == -1) {
                // Add item to selection if not present.
                selection.push(i);
            } else {
                // Remove item from selection if already present.
                selection.splice(arrayIndex, 1);
            }
        } else {
            selection = [i];
        }
        
        this.component.set("selection", selection);
        this.component.doAction();

        this._renderSelection();
    },
    
    _processChange: function(e) {
        if (!this.client.verifyInput(this.component)) {
            WebCore.DOM.preventEventDefault(e);
            this._renderSelection();
            return;
        }
        
        var selection = [];
        if (this._multipleSelect) {
            for (var i = 0; i < this._mainElement.options.length; ++i) {
                if (this._mainElement.options[i].selected) {
                    selection.push(i);
                }
            }
        } else {
            if (this._mainElement.selectedIndex != -1) {
                selection.push(this._mainElement.selectedIndex);
            }
        }
    
        this.component.set("selection", selection);
        this.component.doAction();
    },
    
    _processSelectStart: function(e) {
        WebCore.DOM.preventEventDefault(e);
    },

    _renderMainAsSelect: function(update, parentElement, size) {
        this._mainElement = document.createElement("select");
        this._mainElement.id = this.component.renderId;
        this._mainElement.size = size;

        if (!this._enabled) {
            this._mainElement.disabled = true;
        }
        if (this._multipleSelect) {
            this._mainElement.multiple = "multiple";
        }

        EchoAppRender.Border.render(
                EchoAppRender.getEffectProperty(this.component, "border", "disabledBorder", !this._enabled), 
                this._mainElement);
        EchoAppRender.Color.render(
                EchoAppRender.getEffectProperty(this.component, "foreground", "disabledForeground", !this._enabled), 
                this._mainElement, "color");
        EchoAppRender.Color.render(
                EchoAppRender.getEffectProperty(this.component, "background", "disabledBackground", !this._enabled), 
                this._mainElement, "backgroundColor");
        EchoAppRender.Font.render(
                EchoAppRender.getEffectProperty(this.component, "font", "disabledFont", !this._enabled), 
                this._mainElement);
        EchoAppRender.Insets.render(this.component.render("insets"), this._mainElement, "padding");

        if (this.component.items) {
            for (var i = 0; i < this.component.items.length; ++i) {
                var optionElement = document.createElement("option");
                optionElement.appendChild(document.createTextNode(this.component.items[i].toString()));
                if (this.component.items[i].foreground) {
                    EchoAppRender.Color.render(this.component.items[i].foreground, optionElement, "color");
                }
                if (this.component.items[i].background) {
                    EchoAppRender.Color.render(this.component.items[i].background, optionElement, "backgroundColor");
                }
                if (this.component.items[i].font) {
                    EchoAppRender.Font.render(this.component.items[i].font, optionElement);
                }
                this._mainElement.appendChild(optionElement);
            }
        }
    
        if (this._enabled) {
            WebCore.EventProcessor.add(this._mainElement, "change", Core.method(this, this._processChange), false);
        }

        parentElement.appendChild(this._mainElement);
    },

    _renderMainAsDiv: function(update, parentElement, size) {
        this._mainElement = document.createElement("table");
        this._mainElement.id = this.component.renderId;
        
        var tbodyElement = document.createElement("tbody");
        this._mainElement.appendChild(tbodyElement);
        var trElement = document.createElement("tr");
        tbodyElement.appendChild(trElement);
        var tdElement = document.createElement("td");
        trElement.appendChild(tdElement);
        
        this._divElement = document.createElement("div");
        tdElement.appendChild(this._divElement);
        
        this._divElement.style.cssText = "cursor:default;overflow:auto;";
        
        //FIXME        
        this._divElement.style.height = "6em";
        
        EchoAppRender.Border.render(
                EchoAppRender.getEffectProperty(this.component, "border", "disabledBorder", !this._enabled, 
                EchoAppRender.ListComponentSync.DEFAULT_DIV_BORDER, EchoAppRender.ListComponentSync.DEFAULT_DIV_BORDER), 
                this._divElement);
        EchoAppRender.Color.render(
                EchoAppRender.getEffectProperty(this.component, "foreground", "disabledForeground", !this._enabled), 
                this._divElement, "color");
        EchoAppRender.Color.render(
                EchoAppRender.getEffectProperty(this.component, "background", "disabledBackground", !this._enabled), 
                this._divElement, "backgroundColor");
        EchoAppRender.Font.render(
                EchoAppRender.getEffectProperty(this.component, "font", "disabledFont", !this._enabled), 
                this._divElement);
        EchoAppRender.Insets.render(this.component.render("insets"), this._divElement, "padding");

        if (this.component.items) {
            for (var i = 0; i < this.component.items.length; ++i) {
                var optionElement = document.createElement("div");
                optionElement.appendChild(document.createTextNode(this.component.items[i].toString()));
                if (this.component.items[i].foreground) {
                    EchoAppRender.Color.render(this.component.items[i].foreground, optionElement, "color");
                }
                if (this.component.items[i].background) {
                    EchoAppRender.Color.render(this.component.items[i].background, optionElement, "backgroundColor");
                }
                if (this.component.items[i].font) {
                    EchoAppRender.Font.render(this.component.items[i].font, optionElement);
                }
                this._divElement.appendChild(optionElement);
            }
        }
        
        if (this._enabled) {
            WebCore.EventProcessor.add(this._divElement, "click", Core.method(this, this._processClick), false);
            WebCore.EventProcessor.add(this._divElement, "selectstart", Core.method(this, this._processSelectStart), false);
        }
        
        parentElement.appendChild(this._mainElement);
    },
    
    _renderMain: function(update, parentElement, size) {
        this._multipleSelect = this.component.get("selectionMode") == EchoApp.ListBox.MULTIPLE_SELECTION;
        if (this.component instanceof EchoApp.ListBox && WebCore.Environment.QUIRK_IE_SELECT_LIST_DOM_UPDATE) {
            this._alternateRender = true;
        }
        this._enabled = this.component.isRenderEnabled();
        
        if (this._alternateRender) {
            this._renderMainAsDiv(update, parentElement, size);
        } else {
            this._renderMainAsSelect(update, parentElement, size);
        }
    },
    
    renderDisplay: function() {
        this._renderSelection();
    },
    
    renderDispose: function(update) { 
        WebCore.EventProcessor.removeAll(this._mainElement);
        this._mainElement = null;
    },
    
    _renderSelection: function() {
        // Set selection.
        var selection = this.component.get("selection");
        
        if (!selection) {
            selection = this._multipleSelect ? [] : [0];
        }

        if (this._alternateRender) {
            if (this._hasRenderedSelectedItems) {
                for (var i = 0; i < this.component.items.length; ++i) {
                    EchoAppRender.Color.renderClear(this.component.items[i].foreground, this._divElement.childNodes[i], 
                            "color");
                    EchoAppRender.Color.renderClear(this.component.items[i].background, this._divElement.childNodes[i], 
                            "backgroundColor");
                }
            }
            for (var i = 0; i < selection.length; ++i) {
                if (selection[i] >= 0 && selection[i] < this._divElement.childNodes.length) {
                    EchoAppRender.Color.render(EchoAppRender.ListComponentSync.DEFAULT_SELECTED_FOREGROUND,
                            this._divElement.childNodes[selection[i]], "color");
                    EchoAppRender.Color.render(EchoAppRender.ListComponentSync.DEFAULT_SELECTED_BACKGROUND,
                            this._divElement.childNodes[selection[i]], "backgroundColor");
                }
            }
        } else {
            if (this._hasRenderedSelectedItems) {
                this._mainElement.selectedIndex = -1;
            }
            for (var i = 0; i < selection.length; ++i) {
                if (selection[i] >= 0 && selection[i] < this._mainElement.options.length) {
                    this._mainElement.options[selection[i]].selected = true;
                }
            }
        }
        this._hasRenderedSelectedItems = true;
    },
    
    renderUpdate: function(update) {
        var element = this._mainElement;
        var containerElement = element.parentNode;
        this.renderDispose(update);
        containerElement.removeChild(element);
        this.renderAdd(update, containerElement);
        return false; // Child elements not supported: safe to return false.
    }
});

/**
 * Component rendering peer: ListBox
 */
EchoAppRender.ListBoxSync = Core.extend(EchoAppRender.ListComponentSync, {

    $load: function() {
        EchoRender.registerPeer("ListBox", this);
    },

    renderAdd: function(update, parentElement) {
        this._renderMain(update, parentElement, 6);
    }
});

/**
 * Component rendering peer: SelectField
 */
EchoAppRender.SelectFieldSync = Core.extend(EchoAppRender.ListComponentSync, { 

    $load: function() {
        EchoRender.registerPeer("SelectField", this);
    },
    
    renderAdd: function(update, parentElement) {
        this._renderMain(update, parentElement, 0);
    }
});
/**
 * Component rendering peer: Row
 */
EchoAppRender.RowSync = Core.extend(EchoRender.ComponentSync, {
    
    $static: {
    
        _createRowPrototype: function() {
            var divElement = document.createElement("div");
            divElement.style.outlineStyle = "none";
            divElement.style.overflow = "hidden";
            divElement.tabIndex = "-1";
        
            var tableElement = document.createElement("table");
            tableElement.style.borderCollapse = "collapse";
            divElement.appendChild(tableElement);
        
            var tbodyElement = document.createElement("tbody");
            tableElement.appendChild(tbodyElement);
            
            var trElement = document.createElement("tr");
            tbodyElement.appendChild(trElement);
        
            return divElement;
        }
    },
    
    $load: function() {
        this._rowPrototype = this._createRowPrototype();
        EchoRender.registerPeer("Row", this);
    },

    _processKeyPress: function(e) { 
        switch (e.keyCode) {
        case 37:
        case 39:
            var focusPrevious = e.keyCode == 37;
            var focusedComponent = this.component.application.getFocusedComponent();
            if (focusedComponent && focusedComponent.peer && focusedComponent.peer.getFocusFlags) {
                var focusFlags = focusedComponent.peer.getFocusFlags();
                if ((focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_LEFT)
                        || (!focusPrevious && focusFlags & EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_RIGHT)) {
                    var focusChild = this.component.application.focusManager.findInParent(this.component, focusPrevious);
                    if (focusChild) {
                        this.component.application.setFocusedComponent(focusChild);
                        WebCore.DOM.preventEventDefault(e);
                        return false;
                    }
                }
            }
            break;
        }
        return true;
    },
    
    renderAdd: function(update, parentElement) {
        this._divElement = EchoAppRender.RowSync._rowPrototype.cloneNode(true);
        this._divElement.id = this.component.renderId;
        
        EchoAppRender.Border.render(this.component.render("border"), this._divElement);
        EchoAppRender.Color.renderFB(this.component, this._divElement);
        EchoAppRender.Font.render(this.component.render("font"), this._divElement);
        EchoAppRender.Insets.render(this.component.render("insets"), this._divElement, "padding");
        EchoAppRender.Alignment.render(this.component.render("alignment"), this._divElement, true, this.component);
        
        //                div              table      tbody      tr
        this._trElement = this._divElement.firstChild.firstChild.firstChild;
    
        this._cellSpacing = EchoAppRender.Extent.toPixels(this.component.render("cellSpacing"), false);
        if (this._cellSpacing) {
            this._spacingPrototype = document.createElement("td");
            this._spacingPrototype.style.width = this._cellSpacing + "px";
        }
        
        this._childIdToElementMap = {};
    
        var componentCount = this.component.getComponentCount();
        for (var i = 0; i < componentCount; ++i) {
            var child = this.component.getComponent(i);
            this._renderAddChild(update, child);
        }
        
        WebCore.EventProcessor.add(this._divElement, 
                WebCore.Environment.QUIRK_IE_KEY_DOWN_EVENT_REPEAT ? "keydown" : "keypress",
                Core.method(this, this._processKeyPress), false);
        
        parentElement.appendChild(this._divElement);
    },
    
    _renderAddChild: function(update, child, index) {
        var tdElement = document.createElement("td");
        this._childIdToElementMap[child.renderId] = tdElement;
        EchoRender.renderComponentAdd(update, child, tdElement);
    
        var layoutData = child.render("layoutData");
        var insets;
        if (layoutData) {
            insets = layoutData.insets;
            EchoAppRender.Color.render(layoutData.background, tdElement, "backgroundColor");
            EchoAppRender.FillImage.render(layoutData.backgroundImage, tdElement);
            EchoAppRender.Alignment.render(layoutData.alignment, tdElement, true, this.component);
            if (layoutData.width) {
                if (EchoAppRender.Extent.isPercent(layoutData.width)) {
                    tdElement.style.width = layoutData.width;
                } else {
                    tdElement.style.width = EchoAppRender.Extent.cssValue(layoutData.width, true);
                }
            }
        }
        if (!insets) {
            insets = "0px";
        }
        EchoAppRender.Insets.render(insets, tdElement, "padding");
        
        if (index != null) {
            var currentChildCount;
            if (this._trElement.childNodes.length >= 3 && this._cellSpacing) {
                currentChildCount = (this._trElement.childNodes.length + 1) / 2;
            } else {
                currentChildCount = this._trElement.childNodes.length;
            }
            if (index == currentChildCount) {
                index = null;
            }
        }
        if (index == null) {
            // Full render or append-at-end scenario
            
            // Render spacing td first if index != 0 and cell spacing enabled.
            if (this._cellSpacing && this._trElement.firstChild) {
                this._trElement.appendChild(this._spacingPrototype.cloneNode(false));
            }
    
            // Render child td second.
            this._trElement.appendChild(tdElement);
        } else {
            // Partial render insert at arbitrary location scenario (but not at end)
            var insertionIndex = this._cellSpacing ? index * 2 : index;
            var beforeElement = this._trElement.childNodes[insertionIndex];
            
            // Render child td first.
            this._trElement.insertBefore(tdElement, beforeElement);
            
            // Then render spacing td if required.
            if (this._cellSpacing) {
                this._trElement.insertBefore(this._spacingPrototype.cloneNode(false), beforeElement);
            }
        }
    },
    
    _renderRemoveChild: function(update, child) {
        var childElement = this._childIdToElementMap[child.renderId];
        
        if (this._cellSpacing) {
            // If cell spacing is enabled, remove a spacing element, either before or after the removed child.
            // In the case of a single child existing in the Row, no spacing element will be removed.
            if (childElement.previousSibling) {
                this._trElement.removeChild(childElement.previousSibling);
            } else if (childElement.nextSibling) {
                this._trElement.removeChild(childElement.nextSibling);
            }
        }
        this._trElement.removeChild(childElement);
    
        delete this._childIdToElementMap[child.renderId];
    },
    
    renderDispose: function(update) { 
        WebCore.EventProcessor.removeAll(this._divElement);
        this._divElement = null;
        this._trElement = null;
        this._childIdToElementMap = null;
        this._spacingPrototype = null;
    },
    
    renderUpdate: function(update) {
        var fullRender = false;
        if (update.hasUpdatedProperties() || update.hasUpdatedLayoutDataChildren()) {
            // Full render
            fullRender = true;
        } else {
            var removedChildren = update.getRemovedChildren();
            if (removedChildren) {
                // Remove children.
                for (var i = 0; i < removedChildren.length; ++i) {
                    var child = removedChildren[i];
                    this._renderRemoveChild(update, child);
                }
            }
            var addedChildren = update.getAddedChildren();
            if (addedChildren) {
                // Add children.
                for (var i = 0; i < addedChildren.length; ++i) {
                    this._renderAddChild(update, addedChildren[i], this.component.indexOf(addedChildren[i])); 
                }
            }
        }
        if (fullRender) {
            var element = this._divElement;
            var containerElement = element.parentNode;
            EchoRender.renderComponentDispose(update, update.parent);
            containerElement.removeChild(element);
            this.renderAdd(update, containerElement);
        }
        
        return fullRender;
    }
});
/**
 * Component rendering peer: TextComponent
 */
EchoAppRender.TextComponentSync = Core.extend(EchoRender.ComponentSync, {
    
    $abstract: true,
    
    $static: {
        _supportedPartialProperties: ["text"]
    },
    
    $virtual: {
        
        sanitizeInput: function() {
            var maximumLength = this.component.render("maximumLength", -1);
            if (maximumLength >= 0) {
                if (this._textComponentElement.value && this._textComponentElement.value.length > maximumLength) {
                    this._textComponentElement.value = this._textComponentElement.value.substring(0, maximumLength);
                }
            }
        }
    },
    
    _renderStyle: function() {
        EchoAppRender.Border.render(this.component.render("border"), this._textComponentElement);
        EchoAppRender.Color.renderFB(this.component, this._textComponentElement);
        EchoAppRender.Font.render(this.component.render("font"), this._textComponentElement);
        EchoAppRender.Insets.render(this.component.render("insets"), this._textComponentElement, "padding");
        EchoAppRender.FillImage.render(this.component.render("backgroundImage"), this._textComponentElement);
        var width = this.component.render("width");
        if (width) {
            this._textComponentElement.style.width = width.toString();
        }
        var height = this.component.render("height");
        if (height) {
            this._textComponentElement.style.height = height.toString();
        }
        var toolTipText = this.component.render("toolTipText");
        if (toolTipText) {
            this._textComponentElement.title = toolTipText;
        }
    },
    
    _addEventHandlers: function() {
        WebCore.EventProcessor.add(this._textComponentElement, "click", Core.method(this, this._processClick), false);
        WebCore.EventProcessor.add(this._textComponentElement, "blur", Core.method(this, this._processBlur), false);
        WebCore.EventProcessor.add(this._textComponentElement, "keypress", Core.method(this, this._processKeyPress), false);
        WebCore.EventProcessor.add(this._textComponentElement, "keyup", Core.method(this, this._processKeyUp), false);
    },
    
    renderDispose: function(update) {
        WebCore.EventProcessor.removeAll(this._textComponentElement);
        this._textComponentElement = null;
    },
    
    _processBlur: function(e) {
        if (!this.client.verifyInput(this.component, EchoClient.FLAG_INPUT_PROPERTY)) {
            return;
        }
        this.sanitizeInput();
        this.component.set("text", e.registeredTarget.value);
    },
    
    _processClick: function(e) {
        if (!this.client.verifyInput(this.component, EchoClient.FLAG_INPUT_PROPERTY)) {
            return;
        }
        this.component.application.setFocusedComponent(this.component);
    },

    _processKeyPress: function(e) {
        if (!this.client.verifyInput(this.component, EchoClient.FLAG_INPUT_PROPERTY)) {
            WebCore.DOM.preventEventDefault(e);
            return true;
        }
    },
    
    _processKeyUp: function(e) {
        if (!this.client.verifyInput(this.component, EchoClient.FLAG_INPUT_PROPERTY)) {
            WebCore.DOM.preventEventDefault(e);
            return true;
        }
        this.sanitizeInput();
        
        // Store last updated text in local value, to ensure that we do not attempt to
        // reset it to this value in renderUpdate() and miss any characters that were
        // typed between repaints.
        this._text = e.registeredTarget.value;
        
        this.component.set("text", this._text);
        if (e.keyCode == 13) {
            //FIXME fire from component.
            this.component.fireEvent({type: "action", source: this.component});
        }
        return true;
    },
    
    renderFocus: function() {
        WebCore.DOM.focusElement(this._textComponentElement);
    },
    
    renderUpdate: function(update) {
        var fullRender =  !Core.Arrays.containsAll(EchoAppRender.TextComponentSync._supportedPartialProperties, 
                    update.getUpdatedPropertyNames(), true);
    
        if (fullRender) {
            var element = this._textComponentElement;
            var containerElement = element.parentNode;
            this.renderDispose(update);
            containerElement.removeChild(element);
            this.renderAdd(update, containerElement);
        } else {
            if (update.hasUpdatedProperties()) {
                var textUpdate = update.getUpdatedProperty("text");
                if (textUpdate && textUpdate.newValue != this._text) {
                    this._textComponentElement.value = textUpdate.newValue;
                }
            }
        }
        
        return false; // Child elements not supported: safe to return false.
    }
});

/**
 * Component rendering peer: TextArea
 */
EchoAppRender.TextAreaSync = Core.extend(EchoAppRender.TextComponentSync, {

    $load: function() {
        EchoRender.registerPeer("TextArea", this);
    },

    renderAdd: function(update, parentElement) {
        this._textComponentElement = document.createElement("textarea");
        this._textComponentElement.id = this.component.renderId;
        this._renderStyle(this._textComponentElement);
        this._textComponentElement.style.overflow = "auto";
        this._addEventHandlers(this._textComponentElement);
        if (this.component.get("text")) {
            this._textComponentElement.appendChild(document.createTextNode(this.component.get("text")));
        } else {
            this._textComponentElement.appendChild(document.createTextNode(""));
        }
        parentElement.appendChild(this._textComponentElement);
    }
});

/**
 * Component rendering peer: TextField
 */
EchoAppRender.TextFieldSync = Core.extend(EchoAppRender.TextComponentSync, {
    
    $load: function() {
        EchoRender.registerPeer("TextField", this);
    },
    
    $virtual: {
        _type: "text"
    },

    getFocusFlags: function() {
        return EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_UP | 
                EchoRender.ComponentSync.FOCUS_PERMIT_ARROW_DOWN;
    },

    renderAdd: function(update, parentElement) {
        this._textComponentElement = document.createElement("input");
        this._textComponentElement.id = this.component.renderId;
        this._textComponentElement.type = this._type;
        var maximumLength = this.component.render("maximumLength", -1);
        if (maximumLength >= 0) {
            this._textComponentElement.maxLength = maximumLength;
        }
        this._renderStyle(this._textComponentElement);
        this._addEventHandlers(this._textComponentElement);
        if (this.component.get("text")) {
            this._textComponentElement.value = this.component.get("text");
        }
        parentElement.appendChild(this._textComponentElement);
    },

    sanitizeInput: function() {
        // allow all input
    }
});

/**
 * Component rendering peer: PasswordField
 */
EchoAppRender.PasswordFieldSync = Core.extend(EchoAppRender.TextFieldSync, {
    
    $load: function() {
        EchoRender.registerPeer("PasswordField", this);
    },
    
    _type: "password"
});
